@@ -371,6 +371,8 @@ void GC_print_callers GC_PROTO((struct callinfo info[NFRAMES]));
 #		define GC_MACH_THREAD_STATE_FLAVOR PPC_THREAD_STATE
 #	elif defined(I386)
 #		define GC_MACH_THREAD_STATE_FLAVOR i386_THREAD_STATE
+#	elif defined(X86_64)
+#		define GC_MACH_THREAD_STATE_FLAVOR x86_THREAD_STATE64
 #	else
 #		define GC_MACH_THREAD_STATE_FLAVOR MACHINE_THREAD_STATE
 #	endif
@@ -940,6 +942,7 @@ struct _GC_arrays {
 #   define MAP_SIZE MAP_ENTRIES
 #   define CPP_MAX_OFFSET (OFFSET_TOO_BIG - 1)	
 #   define MAX_OFFSET ((word)CPP_MAX_OFFSET)
+#   define INIT_MAP(map) memset((map), OBJ_INVALID, MAP_SIZE)
     /* The following are used only if GC_all_interior_ptrs != 0 */
 # 	define VALID_OFFSET_SZ \
 	  (CPP_MAX_OFFSET > WORDS_TO_BYTES(CPP_MAXOBJSZ)? \
@@ -1961,6 +1964,8 @@ void GC_err_puts GC_PROTO((GC_CONST char *s));
        /* SPARC/Linux doesn't properly define SIGPWR in <signal.h>.
         * It is aliased to SIGLOST in asm/signal.h, though.		*/
 #      define SIG_SUSPEND SIGLOST
+#    elif defined(NACL)
+#	define SIG_SUSPEND 0
 #    else
        /* Linuxthreads itself uses SIGUSR1 and SIGUSR2.			*/
 #      define SIG_SUSPEND SIGPWR
