@@ -41,29 +41,57 @@ namespace System.Windows.Forms
 		#region Instance Variables
 		private int image_index = -1;
 		private bool is_checked = false;
-		private bool is_focused = false;
 		private int state_image_index = -1;
 		private ListViewSubItemCollection sub_items;
 		private object tag;
 		private bool use_item_style = true;
-#if NET_2_0
+		int display_index = -1;			// actual position in ListView
 		private ListViewGroup group = null;
 		private string name = String.Empty;
 		private string image_key = String.Empty;
-		int index;			// cached index for VirtualMode
-#endif
-		Rectangle bounds;
+		string tooltip_text = String.Empty;
+		int indent_count;
+		Point position = new Point (-1, -1);		// cached to mimic .Net behaviour	
+		Rectangle bounds = Rectangle.Empty;
 		Rectangle checkbox_rect;	// calculated by CalcListViewItem method
 		Rectangle icon_rect;
 		Rectangle item_rect;
 		Rectangle label_rect;
 		ListView owner;
 		Font font;
+		Font hot_font;			// cached font for hot tracking
 		bool selected;
 
 		internal int row;
 		internal int col;
 
+	
+		#region UIA Framework: Methods, Properties and Events
+
+		internal event EventHandler UIATextChanged;
+	
+		internal event LabelEditEventHandler UIASubItemTextChanged;
+
+		internal void OnUIATextChanged ()
+		{
+			if (UIATextChanged != null)
+				UIATextChanged (this, EventArgs.Empty);
+		}
+
+		internal void OnUIASubItemTextChanged (LabelEditEventArgs args)
+		{
+			//If our index is 0 we also generate TextChanged for the ListViewItem
+			//because ListViewItem.Text is the same as ListViewItem.SubItems [0].Text
+			if (args.Item == 0)
+				OnUIATextChanged ();
+
+			if (UIASubItemTextChanged != null)
+				UIASubItemTextChanged (this, args);
+		}
+
+		#endregion // UIA Framework: Methods, Properties and Events
+
+
 		#endregion Instance Variables
 
 		#region Public Constructors
@@ -81,7 +109,7 @@ namespace System.Windows.Forms
 
 		public ListViewItem (ListViewItem.ListViewSubItem [] subItems, int imageIndex)
 		{
-			this.sub_items = new ListViewSubItemCollection (this);
+			this.sub_items = new ListViewSubItemCollection (this, null);
 			for (int i = 0; i < subItems.Length; i++)
 				sub_items.Add (subItems [i]);
 			this.image_index = imageIndex;
@@ -90,13 +118,12 @@ namespace System.Windows.Forms
 		public ListViewItem (string text, int imageIndex)
 		{
 			this.image_index = imageIndex;
-			this.sub_items = new ListViewSubItemCollection (this);
-			this.sub_items.Add (text);
+			this.sub_items = new ListViewSubItemCollection (this, text);
 		}
 
 		public ListViewItem (string [] items, int imageIndex)
 		{
-			this.sub_items = new ListViewSubItemCollection (this);
+			this.sub_items = new ListViewSubItemCollection (this, null);
 			if (items != null) {
 				for (int i = 0; i < items.Length; i++)
 					sub_items.Add (new ListViewSubItem (this, items [i]));
@@ -112,7 +139,6 @@ namespace System.Windows.Forms
 			this.font = font;
 		}
 
-#if NET_2_0
 		public ListViewItem(string[] items, string imageKey) : this(items)
 		{
 			this.ImageKey = imageKey;
@@ -125,7 +151,7 @@ namespace System.Windows.Forms
 
 		public ListViewItem(ListViewSubItem[] subItems, string imageKey)
 		{
-			this.sub_items = new ListViewSubItemCollection (this);
+			this.sub_items = new ListViewSubItemCollection (this, null);
 			for (int i = 0; i < subItems.Length; i++)
 				this.sub_items.Add (subItems [i]);
 			this.ImageKey = imageKey;
@@ -141,71 +167,75 @@ namespace System.Windows.Forms
 
 		public ListViewItem(ListViewGroup group) : this()
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(string text, ListViewGroup group) : this(text)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(string[] items, ListViewGroup group) : this(items)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(ListViewSubItem[] subItems, int imageIndex, ListViewGroup group)
 			: this(subItems, imageIndex)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(ListViewSubItem[] subItems, string imageKey, ListViewGroup group)
 			: this(subItems, imageKey)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(string text, int imageIndex, ListViewGroup group)
 			: this(text, imageIndex)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(string text, string imageKey, ListViewGroup group)
 			: this(text, imageKey)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(string[] items, int imageIndex, ListViewGroup group)
 			: this(items, imageIndex)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(string[] items, string imageKey, ListViewGroup group)
 			: this(items, imageKey)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(string[] items, int imageIndex, Color foreColor, Color backColor,
 				Font font, ListViewGroup group)
 			: this(items, imageIndex, foreColor, backColor, font)
 		{
-			this.group = group;
+			Group = group;
 		}
 
 		public ListViewItem(string[] items, string imageKey, Color foreColor, Color backColor,
 				Font font, ListViewGroup group)
 			: this(items, imageKey, foreColor, backColor, font)
 		{
-			this.group = group;
+			Group = group;
 		}
-#endif
 		#endregion	// Public Constructors
 
+		protected ListViewItem (SerializationInfo info, StreamingContext context)
+		{
+			Deserialize (info, context);
+		}
+
 		#region Public Instance Properties
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		public Color BackColor {
@@ -250,10 +280,8 @@ namespace System.Windows.Forms
 						is_checked = new_value == CheckState.Checked;
 						Invalidate ();
 
-#if NET_2_0
 						ItemCheckedEventArgs args = new ItemCheckedEventArgs (this);
 						owner.OnItemChecked (args);
-#endif
 					}
 				} else
 					is_checked = value;
@@ -263,17 +291,34 @@ namespace System.Windows.Forms
 		[Browsable (false)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		public bool Focused {
-			get { return is_focused; }
+			get { 
+				if (owner == null)
+					return false;
+
+				// In virtual mode the checks are always done using indexes
+				if (owner.VirtualMode)
+					return Index == owner.focused_item_index;
+
+				// Light check
+				return owner.FocusedItem == this;
+
+			}
 			set { 	
-				if (is_focused == value)
+				if (owner == null)
 					return;
 
-				is_focused = value; 
+				if (Focused == value)
+					return;
 
-				Invalidate ();
-				if (owner != null)
-					Layout ();
-				Invalidate ();
+				ListViewItem prev_focused_item = owner.FocusedItem;
+				if (prev_focused_item != null)
+					prev_focused_item.UpdateFocusedState ();
+					
+				owner.focused_item_index = value ? Index : -1;
+				if (value)
+					owner.OnUIAFocusedItemChanged ();
+
+				UpdateFocusedState ();
 			}
 		}
 
@@ -293,6 +338,7 @@ namespace System.Windows.Forms
 					return;
 
 				font = value; 
+				hot_font = null;
 
 				if (owner != null)
 					Layout ();
@@ -319,12 +365,8 @@ namespace System.Windows.Forms
 		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design,
 			 typeof (System.Drawing.Design.UITypeEditor))]
 		[Localizable (true)]
-#if NET_2_0
 		[RefreshProperties (RefreshProperties.Repaint)]
-		// [TypeConverter (typeof (NoneExcludedImageIndexConverter))]
-#else
-		[TypeConverter (typeof (ImageIndexConverter))]
-#endif
+		[TypeConverter (typeof (NoneExcludedImageIndexConverter))]
 		public int ImageIndex {
 			get { return image_index; }
 			set {
@@ -332,9 +374,7 @@ namespace System.Windows.Forms
 					throw new ArgumentException ("Invalid ImageIndex. It must be greater than or equal to -1.");
 				
 				image_index = value;
-#if NET_2_0
 				image_key = String.Empty;
-#endif
 
 				if (owner != null)
 					Layout ();
@@ -342,14 +382,13 @@ namespace System.Windows.Forms
 			}
 		}
 
-#if NET_2_0
 		[DefaultValue ("")]
 		[LocalizableAttribute (true)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design,
 			 typeof (System.Drawing.Design.UITypeEditor))]
 		[RefreshProperties (RefreshProperties.Repaint)]
-		// XXX [TypeConverter (typeof (ImageKeyConverter))
+		[TypeConverter (typeof (ImageKeyConverter))]
 		public string ImageKey {
 			get {
 				return image_key;
@@ -363,7 +402,6 @@ namespace System.Windows.Forms
 				Invalidate ();
 			}
 		}
-#endif
 
 		[Browsable (false)]
 		public ImageList ImageList {
@@ -377,17 +415,35 @@ namespace System.Windows.Forms
 			}
 		}
 
+		[DefaultValue (0)]
+		public int IndentCount {
+			get {
+				return indent_count;
+			}
+			set {
+				if (value < 0)
+					throw new ArgumentOutOfRangeException ("value");
+
+				if (value == indent_count)
+					return;
+
+				indent_count = value;
+				Invalidate ();
+			}
+		}
+
 		[Browsable (false)]
 		public int Index {
 			get {
 				if (owner == null)
 					return -1;
-#if NET_2_0
 				if (owner.VirtualMode)
-					return index;
-#endif
-				else
+					return display_index;
+
+				if (display_index == -1)
 					return owner.Items.IndexOf (this);
+
+				return owner.GetItemIndex (display_index);
 			}
 		}
 
@@ -396,7 +452,6 @@ namespace System.Windows.Forms
 			get { return owner; }
 		}
 
-#if NET_2_0
 		[Browsable (false)]
 		[Localizable (true)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
@@ -408,44 +463,80 @@ namespace System.Windows.Forms
 				name = value == null ? String.Empty : value;
 			}
 		}
-#endif
 
+		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
+		[Browsable (false)]
+		public Point Position {
+			get {
+				if (owner != null && owner.VirtualMode)
+					return owner.GetItemLocation (display_index);
+
+				if (owner != null && !owner.IsHandleCreated)
+					return new Point (-1, -1);
+
+				return position;
+			}
+			set {
+				if (owner == null || owner.View == View.Details || owner.View == View.List)
+					return;
+
+				if (owner.VirtualMode)
+					throw new InvalidOperationException ();
+
+				owner.ChangeItemLocation (display_index, value);
+			}
+		}
+
+		// When ListView uses VirtualMode, selection state info
+		// lives in the ListView, not in the item
+		// Also, in VirtualMode we can't Reset() the selection
 		[Browsable (false)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		public bool Selected {
 			get { 
+				if (owner != null && owner.VirtualMode)
+					return owner.SelectedIndices.Contains (Index);
+
 				return selected; 
 			}
 			set {
-				if (selected == value)
+				if (selected == value && owner != null && !owner.VirtualMode)
 					return;
 
-				if (owner != null) {
-					if (value && !owner.MultiSelect)
-						owner.SelectedItems.Clear ();
-					selected = value;
-					// force re-population of list
-					owner.SelectedIndices.Reset ();
-					Layout ();
-					owner.OnSelectedIndexChanged ();
+				SetSelectedCore (value);
+			}
+		}
+
+		// Expose this method as internal so we can force an update in the selection.
+		internal void SetSelectedCore (bool value)
+		{
+			if (owner != null) {
+				if (value && !owner.MultiSelect)
+					owner.SelectedIndices.Clear ();
+				if (owner.VirtualMode) {
+					if (value)
+						owner.SelectedIndices.InsertIndex (Index);
+					else
+						owner.SelectedIndices.RemoveIndex (Index);
 				} else {
 					selected = value;
+					owner.SelectedIndices.Reset (); // force re-population of list
 				}
+
+				owner.OnItemSelectionChanged (new ListViewItemSelectionChangedEventArgs (this, Index, value));
+				owner.OnSelectedIndexChanged ();
 				Invalidate ();
-			}
+			} else
+				selected = value;
 		}
 
 		[DefaultValue (-1)]
 		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design,
 			 typeof (System.Drawing.Design.UITypeEditor))]
 		[Localizable (true)]
-#if NET_2_0
 		[RefreshProperties (RefreshProperties.Repaint)]
-		// XXX [RelatedImageListAttribute ("ListView.StateImageList")]
-		// XXX [TypeConverter (typeof (NoneExcludedImageIndexConverter))]
-#else
-		[TypeConverter (typeof (ImageIndexConverter))]
-#endif
+		[RelatedImageListAttribute ("ListView.StateImageList")]
+		[TypeConverter (typeof (NoneExcludedImageIndexConverter))]
 		public int StateImageIndex {
 			get { return state_image_index; }
 			set {
@@ -457,10 +548,8 @@ namespace System.Windows.Forms
 		}
 
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
-#if NET_2_0
 		[Editor ("System.Windows.Forms.Design.ListViewSubItemCollectionEditor, " + Consts.AssemblySystem_Design,
 			 typeof (System.Drawing.Design.UITypeEditor))]
-#endif
 		public ListViewSubItemCollection SubItems {
 			get {
 				if (sub_items.Count == 0)
@@ -496,6 +585,10 @@ namespace System.Windows.Forms
 				if (owner != null)
 					Layout ();
 				Invalidate ();
+
+				//UIA Framework: Generates Text changed
+				OnUIATextChanged ();
+
 			}
 		}
 
@@ -505,33 +598,34 @@ namespace System.Windows.Forms
 			set { use_item_style = value; }
 		}
 
-#if NET_2_0
 		[LocalizableAttribute(true)]
 		[DefaultValue (null)]
 		public ListViewGroup Group {
 			get { return this.group; }
-			set
-			{
-				if (this.group != value)
-				{
-					if (value != null)
-					{
-						value.Items.Add(this);
-
-						if (this.group != null)
-							this.group.Items.Remove(this);
-
-						this.group = value;
-					}
+			set {
+				if (group != value) {
+					if (value == null)
+						group.Items.Remove (this);
 					else
-					{
-						if(this.group != null)
-						this.group.Items.Remove(this);
-					}
+						value.Items.Add (this);
+				
+					group = value;
 				}
 			}
 		}
-#endif
+
+		[DefaultValue ("")]
+		public string ToolTipText {
+			get {
+				return tooltip_text;
+			}
+			set {
+				if (value == null)
+					value = String.Empty;
+
+				tooltip_text = value;
+			}
+		}
 
 		#endregion	// Public Instance Properties
 
@@ -554,11 +648,10 @@ namespace System.Windows.Forms
 			ListViewItem clone = new ListViewItem ();
 			clone.image_index = this.image_index;
 			clone.is_checked = this.is_checked;
-			clone.is_focused = this.is_focused;
 			clone.selected = this.selected;
 			clone.font = this.font;
 			clone.state_image_index = this.state_image_index;
-			clone.sub_items = new ListViewSubItemCollection (this);
+			clone.sub_items = new ListViewSubItemCollection (this, null);
 			
 			foreach (ListViewSubItem subItem in this.sub_items)
 				clone.sub_items.Add (subItem.Text, subItem.ForeColor,
@@ -566,9 +659,8 @@ namespace System.Windows.Forms
 			clone.tag = this.tag;
 			clone.use_item_style = this.use_item_style;
 			clone.owner = null;
-#if NET_2_0
 			clone.name = name;
-#endif
+			clone.tooltip_text = tooltip_text;
 
 			return clone;
 		}
@@ -580,6 +672,15 @@ namespace System.Windows.Forms
 			}
 		}
 
+		public ListViewItem FindNearestItem (SearchDirectionHint searchDirection)
+		{
+			if (owner == null)
+				return null;
+
+			Point loc = owner.GetItemLocation (display_index);
+			return owner.FindNearestItem (searchDirection, loc);
+		}
+
 		public Rectangle GetBounds (ItemBoundsPortion portion)
 		{
 			if (owner == null)
@@ -608,7 +709,7 @@ namespace System.Windows.Forms
 				throw new ArgumentException ("Invalid value for portion.");
 			}
 
-			Point item_loc = owner.GetItemLocation (Index);
+			Point item_loc = owner.GetItemLocation (DisplayIndex);
 			rect.X += item_loc.X;
 			rect.Y += item_loc.Y;
 			return rect;
@@ -616,7 +717,19 @@ namespace System.Windows.Forms
 
 		void ISerializable.GetObjectData (SerializationInfo info, StreamingContext context)
 		{
-			// FIXME: TODO
+			Serialize (info, context);
+		}
+
+		public ListViewSubItem GetSubItemAt (int x, int y)
+		{
+			if (owner != null && owner.View != View.Details)
+				return null;
+
+			foreach (ListViewSubItem sub_item in sub_items)
+				if (sub_item.Bounds.Contains (x, y))
+					return sub_item;
+
+			return null;
 		}
 
 		public virtual void Remove ()
@@ -638,12 +751,73 @@ namespace System.Windows.Forms
 		#region Protected Methods
 		protected virtual void Deserialize (SerializationInfo info, StreamingContext context)
 		{
-			// FIXME: TODO
+			sub_items = new ListViewSubItemCollection (this, null);
+			int sub_items_count = 0;
+
+			foreach (SerializationEntry entry in info) {
+				switch (entry.Name) {
+					case "Text":
+						sub_items.Add ((string)entry.Value);
+						break;
+					case "Font":
+						font = (Font)entry.Value;
+						break;
+					case "Checked":
+						is_checked = (bool)entry.Value;
+						break;
+					case "ImageIndex":
+						image_index = (int)entry.Value;
+						break;
+					case "StateImageIndex":
+						state_image_index = (int)entry.Value;
+						break;
+					case "UseItemStyleForSubItems":
+						use_item_style = (bool)entry.Value;
+						break;
+					case "SubItemCount":
+						sub_items_count = (int)entry.Value;
+						break;
+					case "Group":
+						group = (ListViewGroup)entry.Value;
+						break;
+					case "ImageKey":
+						if (image_index == -1)
+							image_key = (string)entry.Value;
+						break;
+				}
+			}
+
+			Type subitem_type = typeof (ListViewSubItem);
+			if (sub_items_count > 0) {
+				sub_items.Clear (); // .net fixup
+				Text = info.GetString ("Text");
+				for (int i = 0; i < sub_items_count - 1; i++)
+					sub_items.Add ((ListViewSubItem)info.GetValue ("SubItem" + (i + 1), subitem_type));
+			}
+
+			// After any sub item has been added.
+			ForeColor = (Color)info.GetValue ("ForeColor", typeof (Color));
+			BackColor = (Color)info.GetValue ("BackColor", typeof (Color));
 		}
 
 		protected virtual void Serialize (SerializationInfo info, StreamingContext context)
 		{
-			// FIXME: TODO
+			info.AddValue ("Text", Text);
+			info.AddValue ("Font", Font);
+			info.AddValue ("ImageIndex", image_index);
+			info.AddValue ("Checked", is_checked);
+			info.AddValue ("StateImageIndex", state_image_index);
+			info.AddValue ("UseItemStyleForSubItems", use_item_style);
+			info.AddValue ("BackColor", BackColor);
+			info.AddValue ("ForeColor", ForeColor);
+			info.AddValue ("ImageKey", image_key);
+			info.AddValue ("Group", group);
+			if (sub_items.Count > 1) {
+				info.AddValue ("SubItemCount", sub_items.Count);
+				for (int i = 1; i < sub_items.Count; i++) {
+					info.AddValue ("SubItem" + i, sub_items [i]);
+				}
+			}
 		}
 		#endregion	// Protected Methods
 
@@ -651,7 +825,7 @@ namespace System.Windows.Forms
 		internal Rectangle CheckRectReal {
 			get {
 				Rectangle rect = checkbox_rect;
-				Point item_loc = owner.GetItemLocation (Index);
+				Point item_loc = owner.GetItemLocation (DisplayIndex);
 				rect.X += item_loc.X;
 				rect.Y += item_loc.Y;
 				return rect;
@@ -661,14 +835,46 @@ namespace System.Windows.Forms
 		Rectangle text_bounds;
 		internal Rectangle TextBounds {
 			get {
+				// Call Layout() if it hasn't been called before.
+				if (owner.VirtualMode && bounds == new Rectangle (-1, -1, -1, -1))
+					Layout ();
 				Rectangle result = text_bounds;
-				Point loc = owner.GetItemLocation (Index);
+				Point loc = owner.GetItemLocation (DisplayIndex);
 				result.X += loc.X;
 				result.Y += loc.Y;
 				return result;
 			}
 		}
 
+		internal int DisplayIndex {
+			get {
+				// Special case for Details view
+				// and no columns (which means no Layout at all)
+				if (display_index == -1)
+					return owner.Items.IndexOf (this);
+
+				return display_index;
+			}
+			set {
+				display_index = value;
+			}
+		}
+
+		internal bool Hot {
+			get {
+				return Index == owner.HotItemIndex;
+			}
+		}
+
+		internal Font HotFont {
+			get {
+				if (hot_font == null)
+					hot_font = new Font (Font, Font.Style | FontStyle.Underline);
+
+				return hot_font;
+			}
+		}
+
 		internal ListView Owner {
 			set {
 				if (owner == value)
@@ -678,23 +884,42 @@ namespace System.Windows.Forms
 			}
 		}
 
-#if NET_2_0
-		internal void SetIndex (int index)
+		internal void SetGroup (ListViewGroup group)
 		{
-			this.index = index;
+			this.group = group;
+		}
+
+		internal void SetPosition (Point position)
+		{
+			this.position = position;
+		}
+
+		// When focus changed, we need to invalidate area
+		// with previous layout and with the new one
+		void UpdateFocusedState ()
+		{
+			if (owner != null) {
+				Invalidate ();
+				Layout ();
+				Invalidate ();
+			}
 		}
-#endif
 
 		internal void Invalidate ()
 		{
-			if (owner == null || owner.item_control == null)
+			if (owner == null || owner.item_control == null || owner.updating)
 				return;
 
-			owner.item_control.Invalidate (Bounds);
+			// Add some padding to bounds (focused extra space, selection)
+			Rectangle rect = Bounds;
+			rect.Inflate (1, 1);
+			owner.item_control.Invalidate (rect);
 		}
 
 		internal void Layout ()
 		{
+			if (owner == null)
+				return;
 			int item_ht;
 			Rectangle total;
 			Size text_size = owner.text_size;
@@ -710,9 +935,13 @@ namespace System.Windows.Forms
 				// returns same bounding rectangles for Item and Entire
 				// values in the case of Details view.
 
+				int x_offset = 0;
+				if (owner.SmallImageList != null)
+					x_offset = indent_count * owner.SmallImageList.ImageSize.Width;
+
 				// Handle reordered column
 				if (owner.Columns.Count > 0)
-					checkbox_rect.X = owner.Columns[0].Rect.X;
+					checkbox_rect.X = owner.Columns[0].Rect.X + x_offset;
 
 				icon_rect = label_rect = Rectangle.Empty;
 				icon_rect.X = checkbox_rect.Right + 2;
@@ -724,14 +953,14 @@ namespace System.Windows.Forms
 				label_rect.Height = icon_rect.Height = item_ht;
 				checkbox_rect.Y = item_ht - checkbox_rect.Height;
 
-				label_rect.X = icon_rect.Right + 1;
+				label_rect.X = icon_rect.Width > 0 ? icon_rect.Right + 1 : icon_rect.Right;
 
 				if (owner.Columns.Count > 0)
 					label_rect.Width = owner.Columns[0].Wd - label_rect.X + checkbox_rect.X;
 				else
 					label_rect.Width = text_size.Width;
 
-				SizeF text_sz = owner.DeviceContext.MeasureString (Text, Font);
+				SizeF text_sz = TextRenderer.MeasureString (Text, Font);
 				text_bounds = label_rect;
 				text_bounds.Width = (int) text_sz.Width;
 
@@ -745,18 +974,25 @@ namespace System.Windows.Forms
 					item_rect.Width += owner.Columns [i].Wd;
 					bounds.Width += owner.Columns [i].Wd;
 				}
+
+				// Bounds for sub items
+				int n = Math.Min (owner.Columns.Count, sub_items.Count);
+				for (int i = 0; i < n; i++) {
+					Rectangle col_rect = owner.Columns [i].Rect;
+					sub_items [i].SetBounds (col_rect.X, 0, col_rect.Width, item_ht);
+				}
 				break;
 
 			case View.LargeIcon:
 				label_rect = icon_rect = Rectangle.Empty;
 
-				SizeF sz = owner.DeviceContext.MeasureString (Text, Font);
+				SizeF sz = TextRenderer.MeasureString (Text, Font);
 				if ((int) sz.Width > text_size.Width) {
-					if (Focused) {
+					if (Focused && owner.InternalContainsFocus) {
 						int text_width = text_size.Width;
 						StringFormat format = new StringFormat ();
 						format.Alignment = StringAlignment.Center;
-						sz = owner.DeviceContext.MeasureString (Text, Font, text_width, format);
+						sz = TextRenderer.MeasureString (Text, Font, text_width, format);
 						text_size.Height = (int) sz.Height;
 					} else
 						text_size.Height = 2 * (int) sz.Height;
@@ -811,8 +1047,10 @@ namespace System.Windows.Forms
 				total = Rectangle.Union (item_rect, checkbox_rect);
 				bounds.Size = total.Size;
 				break;
-#if NET_2_0
 			case View.Tile:
+				if (!Application.VisualStylesEnabled)
+					goto case View.LargeIcon;
+
 				label_rect = icon_rect = Rectangle.Empty;
 
 				if (owner.LargeImageList != null) {
@@ -821,23 +1059,24 @@ namespace System.Windows.Forms
 				}
 
 				int separation = 2;
-				SizeF tsize = owner.DeviceContext.MeasureString (Text, Font);
+				SizeF tsize = TextRenderer.MeasureString (Text, Font);
+				int main_item_height = (int) Math.Ceiling (tsize.Height);
+				int main_item_width = (int) Math.Ceiling (tsize.Width);
+				sub_items [0].bounds.Height = main_item_height;
 
 				// Set initial values for subitem's layout
-				int total_height = (int)Math.Ceiling (tsize.Height);
-				int max_subitem_width = (int)Math.Ceiling (tsize.Width);
-				SubItems [0].bounds.Height = total_height;
+				int total_height = main_item_height;
+				int max_subitem_width = main_item_width;
 			
-				int count = Math.Min (owner.Columns.Count, SubItems.Count);
+				int count = Math.Min (owner.Columns.Count, sub_items.Count);
 				for (int i = 1; i < count; i++) { // Ignore first column and first subitem
-					ListViewSubItem sub_item = SubItems [i];
+					ListViewSubItem sub_item = sub_items [i];
 					if (sub_item.Text == null || sub_item.Text.Length == 0)
 						continue;
 
-					tsize = owner.DeviceContext.MeasureString (sub_item.Text, sub_item.Font);
+					tsize = TextRenderer.MeasureString (sub_item.Text, sub_item.Font);
 				
 					int width = (int)Math.Ceiling (tsize.Width);
-				
 					if (width > max_subitem_width)
 						max_subitem_width = width;
 				
@@ -845,18 +1084,21 @@ namespace System.Windows.Forms
 					total_height += height + separation;
 				
 					sub_item.bounds.Height = height;
-			
 				}
 
+				max_subitem_width = Math.Min (max_subitem_width, owner.TileSize.Width - (icon_rect.Width + 4));
 				label_rect.X = icon_rect.Right + 4;
 				label_rect.Y = owner.TileSize.Height / 2 - total_height / 2;
 				label_rect.Width = max_subitem_width;
 				label_rect.Height = total_height;
 			
-				// Second pass for assigning bounds. This time take first subitem into account.
-				int current_y = label_rect.Y;
-				for (int j = 0; j < count; j++) {
-					ListViewSubItem sub_item = SubItems [j];
+				// Main item - always set bounds for it
+				sub_items [0].SetBounds (label_rect.X, label_rect.Y, max_subitem_width, sub_items [0].bounds.Height);
+
+				// Second pass to assign bounds for every sub item
+				int current_y = sub_items [0].bounds.Bottom + separation;
+				for (int j = 1; j < count; j++) {
+					ListViewSubItem sub_item = sub_items [j];
 					if (sub_item.Text == null || sub_item.Text.Length == 0)
 						continue;
 
@@ -867,7 +1109,6 @@ namespace System.Windows.Forms
 				item_rect = Rectangle.Union (icon_rect, label_rect);
 				bounds.Size = item_rect.Size;
 				break;
-#endif
 			}
 			
 		}
@@ -882,19 +1123,34 @@ namespace System.Windows.Forms
 		[TypeConverter (typeof(ListViewSubItemConverter))]
 		public class ListViewSubItem
 		{
-			private Color back_color;
-			private Font font;
-			private Color fore_color;
+			[NonSerialized]
 			internal ListViewItem owner;
 			private string text = string.Empty;
-#if NET_2_0
-			private string name = String.Empty;
-			private object tag;
+			private string name;
+			private object userData;
+			private SubItemStyle style;
+			[NonSerialized]
 			internal Rectangle bounds;
-#endif
+
+		
+			#region UIA Framework: Methods, Properties and Events
+		
+			[field:NonSerialized]
+			internal event EventHandler UIATextChanged;
+
+			private void OnUIATextChanged ()
+			{
+				if (UIATextChanged != null)
+					UIATextChanged (this, EventArgs.Empty);
+			}
+
+			#endregion // UIA Framework: Methods, Properties and Events
+
 			
 			#region Public Constructors
 			public ListViewSubItem ()
+				: this (null, string.Empty, Color.Empty,
+					Color.Empty, null)
 			{
 			}
 
@@ -909,28 +1165,27 @@ namespace System.Windows.Forms
 			{
 				this.owner = owner;
 				Text = text;
-				this.fore_color = foreColor;
-				this.back_color = backColor;
-				this.font = font;
+				this.style = new SubItemStyle (foreColor,
+					backColor, font);
 			}
 			#endregion // Public Constructors
 
 			#region Public Instance Properties
 			public Color BackColor {
 				get {
-					if (this.back_color != Color.Empty)
-						return this.back_color;
+					if (style.backColor != Color.Empty)
+						return style.backColor;
 					if (this.owner != null && this.owner.ListView != null)
 						return this.owner.ListView.BackColor;
 					return ThemeEngine.Current.ColorWindow;
 				}
 				set { 
-					back_color = value; 
+					style.backColor = value;
 					Invalidate ();
 				}
 			}
 
-#if NET_2_0
+			[Browsable (false)]
 			public Rectangle Bounds {
 				get {
 					Rectangle retval = bounds;
@@ -942,47 +1197,47 @@ namespace System.Windows.Forms
 					return retval;
 				}
 			}
-#endif
 
 			[Localizable (true)]
 			public Font Font {
 				get {
-					if (font != null)
-						return font;
+					if (style.font != null)
+						return style.font;
 					else if (owner != null)
 						return owner.Font;
 					return ThemeEngine.Current.DefaultFont;
 				}
-				set { 
-					if (font == value)
+				set {
+					if (style.font == value)
 						return;
-					font = value; 
+					style.font = value; 
 					Invalidate ();
-				    }
+				}
 			}
 
 			public Color ForeColor {
 				get {
-					if (this.fore_color != Color.Empty)
-						return this.fore_color;
+					if (style.foreColor != Color.Empty)
+						return style.foreColor;
 					if (this.owner != null && this.owner.ListView != null)
 						return this.owner.ListView.ForeColor;
 					return ThemeEngine.Current.ColorWindowText;
 				}
-				set { 
-					fore_color = value; 
+				set {
+					style.foreColor = value;
 					Invalidate ();
 				}
 			}
 
-#if NET_2_0
 			[Localizable (true)]
 			public string Name {
 				get {
+					if (name == null)
+						return string.Empty;
 					return name;
 				}
 				set {
-					name = value == null ? String.Empty : value;
+					name = value;
 				}
 			}
 
@@ -992,13 +1247,12 @@ namespace System.Windows.Forms
 			[Localizable (false)]
 			public object Tag {
 				get {
-					return tag;
+					return userData;
 				}
 				set {
-					tag = value;
+					userData = value;
 				}
 			}
-#endif
 
 			[Localizable (true)]
 			public string Text {
@@ -1013,6 +1267,9 @@ namespace System.Windows.Forms
 					      	text = value; 
 
 					Invalidate ();
+
+					// UIA Framework: Generates SubItem TextChanged
+					OnUIATextChanged ();
 				    }
 			}
 			#endregion // Public Instance Properties
@@ -1020,9 +1277,7 @@ namespace System.Windows.Forms
 			#region Public Methods
 			public void ResetStyle ()
 			{
-				font = ThemeEngine.Current.DefaultFont;
-				back_color = ThemeEngine.Current.DefaultControlBackColor;
-				fore_color = ThemeEngine.Current.DefaultControlForeColor;
+				style.Reset ();
 				Invalidate ();
 			}
 
@@ -1032,17 +1287,22 @@ namespace System.Windows.Forms
 			}
 			#endregion // Public Methods
 
-			
 			#region Private Methods
 			private void Invalidate ()
 			{
 				if (owner == null || owner.owner == null)
 					return;
 
-				owner.owner.Invalidate ();
+				owner.Invalidate ();
+			}
+
+			[OnDeserialized]
+			void OnDeserialized (StreamingContext context)
+			{
+				name = null;
+				userData = null;
 			}
 
-#if NET_2_0
 			internal int Height {
 				get {
 					return bounds.Height;
@@ -1053,8 +1313,34 @@ namespace System.Windows.Forms
 			{
 				bounds = new Rectangle (x, y, width, height);
 			}
-#endif
+
 			#endregion // Private Methods
+
+			[Serializable]
+			class SubItemStyle
+			{
+				public SubItemStyle ()
+				{
+				}
+
+				public SubItemStyle (Color foreColor, Color backColor, Font font)
+				{
+					this.foreColor = foreColor;
+					this.backColor = backColor;
+					this.font = font;
+				}
+
+				public void Reset ()
+				{
+					foreColor = Color.Empty;
+					backColor = Color.Empty;
+					font = null;
+				}
+
+				public Color backColor;
+				public Color foreColor;
+				public Font font;
+			}
 		}
 
 		public class ListViewSubItemCollection : IList, ICollection, IEnumerable
@@ -1063,13 +1349,18 @@ namespace System.Windows.Forms
 			internal ListViewItem owner;
 
 			#region Public Constructors
-			public ListViewSubItemCollection (ListViewItem owner)
+			public ListViewSubItemCollection (ListViewItem owner) : this (owner, owner.Text)
 			{
-				this.owner = owner;
-				this.list = new ArrayList ();
  			}
 			#endregion // Public Constructors
 
+			internal ListViewSubItemCollection (ListViewItem owner, string text)
+			{
+				this.owner = owner;
+				this.list = new ArrayList ();
+				if (text != null)
+					Add (text);
+			}
 			#region Public Properties
 			[Browsable (false)]
 			public int Count {
@@ -1083,12 +1374,13 @@ namespace System.Windows.Forms
 			public ListViewSubItem this [int index] {
 				get { return (ListViewSubItem) list [index]; }
 				set { 
-					value.owner = this.owner;
+					value.owner = owner;
 					list [index] = value;
+					owner.Layout ();
+					owner.Invalidate ();
 				}
 			}
 
-#if NET_2_0
 			public virtual ListViewSubItem this [string key] {
 				get {
 					int idx = IndexOfKey (key);
@@ -1098,7 +1390,6 @@ namespace System.Windows.Forms
 					return (ListViewSubItem) list [idx];
 				}
 			}
-#endif
 
 			bool ICollection.IsSynchronized {
 				get { return list.IsSynchronized; }
@@ -1125,25 +1416,24 @@ namespace System.Windows.Forms
 			#region Public Methods
 			public ListViewSubItem Add (ListViewSubItem item)
 			{
-				item.owner = this.owner;
-				list.Add (item);
+				AddSubItem (item);
+				owner.Layout ();
+				owner.Invalidate ();
 				return item;
 			}
 
 			public ListViewSubItem Add (string text)
 			{
-				ListViewSubItem item = new ListViewSubItem (this.owner, text);
-				list.Add (item);
-				return item;
+				ListViewSubItem item = new ListViewSubItem (owner, text);
+				return Add (item);
 			}
 
 			public ListViewSubItem Add (string text, Color foreColor,
 						    Color backColor, Font font)
 			{
-				ListViewSubItem item = new ListViewSubItem (this.owner, text,
+				ListViewSubItem item = new ListViewSubItem (owner, text,
 									    foreColor, backColor, font);
-				list.Add (item);
-				return item;
+				return Add (item);
 			}
 
 			public void AddRange (ListViewSubItem [] items)
@@ -1154,8 +1444,10 @@ namespace System.Windows.Forms
 				foreach (ListViewSubItem item in items) {
 					if (item == null)
 						continue;
-					this.Add (item);
+					AddSubItem (item);
 				}
+				owner.Layout ();
+				owner.Invalidate ();
 			}
 
 			public void AddRange (string [] items)
@@ -1166,8 +1458,10 @@ namespace System.Windows.Forms
 				foreach (string item in items) {
 					if (item == null)
 						continue;
-					this.Add (item);
+					AddSubItem (new ListViewSubItem (owner, item));
 				}
+				owner.Layout ();
+				owner.Invalidate ();
 			}
 
 			public void AddRange (string [] items, Color foreColor,
@@ -1179,8 +1473,20 @@ namespace System.Windows.Forms
 				foreach (string item in items) {
 					if (item == null)
 						continue;
-					this.Add (item, foreColor, backColor, font);
+
+					AddSubItem (new ListViewSubItem (owner, item, foreColor, backColor, font));
 				}
+				owner.Layout ();
+				owner.Invalidate ();
+			}
+
+			void AddSubItem (ListViewSubItem subItem)
+			{
+				subItem.owner = owner;
+				list.Add (subItem);
+
+				//UIA Framework
+				subItem.UIATextChanged += OnUIASubItemTextChanged;
 			}
 
 			public void Clear ()
@@ -1188,17 +1494,15 @@ namespace System.Windows.Forms
 				list.Clear ();
 			}
 
-			public bool Contains (ListViewSubItem item)
+			public bool Contains (ListViewSubItem subItem)
 			{
-				return list.Contains (item);
+				return list.Contains (subItem);
 			}
 
-#if NET_2_0
 			public virtual bool ContainsKey (string key)
 			{
 				return IndexOfKey (key) != -1;
 			}
-#endif
 
 			public IEnumerator GetEnumerator ()
 			{
@@ -1218,6 +1522,8 @@ namespace System.Windows.Forms
 
 				ListViewSubItem sub_item = (ListViewSubItem) item;
 				sub_item.owner = this.owner;
+				//UIA Framework
+				sub_item.UIATextChanged += OnUIASubItemTextChanged;
 				return list.Add (sub_item);
 			}
 
@@ -1262,7 +1568,6 @@ namespace System.Windows.Forms
 				return list.IndexOf (subItem);
 			}
 
-#if NET_2_0
 			public virtual int IndexOfKey (string key)
 			{
 				if (key == null || key.Length == 0)
@@ -1276,33 +1581,55 @@ namespace System.Windows.Forms
 
 				return -1;
 			}
-#endif
 
 			public void Insert (int index, ListViewSubItem item)
 			{
 				item.owner = this.owner;
 				list.Insert (index, item);
+				owner.Layout ();
+				owner.Invalidate ();
+
+				//UIA Framework
+				item.UIATextChanged += OnUIASubItemTextChanged;
 			}
 
 			public void Remove (ListViewSubItem item)
 			{
 				list.Remove (item);
+				owner.Layout ();
+				owner.Invalidate ();
+
+				//UIA Framework
+				item.UIATextChanged -= OnUIASubItemTextChanged;
 			}
 
-#if NET_2_0
 			public virtual void RemoveByKey (string key)
 			{
 				int idx = IndexOfKey (key);
 				if (idx != -1)
 					RemoveAt (idx);
 			}
-#endif
 
 			public void RemoveAt (int index)
 			{
+				//UIA Framework
+				if (index >= 0 && index < list.Count)
+					((ListViewSubItem) list [index]).UIATextChanged -= OnUIASubItemTextChanged;
+
 				list.RemoveAt (index);
+
 			}
 			#endregion // Public Methods
+			#region UIA Event Handler
+			
+			private void OnUIASubItemTextChanged (object sender, EventArgs args)
+			{
+				owner.OnUIASubItemTextChanged (new LabelEditEventArgs (list.IndexOf (sender)));
+			}
+
+			#endregion
+
+
 		}
 		#endregion // Subclasses
 	}
