@@ -73,7 +73,10 @@ namespace System.Windows.Forms
 			
 				return this.drop_down;
 			}
-			set { this.drop_down = value; }
+			set { 
+				this.drop_down = value;
+				this.drop_down.OwnerItem = this;
+			}
 		}
 
 		[Browsable (false)]
@@ -129,11 +132,11 @@ namespace System.Windows.Forms
 			if (this.drop_down == null || !this.DropDown.Visible)
 				return;
 
+			// OnDropDownHide is called before actually closing DropDown
+			this.OnDropDownHide (EventArgs.Empty);
 			this.DropDown.Close (ToolStripDropDownCloseReason.CloseCalled);
 			this.is_pressed = false;
 			this.Invalidate ();
-			this.OnDropDownHide (EventArgs.Empty);
-			this.OnDropDownClosed (EventArgs.Empty);
 		}
 
 		public void ShowDropDown ()
@@ -152,7 +155,6 @@ namespace System.Windows.Forms
 			
 			this.Invalidate ();
 			this.DropDown.Show (this.DropDownLocation);
-			this.OnDropDownOpened (EventArgs.Empty);
 		}
 		#endregion
 
@@ -224,6 +226,11 @@ namespace System.Windows.Forms
 		protected override void OnFontChanged (EventArgs e)
 		{
 			base.OnFontChanged (e);
+
+			// don't use DropDown directly, since doing that
+			// would created the DropDown control
+			if (drop_down != null)
+				drop_down.Font = Font;
 		}
 
 		protected override void OnRightToLeftChanged (EventArgs e)
@@ -315,7 +322,7 @@ namespace System.Windows.Forms
 			base.Dismiss (reason);
 		}
 
-		internal override void HandleClick (EventArgs e)
+		internal override void HandleClick (int mouse_clicks, EventArgs e)
 		{
 			OnClick (e);
 		}
@@ -325,11 +332,11 @@ namespace System.Windows.Forms
 			if (this.drop_down == null || !this.DropDown.Visible)
 				return;
 
+			// OnDropDownHide is called before actually closing DropDown
+			this.OnDropDownHide (EventArgs.Empty);
 			this.DropDown.Close (reason);
 			this.is_pressed = false;
 			this.Invalidate ();
-			this.OnDropDownHide (EventArgs.Empty);
-			this.OnDropDownClosed (EventArgs.Empty);
 		}
 		
 		private void DropDown_ItemAdded (object sender, ToolStripItemEventArgs e)
