@@ -21,6 +21,16 @@ namespace Mono.Debugger.Soft
 		public int MinorVersion {
 			get; set;
 		}
+
+		/*
+		 * Check that this version is at least major:minor
+		 */
+		public bool AtLeast (int major, int minor) {
+			if ((MajorVersion > major) || ((MajorVersion == major && MinorVersion >= minor)))
+				return true;
+			else
+				return false;
+		}
 	}
 
 	class DebugInfo {
@@ -103,6 +113,48 @@ namespace Mono.Debugger.Soft
 		SINGLE_THREADED = 0x2
 	}
 
+	enum ElementType {
+		End		 = 0x00,
+		Void		= 0x01,
+		Boolean	 = 0x02,
+		Char		= 0x03,
+		I1		  = 0x04,
+		U1		  = 0x05,
+		I2		  = 0x06,
+		U2		  = 0x07,
+		I4		  = 0x08,
+		U4		  = 0x09,
+		I8		  = 0x0a,
+		U8		  = 0x0b,
+		R4		  = 0x0c,
+		R8		  = 0x0d,
+		String	  = 0x0e,
+		Ptr		 = 0x0f,
+		ByRef	   = 0x10,
+		ValueType   = 0x11,
+		Class	   = 0x12,
+		Var        = 0x13,
+		Array	   = 0x14,
+		GenericInst = 0x15,
+		TypedByRef  = 0x16,
+		I		   = 0x18,
+		U		   = 0x19,
+		FnPtr	   = 0x1b,
+		Object	  = 0x1c,
+		SzArray	 = 0x1d,
+		MVar       = 0x1e,
+		CModReqD	= 0x1f,
+		CModOpt	 = 0x20,
+		Internal	= 0x21,
+		Modifier	= 0x40,
+		Sentinel	= 0x41,
+		Pinned	  = 0x45,
+
+		Type		= 0x50,
+		Boxed	   = 0x51,
+		Enum		= 0x55
+	}
+
 	class ValueImpl {
 		public ElementType Type; /* or one of the VALUE_TYPE_ID constants */
 		public long Objid;
@@ -234,7 +286,8 @@ namespace Mono.Debugger.Soft
 		INVALID_ARGUMENT = 102,
 		ERR_UNLOADED = 103,
 		ERR_NO_INVOCATION = 104,
-		ABSENT_INFORMATION = 105
+		ABSENT_INFORMATION = 105,
+		NO_SEQ_POINT_AT_IL_OFFSET = 106
 	}
 
 	public class ErrorHandlerEventArgs : EventArgs {
@@ -265,7 +318,7 @@ namespace Mono.Debugger.Soft
 		 * with newer runtimes, and vice versa.
 		 */
 		public const int MAJOR_VERSION = 2;
-		public const int MINOR_VERSION = 2;
+		public const int MINOR_VERSION = 3;
 
 		enum WPSuspendPolicy {
 			NONE = 0,
@@ -303,7 +356,8 @@ namespace Mono.Debugger.Soft
 			BREAKPOINT = 10,
 			STEP = 11,
 			TYPE_LOAD = 12,
-			EXCEPTION = 13
+			EXCEPTION = 13,
+			KEEPALIVE = 14
 		}
 
 		enum ModifierKind {
@@ -324,7 +378,8 @@ namespace Mono.Debugger.Soft
 			DISPOSE = 6,
 			INVOKE_METHOD = 7,
 			SET_PROTOCOL_VERSION = 8,
-			ABORT_INVOKE = 9
+			ABORT_INVOKE = 9,
+			SET_KEEPALIVE = 10
 		}
 
 		enum CmdEvent {
@@ -337,7 +392,9 @@ namespace Mono.Debugger.Soft
 			GET_STATE = 3,
 			GET_INFO = 4,
 			/* FIXME: Merge into GET_INFO when the major protocol version is increased */
-			GET_ID = 5
+			GET_ID = 5,
+			/* Ditto */
+			GET_TID = 6
 		}
 
 		enum CmdEventRequest {
@@ -394,7 +451,9 @@ namespace Mono.Debugger.Soft
 			GET_FIELD_CATTRS = 11,
 			GET_PROPERTY_CATTRS = 12,
 			/* FIXME: Merge into GET_SOURCE_FILES when the major protocol version is increased */
-			GET_SOURCE_FILES_2 = 13
+			GET_SOURCE_FILES_2 = 13,
+			/* FIXME: Merge into GET_VALUES when the major protocol version is increased */
+			GET_VALUES_2 = 14
 		}
 
 		enum CmdStackFrame {
@@ -1109,6 +1168,8 @@ namespace Mono.Debugger.Soft
 								long id = r.ReadId ();
 								events [i] = new EventInfo (etype, req_id) { ThreadId = thread_id, Id = id };
 								//EventHandler.AppDomainUnload (req_id, thread_id, id);
+							} else if (kind == EventKind.KEEPALIVE) {
+								events [i] = new EventInfo (etype, req_id) { };
 							} else {
 								throw new NotImplementedException ("Unknown event kind: " + kind);
 							}
@@ -1324,6 +1385,13 @@ namespace Mono.Debugger.Soft
 			SendReceive (CommandSet.VM, (int)CmdVM.ABORT_INVOKE, new PacketWriter ().WriteId (thread).WriteInt (id));
 		}
 
+		public void SetSocketTimeouts (int send_timeout, int receive_timeout, int keepalive_interval)
+		{
+			socket.SendTimeout = send_timeout;
+			socket.ReceiveTimeout = receive_timeout;
+			SendReceive (CommandSet.VM, (int)CmdVM.SET_KEEPALIVE, new PacketWriter ().WriteId (keepalive_interval));
+		}
+
 		/*
 		 * DOMAIN
 		 */
@@ -1510,6 +1578,10 @@ namespace Mono.Debugger.Soft
 			return SendReceive (CommandSet.THREAD, (int)CmdThread.GET_ID, new PacketWriter ().WriteId (id)).ReadLong ();
 		}
 
+		public long Thread_GetTID (long id) {
+			return SendReceive (CommandSet.THREAD, (int)CmdThread.GET_TID, new PacketWriter ().WriteId (id)).ReadLong ();
+		}
+
 		/*
 		 * MODULE
 		 */
@@ -1629,9 +1701,13 @@ namespace Mono.Debugger.Soft
 			return SendReceive (CommandSet.TYPE, (int)CmdType.GET_OBJECT, new PacketWriter ().WriteId (id)).ReadId ();
 		}
 
-		public ValueImpl[] Type_GetValues (long id, long[] fields) {
+		public ValueImpl[] Type_GetValues (long id, long[] fields, long thread_id) {
 			int len = fields.Length;
-			PacketReader r = SendReceive (CommandSet.TYPE, (int)CmdType.GET_VALUES, new PacketWriter ().WriteId (id).WriteInt (len).WriteIds (fields));
+			PacketReader r;
+			if (thread_id != 0)
+				r = SendReceive (CommandSet.TYPE, (int)CmdType.GET_VALUES_2, new PacketWriter ().WriteId (id).WriteId (thread_id).WriteInt (len).WriteIds (fields));
+			else
+				r = SendReceive (CommandSet.TYPE, (int)CmdType.GET_VALUES, new PacketWriter ().WriteId (id).WriteInt (len).WriteIds (fields));
 
 			ValueImpl[] res = new ValueImpl [len];
 			for (int i = 0; i < len; ++i)
