@@ -24,8 +24,7 @@
  * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  */
 
-#if (INSIDE_CORLIB && (NET_4_0 || MOONLIGHT)) || (MONOTOUCH && !INSIDE_CORLIB) || (NET_3_5 && !NET_4_0 && !BOOTSTRAP_NET_4_0)
-
+#if (INSIDE_CORLIB && (NET_4_0 || MOONLIGHT || MOBILE)) || (!INSIDE_CORLIB && (NET_3_5 && !NET_4_0 && !MOBILE))
 using System.Runtime.CompilerServices;
 using System.Runtime.Serialization;
 
@@ -33,9 +32,9 @@ namespace System
 {
 	public sealed partial class TimeZoneInfo {
 		[SerializableAttribute]
-#if NET_4_0 || BOOTSTRAP_NET_4_0
+#if NET_4_0
 		[TypeForwardedFrom (Consts.AssemblySystemCore_3_5)]
-#elif MOONLIGHT
+#elif MOONLIGHT || MOBILE
 		[TypeForwardedFrom (Consts.AssemblySystem_Core)]
 #endif
 		public sealed class AdjustmentRule : IEquatable<TimeZoneInfo.AdjustmentRule>, ISerializable, IDeserializationCallback
