@@ -27,18 +27,19 @@
 using System;
 using System.Runtime.CompilerServices;
 
-#if !INSIDE_CORLIB && (NET_4_0 || BOOTSTRAP_NET_4_0 || MOONLIGHT)
+#if !INSIDE_CORLIB && (NET_4_0 || MOONLIGHT || MOBILE)
 
 [assembly:TypeForwardedTo (typeof(TimeZoneInfo))]
 
-#elif NET_3_5 || (MONOTOUCH && !INSIDE_CORLIB) || (MOONLIGHT && INSIDE_CORLIB)
+#elif (INSIDE_CORLIB && (NET_4_0 || MOONLIGHT || MOBILE)) || (!INSIDE_CORLIB && (NET_3_5 && !NET_4_0 && !MOBILE))
 
 using System.Collections.Generic;
 using System.Collections.ObjectModel;
+using System.Linq;
 using System.Runtime.Serialization;
 using System.Text;
 
-#if LIBC
+#if LIBC || MONODROID
 using System.IO;
 using Mono;
 #endif
@@ -47,9 +48,9 @@ using Microsoft.Win32;
 
 namespace System
 {
-#if NET_4_0 || BOOTSTRAP_NET_4_0
+#if NET_4_0
 	[TypeForwardedFrom (Consts.AssemblySystemCore_3_5)]
-#elif MOONLIGHT
+#elif MOONLIGHT || MOBILE
 	[TypeForwardedFrom (Consts.AssemblySystem_Core)]
 #endif
 	[SerializableAttribute]
@@ -83,7 +84,9 @@ namespace System
 		public static TimeZoneInfo Local {
 			get { 
 				if (local == null) {
-#if LIBC
+#if MONODROID
+					local = ZoneInfoDB.Default;
+#elif LIBC
 					try {
 						local = FindSystemTimeZoneByFileName ("Local", "/etc/localtime");	
 					} catch {
@@ -229,7 +232,7 @@ namespace System
 
 			AdjustmentRule rule = GetApplicableRule (dateTime);
 		
-			if (IsDaylightSavingTime (DateTime.SpecifyKind (dateTime, DateTimeKind.Utc)))
+			if (rule != null && IsDaylightSavingTime (DateTime.SpecifyKind (dateTime, DateTimeKind.Utc)))
 				return DateTime.SpecifyKind (dateTime + BaseUtcOffset + rule.DaylightDelta , DateTimeKind.Unspecified);
 			else
 				return DateTime.SpecifyKind (dateTime + BaseUtcOffset, DateTimeKind.Unspecified);
@@ -279,7 +282,10 @@ namespace System
 				return DateTime.SpecifyKind (dateTime - sourceTimeZone.BaseUtcOffset, DateTimeKind.Utc);
 			else {
 				AdjustmentRule rule = sourceTimeZone.GetApplicableRule (dateTime);
-				return DateTime.SpecifyKind (dateTime - sourceTimeZone.BaseUtcOffset - rule.DaylightDelta, DateTimeKind.Utc);
+				if (rule != null)
+					return DateTime.SpecifyKind (dateTime - sourceTimeZone.BaseUtcOffset - rule.DaylightDelta, DateTimeKind.Utc);
+				else
+					return DateTime.SpecifyKind (dateTime - sourceTimeZone.BaseUtcOffset, DateTimeKind.Utc);
 			}
 		}
 
@@ -320,7 +326,9 @@ namespace System
 				return FromRegistryKey(id, key);
 			}
 #endif
-#if LIBC	
+#if MONODROID
+			return ZoneInfoDB.GetTimeZone (id);
+#elif LIBC
 			string filepath = Path.Combine (TimeZoneDirectory, id);
 			return FindSystemTimeZoneByFileName (id, filepath);
 #else
@@ -477,7 +485,10 @@ namespace System
 				throw new ArgumentException ("dateTime is not an ambiguous time");
 
 			AdjustmentRule rule = GetApplicableRule (dateTime);
-			return new TimeSpan[] {baseUtcOffset, baseUtcOffset + rule.DaylightDelta};
+			if (rule != null)
+				return new TimeSpan[] {baseUtcOffset, baseUtcOffset + rule.DaylightDelta};
+			else
+				return new TimeSpan[] {baseUtcOffset, baseUtcOffset};
 		}
 
 		public TimeSpan [] GetAmbiguousTimeOffsets (DateTimeOffset dateTimeOffset)
@@ -522,7 +533,11 @@ namespace System
 					return new ReadOnlyCollection<TimeZoneInfo> (systemTimeZones);
 				}
 #endif
-#if LIBC
+#if MONODROID
+			foreach (string id in ZoneInfoDB.GetAvailableIds ()) {
+				systemTimeZones.Add (ZoneInfoDB.GetTimeZone (id));
+			}
+#elif LIBC
 				string[] continents = new string [] {"Africa", "America", "Antarctica", "Arctic", "Asia", "Atlantic", "Brazil", "Canada", "Chile", "Europe", "Indian", "Mexico", "Mideast", "Pacific", "US"};
 				foreach (string continent in continents) {
 					try {
@@ -550,7 +565,8 @@ namespace System
 		{
 			if (IsDaylightSavingTime (dateTime)) {
 				AdjustmentRule rule = GetApplicableRule (dateTime);
-				return BaseUtcOffset + rule.DaylightDelta;
+				if (rule != null)
+					return BaseUtcOffset + rule.DaylightDelta;
 			}
 			
 			return BaseUtcOffset;
@@ -601,9 +617,11 @@ namespace System
 				dateTime = ConvertTime (dateTime, TimeZoneInfo.Local, this);
 
 			AdjustmentRule rule = GetApplicableRule (dateTime);
-			DateTime tpoint = TransitionPoint (rule.DaylightTransitionEnd, dateTime.Year);
-			if (dateTime > tpoint - rule.DaylightDelta  && dateTime <= tpoint)
-				return true;
+			if (rule != null) {
+				DateTime tpoint = TransitionPoint (rule.DaylightTransitionEnd, dateTime.Year);
+				if (dateTime > tpoint - rule.DaylightDelta  && dateTime <= tpoint)
+					return true;
+			}
 				
 			return false;
 		}
@@ -658,9 +676,11 @@ namespace System
 				return false;
 
 			AdjustmentRule rule = GetApplicableRule (dateTime);
-			DateTime tpoint = TransitionPoint (rule.DaylightTransitionStart, dateTime.Year);
-			if (dateTime >= tpoint && dateTime < tpoint + rule.DaylightDelta)
-				return true;
+			if (rule != null) {
+				DateTime tpoint = TransitionPoint (rule.DaylightTransitionStart, dateTime.Year);
+				if (dateTime >= tpoint && dateTime < tpoint + rule.DaylightDelta)
+					return true;
+			}
 				
 			return false;
 		}
@@ -782,7 +802,7 @@ namespace System
 			return adjustmentRules;
 		}
 
-#if LIBC
+#if LIBC || MONODROID
 		private static bool ValidTZFile (byte [] buffer, int length)
 		{
 			StringBuilder magic = new StringBuilder ();
