@@ -45,13 +45,21 @@ namespace System.Drawing {
 			if ((htmlColor == null) || (htmlColor.Length == 0))
 				return Color.Empty;
 
-			switch (htmlColor.ToLower ()) {
+			switch (htmlColor.ToLowerInvariant ()) {
 			case "buttonface":
+			case "threedface":		
 				return SystemColors.Control;
+			case "buttonhighlight":
+			case "threedlightshadow":
+				return SystemColors.ControlLightLight;
+			case "buttonshadow":
+				return SystemColors.ControlDark;
 			case "captiontext":
 				return SystemColors.ActiveCaptionText;
 			case "threeddarkshadow":
 				return SystemColors.ControlDarkDark;
+			case "threedhighlight":
+				return SystemColors.ControlLight;
 			case "background":
 				return SystemColors.Desktop;
 			case "buttontext":
@@ -62,6 +70,7 @@ namespace System.Drawing {
 			case "lightgrey":
 				return Color.LightGray;
 			}
+			
 			TypeConverter converter = TypeDescriptor.GetConverter (typeof (Color));
 			return (Color) converter.ConvertFromString (htmlColor);
 		}
