@@ -9,7 +9,7 @@
 //   Sebastien Pouliot  <sebastien@ximian.com>
 //
 // Copyright (C) 2002 Ximian, Inc. http://www.ximian.com
-// Copyright (C) 2004-2007 Novell, Inc (http://www.novell.com)
+// Copyright (C) 2004-2008 Novell, Inc (http://www.novell.com)
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the
@@ -31,6 +31,7 @@
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
 
+using System.Collections;
 using System.ComponentModel;
 using System.Drawing.Imaging;
 using System.IO;
@@ -136,38 +137,44 @@ namespace System.Drawing
 
 			iconSize = size;
 			iconDir = original.iconDir;
-			imageData = original.imageData;
 			
-			id = UInt16.MaxValue;
 			int count = iconDir.idCount;
-			for (ushort i=0; i < count; i++) {
-				IconDirEntry ide = iconDir.idEntries [i];
-				if ((ide.height == size.Height) || (ide.width == size.Width)) {
-					id = i;
-					break;
-				}
-			}
+			if (count > 0) {
+				imageData = original.imageData;
+				id = UInt16.MaxValue;
 
-			// if a perfect match isn't found we look for the biggest icon *smaller* than specified
-			if (id == UInt16.MaxValue) {
-				int requested = Math.Min (size.Height, size.Width);
-				IconDirEntry best = iconDir.idEntries [0];
-				for (ushort i=1; i < count; i++) {
+				for (ushort i=0; i < count; i++) {
 					IconDirEntry ide = iconDir.idEntries [i];
-					if ((ide.height < requested) || (ide.width < requested)) {
-						if ((ide.height > best.height) || (ide.width > best.width))
-							id = i;
+					if ((ide.height == size.Height) || (ide.width == size.Width)) {
+						id = i;
+						break;
 					}
 				}
+
+				// if a perfect match isn't found we look for the biggest icon *smaller* than specified
+				if (id == UInt16.MaxValue) {
+					int requested = Math.Min (size.Height, size.Width);
+					IconDirEntry best = iconDir.idEntries [0];
+					for (ushort i=1; i < count; i++) {
+						IconDirEntry ide = iconDir.idEntries [i];
+						if ((ide.height < requested) || (ide.width < requested)) {
+							if ((ide.height > best.height) || (ide.width > best.width))
+								id = i;
+						}
+					}
+				}
+
+				// last one, if nothing better can be found
+				if (id == UInt16.MaxValue)
+					id = (ushort) (count - 1);
+
+				iconSize.Height = iconDir.idEntries [id].height;
+				iconSize.Width = iconDir.idEntries [id].width;
+			} else {
+				iconSize.Height = size.Height;
+				iconSize.Width = size.Width;
 			}
 
-			// last one, if nothing better can be found
-			if (id == UInt16.MaxValue)
-				id = (ushort) (count - 1);
-
-			iconSize.Height = iconDir.idEntries [id].height;
-			iconSize.Width = iconDir.idEntries [id].width;
-
 			if (original.bitmap != null)
 				bitmap = (Bitmap) original.bitmap.Clone ();
 		}
@@ -182,8 +189,10 @@ namespace System.Drawing
 		}
 
 		public Icon (string fileName)
-			: this (new FileStream (fileName, FileMode.Open), 32, 32)
-		{			
+		{
+			using (FileStream fs = File.OpenRead (fileName)) {
+				InitFromStreamWithSize (fs, 32, 32);
+			}
 		}
 
 		public Icon (Type type, string resource)
@@ -215,7 +224,7 @@ namespace System.Drawing
 					height = iconSize.Height;
 				}
 			}
-			if ((dataStream != null) && (width == height)) {
+			if (dataStream != null) {
 				dataStream.Seek (0, SeekOrigin.Begin);
 				InitFromStreamWithSize (dataStream, width, height);
 			}
@@ -233,12 +242,12 @@ namespace System.Drawing
 			this.undisposable = true;
 		}
 
-		void ISerializable.GetObjectData(SerializationInfo info, StreamingContext context)
+		void ISerializable.GetObjectData(SerializationInfo si, StreamingContext context)
 		{
 			MemoryStream ms = new MemoryStream ();
 			Save (ms);
-			info.AddValue ("IconSize", this.Size, typeof (Size));
-			info.AddValue ("IconData", ms.ToArray ());
+			si.AddValue ("IconSize", this.Size, typeof (Size));
+			si.AddValue ("IconData", ms.ToArray ());
 		}
 
 #if NET_2_0
@@ -247,20 +256,24 @@ namespace System.Drawing
 		{
 		}
 		
-		public Icon (string fileName, int width, int height) : 
-			this (new FileStream (fileName, FileMode.Open), width, height)
+		public Icon (string fileName, int width, int height)
 		{
+			using (FileStream fs = File.OpenRead (fileName)) {
+				InitFromStreamWithSize (fs, width, height);
+			}
 		}
 	
-		public Icon (string fileName, Size size) : 
-			this (new FileStream (fileName, FileMode.Open), size)
+		public Icon (string fileName, Size size)
 		{
+			using (FileStream fs = File.OpenRead (fileName)) {
+				InitFromStreamWithSize (fs, size.Width, size.Height);
+			}
 		}
 
 		[MonoLimitation ("The same icon, SystemIcons.WinLogo, is returned for all file types.")]
 		public static Icon ExtractAssociatedIcon (string filePath)
 		{
-			if ((filePath == null) || (filePath.Length == 0))
+			if (String.IsNullOrEmpty (filePath))
 				throw new ArgumentException (Locale.GetText ("Null or empty path."), "filePath");
 			if (!File.Exists (filePath))
 				throw new FileNotFoundException (Locale.GetText ("Couldn't find specified file."), filePath);
@@ -291,7 +304,7 @@ namespace System.Drawing
 
 		public object Clone ()
 		{
-			return new Icon (this, this.Width, this.Height);
+			return new Icon (this, Size);
 		}
 
 		[SecurityPermission (SecurityAction.LinkDemand, UnmanagedCode = true)]
@@ -637,12 +650,9 @@ namespace System.Drawing
 				throw new System.ArgumentException ("Invalid Argument", "stream");
 
 			ushort dirEntryCount = reader.ReadUInt16();
-			iconDir.idCount = dirEntryCount;
-			iconDir.idEntries = new IconDirEntry [dirEntryCount];
-			imageData = new IconImage [dirEntryCount];
+			ArrayList entries = new ArrayList (dirEntryCount);
 			bool sizeObtained = false;
 			// now read in the IconDirEntry structures
-			ushort p = 0;
 			for (int i = 0; i < dirEntryCount; i++) {
 				IconDirEntry ide;
 				ide.width = reader.ReadByte ();
@@ -664,26 +674,35 @@ Console.WriteLine ("\tide.bitCount: {0}", ide.bitCount);
 Console.WriteLine ("\tide.bytesInRes: {0}", ide.bytesInRes);
 Console.WriteLine ("\tide.imageOffset: {0}", ide.imageOffset);
 #endif
+
 				// 256x256 icons are decoded as 0x0 (width and height are encoded as BYTE)
 				// and we ignore them just like MS does (at least up to fx 2.0)
 				if ((ide.width == 0) && (ide.height == 0))
 					continue;
-				iconDir.idEntries [p++] = ide;
+
+				int index = entries.Add (ide);
 
 				//is this is the best fit??
 				if (!sizeObtained) {
 					if ((ide.height == height) || (ide.width == width)) {
-						this.id = (ushort) i;
+						this.id = (ushort) index;
 						sizeObtained = true;
 						this.iconSize.Height = ide.height;
 						this.iconSize.Width = ide.width;
 					}
 				}
 			}
+
 			// Vista 256x256 icons points directly to a PNG bitmap
-			dirEntryCount = p;
+			dirEntryCount = (ushort) entries.Count;
 			if (dirEntryCount == 0)
 				throw new Win32Exception (0, "No valid icon entry were found.");
+
+			iconDir.idCount = dirEntryCount;
+			imageData = new IconImage [dirEntryCount];
+			iconDir.idEntries = new IconDirEntry [dirEntryCount];
+			entries.CopyTo (iconDir.idEntries);
+
 			//if we havent found the best match, return the one with the
 			//largest size. Is this approach correct??
 			if (!sizeObtained){
