@@ -114,25 +114,25 @@ namespace System.Xml.Serialization
 				return;
 			}
 
-			if (ob.GetType ().IsArray && typeof (XmlNode).IsAssignableFrom (ob.GetType ().GetElementType ())) {
-				Writer.WriteStartElement (element, namesp);
-				foreach (XmlNode node in (IEnumerable) ob)
-					node.WriteTo (Writer);
-				Writer.WriteEndElement ();
-				return;
-			}
-
 			if (typeMap.TypeData.SchemaType == SchemaTypes.XmlSerializable)
 			{
 				WriteSerializable ((IXmlSerializable)ob, element, namesp, isNullable);
 				return;
 			}
 
-			XmlTypeMapping map = typeMap.GetRealTypeMap (ob.GetType().FullName);
+			XmlTypeMapping map = typeMap.GetRealTypeMap (ob.GetType());
 
 			if (map == null) 
 			{
-				WriteTypedPrimitive (element, namesp, ob, true);
+				// bug #81539
+				if (ob.GetType ().IsArray && typeof (XmlNode).IsAssignableFrom (ob.GetType ().GetElementType ())) {
+					Writer.WriteStartElement (element, namesp);
+					foreach (XmlNode node in (IEnumerable) ob)
+						node.WriteTo (Writer);
+					Writer.WriteEndElement ();
+				}
+				else
+					WriteTypedPrimitive (element, namesp, ob, true);
 				return;
 			}
 
@@ -279,6 +279,8 @@ namespace System.Xml.Serialization
 		bool MemberHasValue (XmlTypeMapMember member, object ob, bool isValueList)
 		{
 			if (isValueList) {
+				if (member.IsOptionalValueType && !member.GetValueSpecified (ob))
+					return false;
 				return member.GlobalIndex < ((object[])ob).Length;
 			}
 			else if (member.DefaultValue != System.DBNull.Value) {
@@ -342,6 +344,9 @@ namespace System.Xml.Serialization
 					break;
 
 				case SchemaTypes.XmlSerializable:
+					// bug #419973
+					if (!elem.MappedType.TypeData.Type.IsInstanceOfType (memberValue))
+						memberValue = ImplicitConvert (memberValue, elem.MappedType.TypeData.Type);
 					WriteSerializable ((IXmlSerializable) memberValue, elem.ElementName, elem.Namespace, elem.IsNullable);
 					break;
 
@@ -350,6 +355,24 @@ namespace System.Xml.Serialization
 			}
 		}
 
+		object ImplicitConvert (object obj, Type type)
+		{
+			if (obj == null)
+				return null;
+			for (Type t = type; t != typeof (object); t = t.BaseType) {
+				MethodInfo mi = t.GetMethod ("op_Implicit", new Type [] {t});
+				if (mi != null && mi.ReturnType.IsAssignableFrom (obj.GetType ()))
+					return mi.Invoke (null, new object [] {obj});
+			}
+
+			for (Type t = obj.GetType (); t != typeof (object); t = t.BaseType) {
+				MethodInfo mi = t.GetMethod ("op_Implicit", new Type [] {t});
+				if (mi != null && mi.ReturnType == type)
+					return mi.Invoke (null, new object [] {obj});
+			}
+			return obj;
+		}
+
 		void WritePrimitiveValueLiteral (object memberValue, string name, string ns, XmlTypeMapping mappedType, TypeData typeData, bool wrapped, bool isNullable)
 		{
 			if (!wrapped) {
@@ -452,8 +475,11 @@ namespace System.Xml.Serialization
 			}
 
 			Array elems = (Array) memberValue;
-			foreach (XmlNode elem in elems)
+			foreach (var elem_ in elems)
 			{
+				XmlNode elem = elem_ as XmlNode;
+				if (elem == null)
+					throw new InvalidOperationException (String.Format ("XmlAnyElementAttribute can only be applied to members of type XmlElement, XmlElement[] or XmlNode[]. The target object is {0}", elem_ != null ? elem_.GetType () : null));
 				if (elem is XmlElement) 
 				{
 					if (member.IsElementDefined (elem.Name, elem.NamespaceURI))
@@ -499,6 +525,8 @@ namespace System.Xml.Serialization
 
 		string GetEnumXmlValue (XmlTypeMapping typeMap, object ob)
 		{
+			if (ob == null)
+				return null;
 			EnumMap map = (EnumMap)typeMap.ObjectMap;
 			return map.GetXmlName (typeMap.TypeFullName, ob);
 		}
