@@ -129,9 +129,13 @@ namespace System.Xml
 		// argument is uri, not an xml fragment.
 		internal XmlTextReader (bool dummy, XmlResolver resolver, string url, XmlNodeType fragType, XmlParserContext context)
 		{
-			if (resolver == null)
+			if (resolver == null) {
+#if MOONLIGHT
+				resolver = new XmlXapResolver ();
+#else
 				resolver = new XmlUrlResolver ();
-
+#endif
+			}
 			this.XmlResolver = resolver;
 			string uriString;
 			Stream stream = GetStreamFromUrl (url, out uriString);
@@ -178,6 +182,12 @@ namespace System.Xml
 
 		private Stream GetStreamFromUrl (string url, out string absoluteUriString)
 		{
+#if NET_2_1
+			if (url == null)
+				throw new ArgumentNullException ("url");
+			if (url.Length == 0)
+				throw new ArgumentException ("url");
+#endif
 			Uri uri = resolver.ResolveUri (null, url);
 			absoluteUriString = uri != null ? uri.ToString () : String.Empty;
 			return resolver.GetEntity (uri, null, typeof (Stream)) as Stream;
@@ -253,11 +263,9 @@ namespace System.Xml
 			get { return readState == ReadState.EndOfFile; }
 		}
 
-#if !NET_2_1
 		public override bool HasValue {
 			get { return cursorToken.Value != null; }
 		}
-#endif
 
 		public override bool IsDefault {
 			// XmlTextReader does not expand default attributes.
@@ -416,6 +424,7 @@ namespace System.Xml
 
 		private int GetIndexOfQualifiedAttribute (string localName, string namespaceURI)
 		{
+			namespaceURI = namespaceURI ?? String.Empty;
 			for (int i = 0; i < attributeCount; i++) {
 				XmlAttributeTokenInfo ti = attributeTokens [i];
 				if (ti.LocalName == localName && ti.NamespaceURI == namespaceURI)
@@ -617,7 +626,6 @@ namespace System.Xml
 			return more;
 		}
 
-#if !NET_2_1
 		public override bool ReadAttributeValue ()
 		{
 			if (readState == ReadState.Initial && startNodeType == XmlNodeType.Attribute) {
@@ -638,7 +646,6 @@ namespace System.Xml
 			else
 				return false;
 		}
-#endif
 
 		public int ReadBase64 (byte [] buffer, int offset, int length)
 		{
@@ -952,7 +959,11 @@ namespace System.Xml
 		// These values are never re-initialized.
 		private bool namespaces = true;
 		private WhitespaceHandling whitespaceHandling = WhitespaceHandling.All;
+#if MOONLIGHT
+		private XmlResolver resolver = new XmlXapResolver ();
+#else
 		private XmlResolver resolver = new XmlUrlResolver ();
+#endif
 		private bool normalization = false;
 
 		private bool checkCharacters;
@@ -1050,13 +1061,21 @@ namespace System.Xml
 			nsmgr = nsmgr != null ? nsmgr : new XmlNamespaceManager (nameTable);
 
 			if (url != null && url.Length > 0) {
+#if NET_2_1
+				Uri uri = new Uri (url, UriKind.RelativeOrAbsolute);
+#else
 				Uri uri = null;
 				try {
+#if NET_2_0
+					uri = new Uri (url, UriKind.RelativeOrAbsolute);
+#else
 					uri = new Uri (url);
+#endif
 				} catch (Exception) {
 					string path = Path.GetFullPath ("./a");
 					uri = new Uri (new Uri (path), url);
 				}
+#endif
 				parserContext.BaseURI = uri.ToString ();
 			}
 
@@ -1455,6 +1474,10 @@ namespace System.Xml
 						Uri buri =
 							BaseURI != String.Empty ?
 							new Uri (BaseURI) : null;
+						// xml:base="" without any base URI -> pointless. However there are
+						// some people who use such xml:base. Seealso bug #608391.
+						if (buri == null && String.IsNullOrEmpty (value))
+							break;
 						Uri uri = resolver.ResolveUri (
 							buri, value);
 						parserContext.BaseURI =
@@ -1562,7 +1585,7 @@ namespace System.Xml
 
 		private void AppendValueChar (int ch)
 		{
-			if (ch < Char.MaxValue)
+			if (ch <= Char.MaxValue)
 				valueBuffer.Append ((char) ch);
 			else
 				AppendSurrogatePairValueChar (ch);
@@ -1644,7 +1667,7 @@ namespace System.Xml
 				// FIXME: it might be optimized by the JIT later,
 //				AppendValueChar (ch);
 				{
-					if (ch < Char.MaxValue)
+					if (ch <= Char.MaxValue)
 						valueBuffer.Append ((char) ch);
 					else
 						AppendSurrogatePairValueChar (ch);
@@ -1815,7 +1838,7 @@ namespace System.Xml
 		{
 			IncrementAttributeToken ();
 			XmlAttributeTokenInfo ati = attributeTokens [currentAttribute];
-			ati.Name = parserContext.NameTable.Add (name);
+			ati.Name = NameTable.Add (name);
 			ati.Prefix = String.Empty;
 			ati.NamespaceURI = String.Empty;
 			IncrementAttributeValueToken ();
@@ -1965,7 +1988,7 @@ namespace System.Xml
 					// FIXME: it might be optimized by the JIT later,
 //					AppendValueChar (ch);
 					{
-						if (ch < Char.MaxValue)
+						if (ch <= Char.MaxValue)
 							valueBuffer.Append ((char) ch);
 						else
 							AppendSurrogatePairValueChar (ch);
@@ -2358,7 +2381,7 @@ namespace System.Xml
 				// FIXME: it might be optimized by the JIT later,
 //				AppendValueChar (ch);
 				{
-					if (ch < Char.MaxValue)
+					if (ch <= Char.MaxValue)
 						valueBuffer.Append ((char) ch);
 					else
 						AppendSurrogatePairValueChar (ch);
@@ -2756,7 +2779,7 @@ namespace System.Xml
 			// AppendNameChar (ch);
 			{
 				// nameBuffer.Length is always non-0 so no need to ExpandNameCapacity () here
-				if (ch < Char.MaxValue)
+				if (ch <= Char.MaxValue)
 					nameBuffer [nameLength++] = (char) ch;
 				else
 					AppendSurrogatePairNameChar (ch);
@@ -2773,7 +2796,7 @@ namespace System.Xml
 				{
 					if (nameLength == nameCapacity)
 						ExpandNameCapacity ();
-					if (ch < Char.MaxValue)
+					if (ch <= Char.MaxValue)
 						nameBuffer [nameLength++] = (char) ch;
 					else
 						AppendSurrogatePairNameChar (ch);
@@ -2902,11 +2925,6 @@ namespace System.Xml
 				case -1:
 					throw NotWFError ("Unexpected end of xml.");
 				case '<':
-					if (i + 1 == length)
-						// if it does not end here,
-						// it cannot store another
-						// character, so stop here.
-						return i;
 					Advance (c);
 					if (PeekChar () != '/') {
 						nestLevel++;
@@ -2929,7 +2947,7 @@ namespace System.Xml
 					return i;
 				default:
 					Advance (c);
-					if (c < Char.MaxValue)
+					if (c <= Char.MaxValue)
 						buffer [bufIndex++] = (char) c;
 					else {
 						buffer [bufIndex++] = (char) ((c - 0x10000) / 0x400 + 0xD800);
