@@ -56,7 +56,46 @@ namespace Microsoft.VisualBasic
 {
 	internal class VBCodeGenerator : CodeGenerator
 	{
-		private string[] Keywords = new string[] {
+#if NET_2_0
+		private string [] Keywords = new string [] {
+			"AddHandler", "AddressOf", "Alias", "And",
+			"AndAlso", "Ansi", "As", "Assembly",
+			"Auto", "Boolean", "ByRef", "Byte", 
+			"ByVal", "Call", "Case", "Catch", 
+			"CBool", "CByte", "CChar", "CDate", 
+			"CDec", "CDbl", "Char", "CInt", 
+			"Class", "CLng", "CObj", "Const", 
+			"CShort", "CSng", "CStr", "CType", 
+			"Date", "Decimal", "Declare", "Default", 
+			"Delegate", "Dim", "DirectCast", "Do", 
+			"Double", "Each", "Else", "ElseIf", 
+			"End", "Enum", "Erase", "Error", 
+			"Event", "Exit", "False", "Finally", 
+			"For", "Friend", "Function", "Get", 
+			"GetType", "Global", "GoSub", "GoTo", "Handles", 
+			"If", "Implements", "Imports", "In", 
+			"Inherits", "Integer", "Interface", "Is", 
+			"Let", "Lib", "Like", "Long", 
+			"Loop", "Me", "Mod", "Module", 
+			"MustInherit", "MustOverride", "MyBase", "MyClass", 
+			"Namespace", "New", "Next", "Not", 
+			"Nothing", "NotInheritable", "NotOverridable", "Object", 
+			"On", "Option", "Optional", "Or", 
+			"OrElse", "Overloads", "Overridable", "Overrides", 
+			"ParamArray", "Partial", "Preserve", "Private", "Property", 
+			"Protected", "Public", "RaiseEvent", "ReadOnly", 
+			"ReDim", "REM", "RemoveHandler", "Resume", 
+			"Return", "Select", "Set", "Shadows", 
+			"Shared", "Short", "Single", "Static", 
+			"Step", "Stop", "String", "Structure", 
+			"Sub", "SyncLock", "Then", "Throw", 
+			"To", "True", "Try", "TypeOf", 
+			"Unicode", "Until", "Variant", "When", 
+			"While", "With", "WithEvents", "WriteOnly", 
+			"Xor" 
+		};
+#else
+		private string [] Keywords = new string [] {
 			"AddHandler", "AddressOf", "Alias", "And",
 			"AndAlso", "Ansi", "As", "Assembly",
 			"Auto", "Boolean", "ByRef", "Byte", 
@@ -93,6 +132,7 @@ namespace Microsoft.VisualBasic
 			"While", "With", "WithEvents", "WriteOnly", 
 			"Xor" 
 		};
+#endif
 
 		public VBCodeGenerator()
 		{
@@ -133,9 +173,9 @@ namespace Microsoft.VisualBasic
 
 				output.Write ("(Not (");
 				GenerateExpression (reverse ? e.Right : e.Left);
-    				output.Write (") Is ");
-    				GenerateExpression (reverse ? e.Left : e.Right);
-    				output.Write (')');
+				output.Write (") Is ");
+				GenerateExpression (reverse ? e.Left : e.Right);
+				output.Write (')');
 			} else {
 				base.GenerateBinaryOperatorExpression (e);
 			}
@@ -154,24 +194,21 @@ namespace Microsoft.VisualBasic
 
 				OutputType (createType);
 				
-				output.WriteLine (" {");
+				output.Write ("() {");
 				++Indent;
-				OutputExpressionList (initializers, true);
+				OutputExpressionList (initializers);
 				--Indent;
 				output.Write ("}");
-
-			} 
-			else {
+			} else {
 				CodeTypeReference arrayType = createType.ArrayElementType;
-				while (arrayType != null) 
-				{
+				while (arrayType != null) {
 					createType = arrayType;
 					arrayType = arrayType.ArrayElementType;
 				}
 
 				OutputType (createType);
 
-				output.Write ('(');
+				output.Write ("((");
 
 				CodeExpression size = expression.SizeExpression;
 				if (size != null)
@@ -179,7 +216,7 @@ namespace Microsoft.VisualBasic
 				else
 					output.Write (expression.Size);
 
-				output.Write (')');
+				output.Write (") - 1) {}");
 			}
 		}
 
@@ -199,14 +236,14 @@ namespace Microsoft.VisualBasic
 			output.Write (")");
 		}
 
-		private bool AsBool(object datavalue)
+		private bool AsBool (object datavalue)
 		{
-			return datavalue != null && datavalue is bool && (bool)datavalue;
+			return datavalue != null && datavalue is bool && (bool) datavalue;
 		}
 		
-		private string OnOff(bool datavalue)
+		private string OnOff (bool datavalue)
 		{
-			return datavalue?"On":"Off";
+			return datavalue ? "On" : "Off";
 		}
 
 		protected override void GenerateCompileUnitStart (CodeCompileUnit compileUnit)
@@ -221,8 +258,8 @@ namespace Microsoft.VisualBasic
 			GenerateComment (new CodeComment (" </autogenerated>"));
 			GenerateComment (new CodeComment ("------------------------------------------------------------------------------"));
 			Output.WriteLine ();
-			if (AsBool(compileUnit.UserData["AllowLateBound"])) {
-				Output.WriteLine("Option Explicit {0}",OnOff(AsBool(compileUnit.UserData["RequireVariableDeclaration"])));
+			if (AsBool (compileUnit.UserData ["AllowLateBound"])) {
+				Output.WriteLine("Option Explicit {0}", OnOff (AsBool (compileUnit.UserData ["RequireVariableDeclaration"])));
 				Output.WriteLine("Option Strict Off");
 			} else {
 				Output.WriteLine("Option Explicit On"); // Strict On implies Explicit On
@@ -264,17 +301,17 @@ namespace Microsoft.VisualBasic
 				GenerateExpression (targetObject);
 				Output.Write ('.');
 			}
-			Output.Write (expression.FieldName);
+			Output.Write (CreateEscapedIdentifier (expression.FieldName));
 		}
 		
 		protected override void GenerateArgumentReferenceExpression (CodeArgumentReferenceExpression expression)
 		{
-			Output.Write (expression.ParameterName);
+			Output.Write (CreateEscapedIdentifier (expression.ParameterName));
 		}
 
 		protected override void GenerateVariableReferenceExpression (CodeVariableReferenceExpression expression)
 		{
-			Output.Write (expression.VariableName);
+			Output.Write (CreateEscapedIdentifier (expression.VariableName));
 		}
 		
 		protected override void GenerateIndexerExpression (CodeIndexerExpression expression)
@@ -292,7 +329,7 @@ namespace Microsoft.VisualBasic
 			TextWriter output = Output;
 
 			GenerateExpression (expression.TargetObject);
-			output.Write (".Item(");
+			output.Write ("(");
 			OutputExpressionList (expression.Indices);
 			output.Write (')');
 		}
@@ -315,9 +352,11 @@ namespace Microsoft.VisualBasic
 
 		protected override void GenerateMethodReferenceExpression (CodeMethodReferenceExpression expression)
 		{
-			GenerateExpression (expression.TargetObject);
-			Output.Write ('.');
-			Output.Write (expression.MethodName);
+			if (expression.TargetObject != null) {
+				GenerateExpression (expression.TargetObject);
+				Output.Write ('.');
+			}
+			Output.Write (CreateEscapedIdentifier (expression.MethodName));
 		}
 
 		protected override void GenerateEventReferenceExpression (CodeEventReferenceExpression expression)
@@ -325,17 +364,39 @@ namespace Microsoft.VisualBasic
 			if (expression.TargetObject != null) {
 				GenerateExpression (expression.TargetObject);
 				Output.Write ('.');
+				if (expression.TargetObject is CodeThisReferenceExpression) {
+					// We're actually creating a reference to a compiler-generated field here...
+					Output.Write (expression.EventName + "Event");
+				} else {
+					Output.Write (CreateEscapedIdentifier (expression.EventName));
+				}
+			} else {
+				Output.Write (CreateEscapedIdentifier (expression.EventName + "Event"));
 			}
-			Output.Write (CreateEscapedIdentifier(expression.EventName));
 		}
 
 		protected override void GenerateDelegateInvokeExpression (CodeDelegateInvokeExpression expression)
 		{
+			CodeEventReferenceExpression ev = expression.TargetObject as CodeEventReferenceExpression;
+			
+#if ONLY_1_1
 			Output.Write ("RaiseEvent ");
-			GenerateExpression (expression.TargetObject);
+#endif
+			if (ev != null) {
+#if NET_2_0
+				Output.Write ("RaiseEvent ");
+#endif
+				if (ev.TargetObject != null && !(ev.TargetObject is CodeThisReferenceExpression)) {
+					GenerateExpression (ev.TargetObject);
+					Output.Write (".");
+				}
+				Output.Write (ev.EventName);
+			} else if (expression.TargetObject != null) {
+				GenerateExpression (expression.TargetObject);
+			}
 			Output.Write ('(');
 			OutputExpressionList (expression.Parameters);
-			Output.WriteLine (')');
+			Output.Write (')');
 		}
 		
 		protected override void GenerateObjectCreateExpression (CodeObjectCreateExpression expression)
@@ -394,7 +455,11 @@ namespace Microsoft.VisualBasic
 
 		protected override void GeneratePropertyReferenceExpression (CodePropertyReferenceExpression expression)
 		{
-			GenerateMemberReferenceExpression (expression.TargetObject, expression.PropertyName);
+			if (expression.TargetObject != null) {
+				GenerateMemberReferenceExpression (expression.TargetObject, expression.PropertyName);
+			} else {
+				Output.Write (CreateEscapedIdentifier (expression.PropertyName));
+			}
 		}
 
 		protected override void GeneratePropertySetValueReferenceExpression (CodePropertySetValueReferenceExpression expression)
@@ -453,13 +518,13 @@ namespace Microsoft.VisualBasic
 			string text = comment.Text;
 
 			for (int i = 0; i < text.Length; i++) {
-				output.Write (text[i]);
+				output.Write (text [i]);
 				if (text[i] == '\r') {
-					if (i < (text.Length - 1) && text[i + 1] == '\n') {
+					if (i < (text.Length - 1) && text [i + 1] == '\n') {
 						continue;
 					}
 					output.Write (commentChars);
-				} else if (text[i] == '\n') {
+				} else if (text [i] == '\n') {
 					output.Write (commentChars);
 				}
 			}
@@ -526,7 +591,6 @@ namespace Microsoft.VisualBasic
 
 			CodeStatementCollection finallies = statement.FinallyStatements;
 			if (finallies.Count > 0) {
-
 				output.WriteLine ("Finally");
 				++Indent;
 				GenerateStatements (finallies);
@@ -537,7 +601,7 @@ namespace Microsoft.VisualBasic
 		}
 
 		protected override void GenerateAssignStatement (CodeAssignStatement statement)
-		{			
+		{
 			TextWriter output = Output;
 			GenerateExpression (statement.Left);
 			output.Write (" = ");
@@ -550,7 +614,11 @@ namespace Microsoft.VisualBasic
 			TextWriter output = Output;
 
 			Output.Write ("AddHandler ");
-			GenerateEventReferenceExpression (statement.Event);
+			if (statement.Event.TargetObject != null) {
+				GenerateEventReferenceExpression (statement.Event);
+			} else {
+				Output.Write (CreateEscapedIdentifier (statement.Event.EventName));
+			}
 			Output.Write ( ", ");
 			GenerateExpression (statement.Listener);
 			output.WriteLine ();
@@ -561,7 +629,11 @@ namespace Microsoft.VisualBasic
 			TextWriter output = Output;
 
 			Output.Write ("RemoveHandler ");
-			GenerateEventReferenceExpression (statement.Event);
+			if (statement.Event.TargetObject != null) {
+				GenerateEventReferenceExpression (statement.Event);
+			} else {
+				Output.Write (CreateEscapedIdentifier (statement.Event.EventName));
+			}
 			Output.Write ( ", ");
 			GenerateExpression (statement.Listener);
 			output.WriteLine ();
@@ -605,8 +677,7 @@ namespace Microsoft.VisualBasic
 			OutputTypeNamePair (statement.Type, statement.Name);
 
 			CodeExpression initExpression = statement.InitExpression;
-			if (initExpression != null) 
-			{
+			if (initExpression != null) {
 				output.Write (" = ");
 				GenerateExpression (initExpression);
 			}
@@ -617,11 +688,12 @@ namespace Microsoft.VisualBasic
 		protected override void GenerateLinePragmaStart (CodeLinePragma linePragma)
 		{
 			Output.WriteLine ();
-			Output.Write ("#ExternalSource(");
+			Output.Write ("#ExternalSource(\"");
 			Output.Write (linePragma.FileName);
-			Output.Write (", ");
+			Output.Write ("\",");
 			Output.Write (linePragma.LineNumber);
 			Output.WriteLine (")");
+			Output.WriteLine ("");
 		}
 
 		protected override void GenerateLinePragmaEnd (CodeLinePragma linePragma)
@@ -631,13 +703,12 @@ namespace Microsoft.VisualBasic
 
 		protected override void GenerateEvent (CodeMemberEvent eventRef, CodeTypeDeclaration declaration)
 		{
-			if (IsCurrentDelegate || IsCurrentEnum) {
+			if (IsCurrentDelegate || IsCurrentEnum)
 				return;
-			}
 
 			TextWriter output = Output;
 
-			OutputAttributes (eventRef.CustomAttributes, null, 
+			OutputAttributes (eventRef.CustomAttributes, null,
 				LineHandling.ContinueLine);
 
 			OutputMemberAccessModifier (eventRef.Attributes);
@@ -661,9 +732,8 @@ namespace Microsoft.VisualBasic
 
 		protected override void GenerateField (CodeMemberField field)
 		{
-			if (IsCurrentDelegate || IsCurrentInterface) {
+			if (IsCurrentDelegate || IsCurrentInterface)
 				return;
-			}
 
 			TextWriter output = Output;
 
@@ -697,7 +767,7 @@ namespace Microsoft.VisualBasic
 		protected override void GenerateEntryPointMethod (CodeEntryPointMethod method, CodeTypeDeclaration declaration)
 		{
 #if NET_2_0
-			OutputAttributes (method.CustomAttributes, null, 
+			OutputAttributes (method.CustomAttributes, null,
 				LineHandling.ContinueLine);
 #endif
 
@@ -711,9 +781,8 @@ namespace Microsoft.VisualBasic
 		[MonoTODO ("partially implemented")]
 		protected override void GenerateMethod (CodeMemberMethod method, CodeTypeDeclaration declaration)
 		{
-			if (IsCurrentDelegate || IsCurrentEnum) {
+			if (IsCurrentDelegate || IsCurrentEnum)
 				return;
-			}
 
 			bool isSub = method.ReturnType.BaseType == typeof(void).FullName;
 
@@ -743,6 +812,11 @@ namespace Microsoft.VisualBasic
 				output.Write ("Function ");
 
 			output.Write (GetMethodName(method));
+
+#if NET_2_0
+			OutputTypeParameters (method.TypeParameters);
+#endif
+
 			output.Write ('(');
 			OutputParameters (method.Parameters);
 			output.Write (')');
@@ -779,9 +853,8 @@ namespace Microsoft.VisualBasic
 
 		protected override void GenerateProperty (CodeMemberProperty property, CodeTypeDeclaration declaration)
 		{
-			if (IsCurrentDelegate || IsCurrentEnum) {
+			if (IsCurrentDelegate || IsCurrentEnum)
 				return;
-			}
 
 			TextWriter output = Output;
 
@@ -847,18 +920,22 @@ namespace Microsoft.VisualBasic
 				++Indent;
 				if (property.HasGet) {
 					output.WriteLine ("Get");
-					++Indent;
-					GenerateStatements (property.GetStatements);
-					--Indent;
-					output.WriteLine ("End Get");
+					if (!IsAbstract (property.Attributes)) {
+						++Indent;
+						GenerateStatements (property.GetStatements);
+						--Indent;
+						output.WriteLine ("End Get");
+					}
 				}
 
 				if (property.HasSet) {
 					output.WriteLine ("Set");
-					++Indent;
-					GenerateStatements (property.SetStatements);
-					--Indent;
-					output.WriteLine ("End Set");
+					if (!IsAbstract (property.Attributes)) {
+						++Indent;
+						GenerateStatements (property.SetStatements);
+						--Indent;
+						output.WriteLine ("End Set");
+					}
 				}
 
 				--Indent;
@@ -868,9 +945,8 @@ namespace Microsoft.VisualBasic
 
 		protected override void GenerateConstructor (CodeConstructor constructor, CodeTypeDeclaration declaration)
 		{
-			if (IsCurrentDelegate || IsCurrentEnum || IsCurrentInterface) {
+			if (IsCurrentDelegate || IsCurrentEnum || IsCurrentInterface)
 				return;
-			}
 
 			OutputAttributes (constructor.CustomAttributes, null,
 				LineHandling.ContinueLine);
@@ -908,9 +984,8 @@ namespace Microsoft.VisualBasic
 		
 		protected override void GenerateTypeConstructor (CodeTypeConstructor constructor)
 		{
-			if (IsCurrentDelegate || IsCurrentEnum || IsCurrentInterface) {
+			if (IsCurrentDelegate || IsCurrentEnum || IsCurrentInterface)
 				return;
-			}
 
 #if NET_2_0
 			OutputAttributes (constructor.CustomAttributes, null,
@@ -948,8 +1023,12 @@ namespace Microsoft.VisualBasic
 					output.Write ("Delegate Function ");
 				}
 
-				output.Write (delegateDecl.Name);
+				output.Write (CreateEscapedIdentifier (delegateDecl.Name));
+#if NET_2_0
+				OutputTypeParameters (delegateDecl.TypeParameters);
+#endif
 				output.Write ("(");
+				OutputParameters (delegateDecl.Parameters);
 				Output.Write (")");
 				if (!isSub) {
 					Output.Write (" As ");
@@ -958,8 +1037,10 @@ namespace Microsoft.VisualBasic
 				Output.WriteLine ("");
 			} else {
 				OutputTypeAttributes (declaration);
-
-				output.Write (declaration.Name);
+				output.Write (CreateEscapedIdentifier (declaration.Name));
+#if NET_2_0
+				OutputTypeParameters (declaration.TypeParameters);
+#endif
 
 				if (IsCurrentEnum) {
 					if (declaration.BaseTypes.Count > 0) {
@@ -1154,12 +1235,12 @@ namespace Microsoft.VisualBasic
 		protected override void OutputFieldScopeModifier (MemberAttributes attributes)
 		{
 			switch (attributes & MemberAttributes.ScopeMask) {
-				case MemberAttributes.Static:
-					Output.Write ("Shared ");
-					break;
-				case MemberAttributes.Const:
-					Output.Write ("Const ");
-					break;
+			case MemberAttributes.Static:
+				Output.Write ("Shared ");
+				break;
+			case MemberAttributes.Const:
+				Output.Write ("Const ");
+				break;
 			}
 		}
 
@@ -1187,79 +1268,77 @@ namespace Microsoft.VisualBasic
 		protected override void OutputMemberAccessModifier (MemberAttributes attributes)
 		{
 			switch (attributes & MemberAttributes.AccessMask) {
-				case MemberAttributes.Assembly:
-				case MemberAttributes.FamilyAndAssembly:
-					Output.Write ("Friend "); 
-					break;
-				case MemberAttributes.Family:
-					Output.Write ("Protected ");
-					break;
-				case MemberAttributes.FamilyOrAssembly:
-					Output.Write ("Protected Friend ");
-					break;
-				case MemberAttributes.Private:
-					Output.Write ("Private ");
-					break;
-				case MemberAttributes.Public:
-					Output.Write ("Public ");
-					break;
+			case MemberAttributes.Assembly:
+			case MemberAttributes.FamilyAndAssembly:
+				Output.Write ("Friend "); 
+				break;
+			case MemberAttributes.Family:
+				Output.Write ("Protected ");
+				break;
+			case MemberAttributes.FamilyOrAssembly:
+				Output.Write ("Protected Friend ");
+				break;
+			case MemberAttributes.Private:
+				Output.Write ("Private ");
+				break;
+			case MemberAttributes.Public:
+				Output.Write ("Public ");
+				break;
 			}
 		}
 
 		private void OutputVTableModifier (MemberAttributes attributes)
 		{
-			if ((attributes & MemberAttributes.VTableMask) == MemberAttributes.New) {
+			if ((attributes & MemberAttributes.VTableMask) == MemberAttributes.New)
 				Output.Write ("Shadows ");
-			}
 		}
 
 		protected override void OutputMemberScopeModifier (MemberAttributes attributes)
 		{
 			switch (attributes & MemberAttributes.ScopeMask) {
-				case MemberAttributes.Abstract:
-					Output.Write ("MustOverride ");
-					break;
-				case MemberAttributes.Final:
-					// do nothing
-					break;
-				case MemberAttributes.Static:
-					Output.Write ("Shared ");
-					break;
-				case MemberAttributes.Override:
-					Output.Write ("Overrides ");
-					break;
-				case MemberAttributes.Overloaded:
-					// based on http://gendotnet.com/Code%20Gen%20Articles/codedom.htm
-					Output.Write ("Overloads ");
+			case MemberAttributes.Abstract:
+				Output.Write ("MustOverride ");
+				break;
+			case MemberAttributes.Final:
+				// do nothing
+				break;
+			case MemberAttributes.Static:
+				Output.Write ("Shared ");
+				break;
+			case MemberAttributes.Override:
+				Output.Write ("Overrides ");
+				break;
+			case MemberAttributes.Overloaded:
+				// based on http://gendotnet.com/Code%20Gen%20Articles/codedom.htm
+				Output.Write ("Overloads ");
 
-					MemberAttributes access_ovl = attributes & MemberAttributes.AccessMask;
-					if (access_ovl == MemberAttributes.Public || access_ovl == MemberAttributes.Family) {
-						Output.Write ("Overridable ");
-					}
-					break;
-				default:
-					//
-					// FUNNY! if the scope value is
-					// rubbish (0 or >Const), and access
-					// is public, protected make it
-					// "virtual".
-					//
-					// i'm not sure whether this is 100%
-					// correct, but it seems to be MS
-					// behavior.
-					//
-					// On MS.NET 2.0, internal properties
-					// are also marked "virtual".
-					//
-					MemberAttributes access = attributes & MemberAttributes.AccessMask;
-					if (access == MemberAttributes.Public || 
+				MemberAttributes access_ovl = attributes & MemberAttributes.AccessMask;
+				if (access_ovl == MemberAttributes.Public || access_ovl == MemberAttributes.Family)
+					Output.Write ("Overridable ");
+				break;
+			default:
+				//
+				// FUNNY! if the scope value is
+				// rubbish (0 or >Const), and access
+				// is public, protected make it
+				// "virtual".
+				//
+				// i'm not sure whether this is 100%
+				// correct, but it seems to be MS
+				// behavior.
+				//
+				// On MS.NET 2.0, internal properties
+				// are also marked "virtual".
+				//
+				MemberAttributes access = attributes & MemberAttributes.AccessMask;
+				if (access == MemberAttributes.Public || 
 #if NET_2_0
-						access == MemberAttributes.Family || access == MemberAttributes.Assembly)
+					access == MemberAttributes.Family || access == MemberAttributes.Assembly)
 #else
-						access == MemberAttributes.Family)
+					access == MemberAttributes.Family)
 #endif
-						Output.Write ("Overridable ");
-					break;
+					Output.Write ("Overridable ");
+				break;
 			}
 		}
 
@@ -1325,26 +1404,31 @@ namespace Microsoft.VisualBasic
 			TextWriter output = Output;
 			TypeAttributes attributes = declaration.TypeAttributes;
 
-			switch (attributes & TypeAttributes.VisibilityMask) {
-				case TypeAttributes.Public:
-				case TypeAttributes.NestedPublic:
-					output.Write ("Public ");
-					break;
-				case TypeAttributes.NestedPrivate:
-					output.Write ("Private ");
-					break;
 #if NET_2_0
-				case TypeAttributes.NotPublic:
-				case TypeAttributes.NestedFamANDAssem:
-				case TypeAttributes.NestedAssembly:
-					output.Write ("Friend ");
-					break; 
-				case TypeAttributes.NestedFamily:
-					output.Write ("Protected ");
-					break;
-				case TypeAttributes.NestedFamORAssem:
-					output.Write ("Protected Friend ");
-					break;
+			if (declaration.IsPartial)
+				output.Write ("Partial ");
+#endif
+			
+			switch (attributes & TypeAttributes.VisibilityMask) {
+			case TypeAttributes.Public:
+			case TypeAttributes.NestedPublic:
+				output.Write ("Public ");
+				break;
+			case TypeAttributes.NestedPrivate:
+				output.Write ("Private ");
+				break;
+#if NET_2_0
+			case TypeAttributes.NotPublic:
+			case TypeAttributes.NestedFamANDAssem:
+			case TypeAttributes.NestedAssembly:
+				output.Write ("Friend ");
+				break; 
+			case TypeAttributes.NestedFamily:
+				output.Write ("Protected ");
+				break;
+			case TypeAttributes.NestedFamORAssem:
+				output.Write ("Protected Friend ");
+				break;
 #endif
 			}
 
@@ -1367,12 +1451,59 @@ namespace Microsoft.VisualBasic
 			}
 		}
 
+#if NET_2_0
+		void OutputTypeParameters (CodeTypeParameterCollection parameters)
+		{
+			int count = parameters.Count;
+			if (count == 0)
+				return;
+
+			Output.Write ("(Of ");
+			for (int i = 0; i < count; ++i) {
+				if (i > 0)
+					Output.Write (", ");
+				CodeTypeParameter p = parameters [i];
+				Output.Write (p.Name);
+				OutputTypeParameterConstraints (p);
+			}
+			Output.Write (')');
+		}
+
+		void OutputTypeParameterConstraints (CodeTypeParameter parameter)
+		{
+			int constraint_count = parameter.Constraints.Count +
+				(parameter.HasConstructorConstraint ? 1 : 0);
+
+			if (constraint_count == 0)
+				return;
+
+			Output.Write (" As ");
+
+			if (constraint_count > 1)
+				Output.Write (" {");
+
+			for (int i = 0; i < parameter.Constraints.Count; i++) {
+				if (i > 0)
+					Output.Write (", ");
+				OutputType (parameter.Constraints [i]);
+			}
+
+			if (parameter.HasConstructorConstraint) {
+				if (constraint_count > 1)
+					Output.Write (", ");
+				Output.Write ("New");
+			}
+
+			if (constraint_count > 1)
+				Output.Write ("}");
+		}
+#endif
+
 		protected override void OutputTypeNamePair (CodeTypeReference typeRef, String name)
 		{
 #if NET_2_0
-			if (name.Length == 0) {
+			if (name.Length == 0)
 				name = "__exception";
-			}
 #endif
 			Output.Write (CreateEscapedIdentifier(name) + " As " + GetTypeOutput (typeRef));
 		}
@@ -1387,12 +1518,10 @@ namespace Microsoft.VisualBasic
 			StringBuilder mySBuilder = new StringBuilder(value.Length);
 			mySBuilder.Append ("\"");
 			bool inQuotes = true;
-			for (int MyCounter = 0; MyCounter < value.Length; MyCounter++)
-			{
+			for (int MyCounter = 0; MyCounter < value.Length; MyCounter++) {
 				if (value[MyCounter] == 34) //quotation mark
 				{
-					if (!inQuotes)
-					{
+					if (!inQuotes) {
 						mySBuilder.Append ("&\"");
 						inQuotes = true;
 					}
@@ -1401,8 +1530,7 @@ namespace Microsoft.VisualBasic
 				}
 				else if (value[MyCounter] >= 32) //standard ansi/unicode characters
 				{
-					if (!inQuotes)
-					{
+					if (!inQuotes) {
 						mySBuilder.Append ("&\"");
 						inQuotes = true;
 					}
@@ -1410,15 +1538,14 @@ namespace Microsoft.VisualBasic
 				}
 				else //special chars, e.g. line break
 				{
-					if (inQuotes)
-					{ 
+					if (inQuotes) {
 						mySBuilder.Append ("\"");
 						inQuotes = false;
 					}
 					mySBuilder.Append ("&Microsoft.VisualBasic.ChrW(");
 					mySBuilder.Append ((int)value[MyCounter]); 
 					mySBuilder.Append (")");
-				}			
+				}
 			}
 			if (inQuotes)
 				mySBuilder.Append ("\"");
@@ -1460,7 +1587,7 @@ namespace Microsoft.VisualBasic
 			arrayType = type.ArrayElementType;
 			if (arrayType != null)
 				output = GetTypeOutput (arrayType);
-			else { 
+			else {
 				switch (type.BaseType) {
 				case "System.DateTime":
 					output = "Date";
@@ -1514,6 +1641,7 @@ namespace Microsoft.VisualBasic
 #endif
 				default:
 					output = type.BaseType.Replace('+', '.');
+					output = CreateEscapedIdentifier (output);
 					break;
 				}
 			}
@@ -1555,9 +1683,8 @@ namespace Microsoft.VisualBasic
 					continue;
 				}
 
-				if (p != property && p.Name == property.Name && p.PrivateImplementationType == null) {
+				if (p != property && p.Name == property.Name && p.PrivateImplementationType == null)
 					return true;
-				}
 			}
 			return false;
 		}
@@ -1575,9 +1702,8 @@ namespace Microsoft.VisualBasic
 					continue;
 				}
 
-				if (!(m is CodeTypeConstructor) && !(m is CodeConstructor) && m != method && m.Name == method.Name && m.PrivateImplementationType == null) {
+				if (!(m is CodeTypeConstructor) && !(m is CodeConstructor) && m != method && m.Name == method.Name && m.PrivateImplementationType == null)
 					return true;
-				}
 			}
 			return false;
 		}
@@ -1585,9 +1711,8 @@ namespace Microsoft.VisualBasic
 		private string GetEventName (CodeMemberEvent evt)
 		{
 #if NET_2_0
-			if (evt.PrivateImplementationType == null) {
+			if (evt.PrivateImplementationType == null)
 				return evt.Name;
-			}
 
 			string baseType = evt.PrivateImplementationType.BaseType.Replace ('.', '_');
 			return baseType + "_" + evt.Name;
@@ -1598,9 +1723,8 @@ namespace Microsoft.VisualBasic
 
 		private string GetMethodName (CodeMemberMethod method)
 		{
-			if (method.PrivateImplementationType == null) {
+			if (method.PrivateImplementationType == null)
 				return method.Name;
-			}
 
 			string baseType = method.PrivateImplementationType.BaseType.Replace ('.', '_');
 			return baseType + "_" + method.Name;
@@ -1608,14 +1732,18 @@ namespace Microsoft.VisualBasic
 
 		private string GetPropertyName (CodeMemberProperty property)
 		{
-			if (property.PrivateImplementationType == null) {
+			if (property.PrivateImplementationType == null)
 				return property.Name;
-			}
 
 			string baseType = property.PrivateImplementationType.BaseType.Replace ('.', '_');
 			return baseType + "_" + property.Name;
 		}
 
+		static bool IsAbstract (MemberAttributes attributes)
+		{
+			return (attributes & MemberAttributes.ScopeMask) == MemberAttributes.Abstract;
+		}
+
 		private enum LineHandling
 		{
 			InLine,
