@@ -39,6 +39,11 @@ using System.Runtime.InteropServices;
 
 namespace System.ComponentModel
 {
+#if MOONLIGHT
+	public abstract class PropertyDescriptor
+	{
+	}
+#else
 	[ComVisible (true)]
 	public abstract class PropertyDescriptor : MemberDescriptor
 	{
@@ -313,6 +318,5 @@ namespace System.ComponentModel
 			return type;
 		}
 	}
+#endif
 }
-
-
