@@ -42,9 +42,6 @@ using System.Globalization;
 
 namespace System.Net.NetworkInformation {
 	public abstract class NetworkInterface {
-		[DllImport ("libc")]
-		static extern int uname (IntPtr buf);
-
 		static Version windowsVer51 = new Version (5, 1);
 		static internal readonly bool runningOnUnix = (Environment.OSVersion.Platform == PlatformID.Unix);
 		
@@ -56,17 +53,8 @@ namespace System.Net.NetworkInformation {
 		public static NetworkInterface [] GetAllNetworkInterfaces ()
 		{
 			if (runningOnUnix) {
-				bool darwin = false;
-				IntPtr buf = Marshal.AllocHGlobal (8192);
-				if (uname (buf) == 0) {
-					string os = Marshal.PtrToStringAnsi (buf);
-					if (os == "Darwin")
-						darwin = true;
-				}
-				Marshal.FreeHGlobal (buf);
-
 				try {
-					if (darwin)
+					if (Platform.IsMacOS)
 						return MacOsNetworkInterface.ImplGetAllNetworkInterfaces ();
 					else
 						return LinuxNetworkInterface.ImplGetAllNetworkInterfaces ();
