@@ -31,7 +31,11 @@
 //
 
 #if NET_2_0 && SECURITY_DEP
+
+#if !MOONLIGHT
 extern alias PrebuiltSystem;
+using X509CertificateCollection = PrebuiltSystem::System.Security.Cryptography.X509Certificates.X509CertificateCollection;
+#endif
 
 using System;
 using System.IO;
@@ -51,8 +55,6 @@ using MonoHashAlgorithmType = Mono.Security.Protocol.Tls.HashAlgorithmType;
 using MonoExchangeAlgorithmType = Mono.Security.Protocol.Tls.ExchangeAlgorithmType;
 using MonoSecurityProtocolType = Mono.Security.Protocol.Tls.SecurityProtocolType;
 
-using X509CertificateCollection = PrebuiltSystem::System.Security.Cryptography.X509Certificates.X509CertificateCollection;
-
 namespace System.Net.Security 
 {
 	[MonoTODO ("Non-X509Certificate2 certificate is not supported")]
@@ -138,7 +140,11 @@ namespace System.Net.Security
 		}
 
 		public override bool IsServer { 
+#if MOONLIGHT
+			get { return false; }
+#else
 			get { return ssl_stream is SslServerStream; }
+#endif
 		}
 
 		public override bool IsSigned { 
@@ -272,8 +278,11 @@ namespace System.Net.Security
 		public virtual X509Certificate RemoteCertificate {
 			get {
 				CheckConnectionAuthenticated ();
-
+#if MOONLIGHT
+				return ssl_stream.ServerCertificate;
+#else
 				return !IsServer ? ssl_stream.ServerCertificate : ((SslServerStream) ssl_stream).ClientCertificate;
+#endif
 			}
 		}
 
@@ -400,7 +409,7 @@ namespace System.Net.Security
 
 			return ssl_stream.BeginRead (buffer, offset, count, asyncCallback, asyncState);
 		}
-
+#if !MOONLIGHT
 		public virtual IAsyncResult BeginAuthenticateAsServer (X509Certificate serverCertificate, AsyncCallback callback, object asyncState)
 		{
 			return BeginAuthenticateAsServer (serverCertificate, false, SslProtocols.Tls, false, callback, asyncState);
@@ -440,7 +449,7 @@ namespace System.Net.Security
 
 			return BeginRead (new byte [0], 0, 0, callback, asyncState);
 		}
-
+#endif
 		MonoSecurityProtocolType GetMonoSslProtocol (SslProtocols ms)
 		{
 			switch (ms) {
@@ -472,7 +481,7 @@ namespace System.Net.Security
 			EndAuthenticateAsClient (BeginAuthenticateAsClient (
 				targetHost, clientCertificates, sslProtocolType, checkCertificateRevocation, null, null));
 		}
-
+#if !MOONLIGHT
 		public virtual void AuthenticateAsServer (X509Certificate serverCertificate)
 		{
 			AuthenticateAsServer (serverCertificate, false, SslProtocols.Tls, false);
@@ -483,7 +492,7 @@ namespace System.Net.Security
 			EndAuthenticateAsServer (BeginAuthenticateAsServer (
 				serverCertificate, clientCertificateRequired, sslProtocolType, checkCertificateRevocation, null, null));
 		}
-
+#endif
 		protected override void Dispose (bool disposing)
 		{
 			if (disposing) {
@@ -503,7 +512,7 @@ namespace System.Net.Security
 			else
 				ssl_stream.EndWrite (asyncResult);
 		}
-
+#if !MOONLIGHT
 		public virtual void EndAuthenticateAsServer (IAsyncResult asyncResult)
 		{
 			CheckConnectionAuthenticated ();
@@ -513,7 +522,7 @@ namespace System.Net.Security
 			else
 				ssl_stream.EndWrite (asyncResult);
 		}
-
+#endif
 		public override int EndRead (IAsyncResult asyncResult)
 		{
 			CheckConnectionAuthenticated ();
