@@ -26,12 +26,13 @@
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
 
-#if NET_2_0 && SECURITY_DEP
+#if SECURITY_DEP
 
 using System.IO;
 using System.Net.Sockets;
 using System.Reflection;
 using System.Text;
+using System.Threading;
 using System.Security.Cryptography;
 using System.Security.Cryptography.X509Certificates;
 using Mono.Security.Protocol.Tls;
@@ -39,6 +40,7 @@ using Mono.Security.Protocol.Tls;
 namespace System.Net {
 	sealed class HttpConnection
 	{
+		static AsyncCallback onread_cb = new AsyncCallback (OnRead);
 		const int BufferSize = 8192;
 		Socket sock;
 		Stream stream;
@@ -51,10 +53,14 @@ namespace System.Net {
 		RequestStream i_stream;
 		ResponseStream o_stream;
 		bool chunked;
-		int chunked_uses;
+		int reuses;
 		bool context_bound;
 		bool secure;
 		AsymmetricAlgorithm key;
+		int s_timeout = 90000; // 90k ms for first request, 15k ms from then on
+		Timer timer;
+		IPEndPoint local_ep;
+		HttpListener last_listener;
 
 		public HttpConnection (Socket sock, EndPointListener epl, bool secure, X509Certificate2 cert, AsymmetricAlgorithm key)
 		{
@@ -65,14 +71,11 @@ namespace System.Net {
 			if (secure == false) {
 				stream = new NetworkStream (sock, false);
 			} else {
-#if EMBEDDED_IN_1_0
-				throw new NotImplementedException ();
-#else
 				SslServerStream ssl_stream = new SslServerStream (new NetworkStream (sock, false), cert, false, false);
 				ssl_stream.PrivateKeyCertSelectionDelegate += OnPVKSelection;
 				stream = ssl_stream;
-#endif
 			}
+			timer = new Timer (OnTimeout, null, Timeout.Infinite, Timeout.Infinite);
 			Init ();
 		}
 
@@ -81,7 +84,6 @@ namespace System.Net {
 			return key;
 		}
 
-
 		void Init ()
 		{
 			context_bound = false;
@@ -96,8 +98,12 @@ namespace System.Net {
 			context = new HttpListenerContext (this);
 		}
 
-		public int ChunkedUses {
-			get { return chunked_uses; }
+		public bool IsClosed {
+			get { return (sock == null); }
+		}
+
+		public int Reuses {
+			get { return reuses; }
 		}
 
 		public IPEndPoint LocalEndPoint {
@@ -117,14 +123,25 @@ namespace System.Net {
 			set { prefix = value; }
 		}
 
+		void OnTimeout (object unused)
+		{
+			CloseSocket ();
+			Unbind ();
+		}
+
 		public void BeginReadRequest ()
 		{
 			if (buffer == null)
 				buffer = new byte [BufferSize];
 			try {
-				stream.BeginRead (buffer, 0, BufferSize, OnRead, this);
+				if (reuses == 1)
+					s_timeout = 15000;
+				timer.Change (s_timeout, Timeout.Infinite);
+				stream.BeginRead (buffer, 0, BufferSize, onread_cb, this);
 			} catch {
-				sock.Close (); // stream disposed
+				timer.Change (Timeout.Infinite, Timeout.Infinite);
+				CloseSocket ();
+				Unbind ();
 			}
 		}
 
@@ -156,26 +173,39 @@ namespace System.Net {
 			return o_stream;
 		}
 
-		void OnRead (IAsyncResult ares)
+		static void OnRead (IAsyncResult ares)
 		{
-			// TODO: set a limit on ms length.
 			HttpConnection cnc = (HttpConnection) ares.AsyncState;
+			cnc.OnReadInternal (ares);
+		}
+
+		void OnReadInternal (IAsyncResult ares)
+		{
+			timer.Change (Timeout.Infinite, Timeout.Infinite);
 			int nread = -1;
 			try {
 				nread = stream.EndRead (ares);
 				ms.Write (buffer, 0, nread);
-			} catch (Exception e) {
-				//Console.WriteLine (e);
-				if (ms.Length > 0)
+				if (ms.Length > 32768) {
+					SendError ("Bad request", 400);
+					Close (true);
+					return;
+				}
+			} catch {
+				if (ms != null && ms.Length > 0)
 					SendError ();
-				sock.Close ();
+				if (sock != null) {
+					CloseSocket ();
+					Unbind ();
+				}
 				return;
 			}
 
 			if (nread == 0) {
 				//if (ms.Length > 0)
 				//	SendError (); // Why bother?
-				sock.Close ();
+				CloseSocket ();
+				Unbind ();
 				return;
 			}
 
@@ -185,18 +215,35 @@ namespace System.Net {
 
 				if (context.HaveError) {
 					SendError ();
-					Close ();
+					Close (true);
 					return;
 				}
 
 				if (!epl.BindContext (context)) {
 					SendError ("Invalid host", 400);
-					Close ();
+					Close (true);
+					return;
 				}
+				HttpListener listener = context.Listener;
+				if (last_listener != listener) {
+					RemoveConnection ();
+					listener.AddConnection (this);
+					last_listener = listener;
+				}
+
 				context_bound = true;
+				listener.RegisterContext (context);
 				return;
 			}
-			stream.BeginRead (buffer, 0, BufferSize, OnRead, cnc);
+			stream.BeginRead (buffer, 0, BufferSize, onread_cb, this);
+		}
+
+		void RemoveConnection ()
+		{
+			if (last_listener == null)
+				epl.RemoveConnection (this);
+			else
+				last_listener.RemoveConnection (this);
 		}
 
 		enum InputState {
@@ -222,8 +269,19 @@ namespace System.Net {
 			int len = (int) ms.Length;
 			int used = 0;
 			string line;
-			while ((line = ReadLine (buffer, position, len - position, ref used)) != null) {
+
+			try {
+				line = ReadLine (buffer, position, len - position, ref used);
 				position += used;
+			} catch {
+				context.ErrorMessage = "Bad request";
+				context.ErrorStatus = 400;
+				return true;
+			}
+
+			do {
+				if (line == null)
+					break;
 				if (line == "") {
 					if (input_state == InputState.RequestLine)
 						continue;
@@ -236,7 +294,13 @@ namespace System.Net {
 					context.Request.SetRequestLine (line);
 					input_state = InputState.Headers;
 				} else {
-					context.Request.AddHeader (line);
+					try {
+						context.Request.AddHeader (line);
+					} catch (Exception e) {
+						context.ErrorMessage = e.Message;
+						context.ErrorStatus = 400;
+						return true;
+					}
 				}
 
 				if (context.HaveError)
@@ -244,7 +308,15 @@ namespace System.Net {
 
 				if (position >= len)
 					break;
-			}
+				try {
+					line = ReadLine (buffer, position, len - position, ref used);
+					position += used;
+				} catch {
+					context.ErrorMessage = "Bad request";
+					context.ErrorStatus = 400;
+					return true;
+				}
+			} while (line != null);
 
 			if (used == len) {
 				ms.SetLength (0);
@@ -283,18 +355,22 @@ namespace System.Net {
 
 		public void SendError (string msg, int status)
 		{
-			HttpListenerResponse response = context.Response;
-			response.StatusCode = status;
-			response.ContentType = "text/html";
-			string description = HttpListenerResponse.GetStatusDescription (status);
-			string str;
-			if (msg != null)
-				str = String.Format ("<h1>{0} ({1})</h1>", description, msg);
-			else
-				str = String.Format ("<h1>{0}</h1>", description);
+			try {
+				HttpListenerResponse response = context.Response;
+				response.StatusCode = status;
+				response.ContentType = "text/html";
+				string description = HttpListenerResponse.GetStatusDescription (status);
+				string str;
+				if (msg != null)
+					str = String.Format ("<h1>{0} ({1})</h1>", description, msg);
+				else
+					str = String.Format ("<h1>{0}</h1>", description);
 
-			byte [] error = context.Response.ContentEncoding.GetBytes (str);
-			response.Close (error, false);
+				byte [] error = context.Response.ContentEncoding.GetBytes (str);
+				response.Close (error, false);
+			} catch {
+				// response was already closed
+			}
 		}
 
 		public void SendError ()
@@ -302,7 +378,34 @@ namespace System.Net {
 			SendError (context.ErrorMessage, context.ErrorStatus);
 		}
 
+		void Unbind ()
+		{
+			if (context_bound) {
+				epl.UnbindContext (context);
+				context_bound = false;
+			}
+		}
+
 		public void Close ()
+		{
+			Close (false);
+		}
+
+		void CloseSocket ()
+		{
+			if (sock == null)
+				return;
+
+			try {
+				sock.Close ();
+			} catch {
+			} finally {
+				sock = null;
+			}
+			RemoveConnection ();
+		}
+
+		internal void Close (bool force_close)
 		{
 			if (sock != null) {
 				Stream st = GetResponseStream ();
@@ -311,31 +414,49 @@ namespace System.Net {
 			}
 
 			if (sock != null) {
-				if (chunked && context.Response.ForceCloseChunked == false) {
-					// Don't close. Keep working.
-					chunked_uses++;
-					Init ();
-					BeginReadRequest ();
-					return;
-				}
+				force_close |= !context.Request.KeepAlive;
+				if (!force_close)
+					force_close = (context.Response.Headers ["connection"] == "close");
+				/*
+				if (!force_close) {
+//					bool conn_close = (status_code == 400 || status_code == 408 || status_code == 411 ||
+//							status_code == 413 || status_code == 414 || status_code == 500 ||
+//							status_code == 503);
 
-				if (context.Response.Headers ["connection"] == "close") {
-					Socket s = sock;
-					sock = null;
-					try {
-						s.Shutdown (SocketShutdown.Both);
-					} catch {
-					} finally {
-						s.Close ();
+					force_close |= (context.Request.ProtocolVersion <= HttpVersion.Version10);
+				}
+				*/
+
+				if (!force_close && context.Request.FlushInput ()) {
+					if (chunked && context.Response.ForceCloseChunked == false) {
+						// Don't close. Keep working.
+						reuses++;
+						Unbind ();
+						Init ();
+						BeginReadRequest ();
+						return;
 					}
-				} else {
+
+					reuses++;
+					Unbind ();
 					Init ();
 					BeginReadRequest ();
 					return;
 				}
 
-				if (context_bound)
-					epl.UnbindContext (context);
+				Socket s = sock;
+				sock = null;
+				try {
+					if (s != null)
+						s.Shutdown (SocketShutdown.Both);
+				} catch {
+				} finally {
+					if (s != null)
+						s.Close ();
+				}
+				Unbind ();
+				RemoveConnection ();
+				return;
 			}
 		}
 	}
