@@ -246,10 +246,10 @@ namespace System.Net
 		[MonoTODO("Needs to respect Module, Proxy.AutoDetect, and Proxy.ScriptLocation config settings")]
 		static IWebProxy GetDefaultWebProxy ()
 		{
-			WebProxy p = null;
-			
 #if CONFIGURATION_DEP
 			DefaultProxySection sec = ConfigurationManager.GetSection ("system.net/defaultProxy") as DefaultProxySection;
+			WebProxy p;
+			
 			if (sec == null)
 				return GetSystemWebProxy ();
 			
@@ -265,8 +265,11 @@ namespace System.Net
 			
 			if (pe.BypassOnLocal != ProxyElement.BypassOnLocalValues.Unspecified)
 				p.BypassProxyOnLocal = (pe.BypassOnLocal == ProxyElement.BypassOnLocalValues.True);
-#endif
+			
 			return p;
+#else
+			return GetSystemWebProxy ();
+#endif
 		}
 #endif
 
