@@ -108,7 +108,7 @@ namespace System.IO
 				return new IOException (message, unchecked((int)0x80070000) | (int)error);
 			case MonoIOError.ERROR_INVALID_DRIVE:
 				message = String.Format ("Could not find the drive  '{0}'. The drive might not be ready or might not be mapped.", path);
-#if NET_2_0 && !NET_2_1
+#if !NET_2_1
 				return new DriveNotFoundException (message);
 #else
 				return new IOException (message, unchecked((int)0x80070000) | (int)error);
@@ -210,6 +210,18 @@ namespace System.IO
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		public extern static MonoFileType GetFileType (IntPtr handle, out MonoIOError error);
 
+		//
+		// Find file methods
+		//
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		public extern static string FindFirst (string path, string pattern, out FileAttributes result_attr, out MonoIOError error, out IntPtr handle);
+		
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		public extern static string FindNext (IntPtr handle, out FileAttributes result_attr, out MonoIOError error);
+		
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		public extern static int FindClose (IntPtr handle);
+		
 		public static bool Exists (string path, out MonoIOError error)
 		{
 			FileAttributes attrs = GetFileAttributes (path,
