@@ -48,21 +48,11 @@ using System.Text;
 
 namespace System.IO {
 
-#if NET_2_0
 	[ComVisible (true)]
 	public static class Path {
 
 		[Obsolete ("see GetInvalidPathChars and GetInvalidFileNameChars methods.")]
 		public static readonly char[] InvalidPathChars;
-#else
-	public sealed class Path {
-
-		private Path ()
-		{
-		}
-
-		public static readonly char[] InvalidPathChars;
-#endif
 		public static readonly char AltDirectorySeparatorChar;
 		public static readonly char DirectorySeparatorChar;
 		public static readonly char PathSeparator;
@@ -285,9 +275,14 @@ namespace System.IO {
 		public static string GetFullPath (string path)
 		{
 			string fullpath = InsecureGetFullPath (path);
+
+			SecurityManager.EnsureElevatedPermissions (); // this is a no-op outside moonlight
+
+#if !NET_2_1
 			if (SecurityManager.SecurityEnabled) {
 				new FileIOPermission (FileIOPermissionAccess.PathDiscovery, fullpath).Demand ();
 			}
+#endif
 			return fullpath;
 		}
 
@@ -339,20 +334,29 @@ namespace System.IO {
 			// if the supplied path ends with a separator...
 			char end = path [path.Length - 1];
 
+			var canonicalize = true;
 			if (path.Length >= 2 &&
 				IsDsc (path [0]) &&
 				IsDsc (path [1])) {
 				if (path.Length == 2 || path.IndexOf (path [0], 2) < 0)
-					throw new ArgumentException ("UNC pass should be of the form \\\\server\\share.");
+					throw new ArgumentException ("UNC paths should be of the form \\\\server\\share.");
 
 				if (path [0] != DirectorySeparatorChar)
 					path = path.Replace (AltDirectorySeparatorChar, DirectorySeparatorChar);
 
-				path = CanonicalizePath (path);
 			} else {
-				if (!IsPathRooted (path))
+				if (!IsPathRooted (path)) {
+					
+					// avoid calling expensive CanonicalizePath when possible
+					var start = 0;
+					while ((start = path.IndexOf ('.', start)) != -1) {
+						if (++start == path.Length || path [start] == DirectorySeparatorChar || path [start] == AltDirectorySeparatorChar)
+							break;
+					}
+					canonicalize = start > 0;
+					
 					path = Directory.GetCurrentDirectory () + DirectorySeparatorStr + path;
-				else if (DirectorySeparatorChar == '\\' &&
+				} else if (DirectorySeparatorChar == '\\' &&
 					path.Length >= 2 &&
 					IsDsc (path [0]) &&
 					!IsDsc (path [1])) { // like `\abc\def'
@@ -362,8 +366,10 @@ namespace System.IO {
 					else
 						path = current.Substring (0, current.IndexOf ('\\', current.IndexOf ("\\\\") + 1));
 				}
-				path = CanonicalizePath (path);
 			}
+			
+			if (canonicalize)
+			    path = CanonicalizePath (path);
 
 			// if the original ended with a [Alt]DirectorySeparatorChar then ensure the full path also ends with one
 			if (IsDsc (end) && (path [path.Length - 1] != DirectorySeparatorChar))
@@ -434,6 +440,8 @@ namespace System.IO {
 			Random rnd;
 			int num = 0;
 
+			SecurityManager.EnsureElevatedPermissions (); // this is a no-op outside moonlight
+
 			rnd = new Random ();
 			do {
 				num = rnd.Next ();
@@ -448,6 +456,14 @@ namespace System.IO {
 					// avoid an endless loop
 					throw;
 				}
+				catch (UnauthorizedAccessException) {
+					// This can happen if we don't have write permission to /tmp
+					throw;
+				}
+				catch (DirectoryNotFoundException) {
+					// This happens when TMPDIR does not exist
+					throw;
+				}
 				catch {
 				}
 			} while (f == null);
@@ -459,6 +475,8 @@ namespace System.IO {
 		[EnvironmentPermission (SecurityAction.Demand, Unrestricted = true)]
 		public static string GetTempPath ()
 		{
+			SecurityManager.EnsureElevatedPermissions (); // this is a no-op outside moonlight
+
 			string p = get_temp_path ();
 			if (p.Length > 0 && p [p.Length - 1] != DirectorySeparatorChar)
 				return p + DirectorySeparatorChar;
@@ -495,7 +513,6 @@ namespace System.IO {
 				(!dirEqualsVolume && path.Length > 1 && path [1] == VolumeSeparatorChar));
 		}
 
-#if NET_2_0
 		public static char[] GetInvalidFileNameChars ()
 		{
 			// return a new array as we do not want anyone to be able to change the values
@@ -543,7 +560,7 @@ namespace System.IO {
 
 			return sb.ToString ();
 		}
-#endif
+
 		// private class methods
 
 		private static int findExtension (string path)
@@ -567,17 +584,8 @@ namespace System.IO {
 			AltDirectorySeparatorChar = MonoIO.AltDirectorySeparatorChar;
 
 			PathSeparator = MonoIO.PathSeparator;
-#if NET_2_0
 			// this copy will be modifiable ("by design")
 			InvalidPathChars = GetInvalidPathChars ();
-#else
-			if (Environment.IsRunningOnWindows) {
-				InvalidPathChars = new char [15] { '\x00', '\x08', '\x10', '\x11', '\x12', '\x14', '\x15', '\x16',
-					'\x17', '\x18', '\x19', '\x22', '\x3C', '\x3E', '\x7C' };
-			} else {
-				InvalidPathChars = new char [1] { '\x00' };
-			}
-#endif
 			// internal fields
 
 			DirectorySeparatorStr = DirectorySeparatorChar.ToString ();
@@ -667,7 +675,7 @@ namespace System.IO {
 			for (int i = 0; i < dirs.Length; i++) {
 				// WIN32 path components must be trimmed
 				if (Environment.IsRunningOnWindows)
-					dirs[i] = dirs[i].Trim ();
+					dirs[i] = dirs[i].TrimEnd ();
 				
 				if (dirs[i] == "." || (i != 0 && dirs[i].Length == 0))
 					continue;
@@ -738,5 +746,106 @@ namespace System.IO {
 
 			return String.Compare (subset, slast, path, slast, subset.Length - slast) == 0;
 		}
+
+#if NET_4_0 || MOONLIGHT || MOBILE
+		public
+#else
+                internal
+#endif
+		static string Combine (params string [] paths)
+		{
+			if (paths == null)
+				throw new ArgumentNullException ("paths");
+
+			bool need_sep;
+			var ret = new StringBuilder ();
+			int pathsLen = paths.Length;
+			int slen;
+			foreach (var s in paths) {
+				need_sep = false;
+				if (s == null)
+					throw new ArgumentNullException ("One of the paths contains a null value", "paths");
+				if (s.IndexOfAny (InvalidPathChars) != -1)
+					throw new ArgumentException ("Illegal characters in path.");
+				
+				pathsLen--;
+				if (IsPathRooted (s))
+					ret.Length = 0;
+				
+				ret.Append (s);
+				slen = s.Length;
+				if (slen > 0 && pathsLen > 0) {
+					char p1end = s [slen - 1];
+					if (p1end != DirectorySeparatorChar && p1end != AltDirectorySeparatorChar && p1end != VolumeSeparatorChar)
+						need_sep = true;
+				}
+				
+				if (need_sep)
+					ret.Append (DirectorySeparatorStr);
+			}
+
+			return ret.ToString ();
+		}
+
+#if NET_4_0 || MOONLIGHT || MOBILE
+		public
+#else
+                internal
+#endif
+		static string Combine (string path1, string path2, string path3)
+		{
+			if (path1 == null)
+				throw new ArgumentNullException ("path1");
+
+			if (path2 == null)
+				throw new ArgumentNullException ("path2");
+
+			if (path3 == null)
+				throw new ArgumentNullException ("path3");
+			
+			return Combine (new string [] { path1, path2, path3 });
+		}
+
+#if NET_4_0 || MOONLIGHT || MOBILE
+		public
+#else
+                internal
+#endif
+		static string Combine (string path1, string path2, string path3, string path4)
+		{
+			if (path1 == null)
+				throw new ArgumentNullException ("path1");
+
+			if (path2 == null)
+				throw new ArgumentNullException ("path2");
+
+			if (path3 == null)
+				throw new ArgumentNullException ("path3");
+
+			if (path4 == null)
+				throw new ArgumentNullException ("path4");
+			
+			return Combine (new string [] { path1, path2, path3, path4 });
+		}
+
+		internal static void Validate (string path)
+		{
+			Validate (path, "path");
+		}
+
+		internal static void Validate (string path, string parameterName)
+		{
+			if (path == null)
+				throw new ArgumentNullException (parameterName);
+			if (String.IsNullOrWhiteSpace (path))
+				throw new ArgumentException (Locale.GetText ("Path is empty"));
+			if (path.IndexOfAny (Path.InvalidPathChars) != -1)
+				throw new ArgumentException (Locale.GetText ("Path contains invalid chars"));
+#if MOONLIGHT
+			// On Moonlight (SL4+) there are some limitations in "Elevated Trust"
+			if (SecurityManager.HasElevatedPermissions) {
+			}
+#endif
+		}
 	}
 }
