@@ -16,9 +16,10 @@ namespace System.Runtime.InteropServices
 	[InterfaceType (ComInterfaceType.InterfaceIsIUnknown)]
 	[Guid ("f7102fa9-cabb-3a74-a6da-b4567ef1b079")]
 	[TypeLibImportClass (typeof (MemberInfo))]
+	[ComVisible (true)]
 	public interface _MemberInfo
 	{
-		bool Equals (object obj);
+		bool Equals (object other);
 
 		object[] GetCustomAttributes (bool inherit);
 
