@@ -22,7 +22,7 @@
 //
 //
 
-#if NET_4_0 || BOOTSTRAP_NET_4_0
+#if NET_4_0 || MOBILE
 using System;
 
 namespace System.Threading.Tasks
@@ -42,6 +42,8 @@ namespace System.Threading.Tasks
 			get {
 				if (function != null)
 					Wait ();
+				else if (Exception != null)
+					throw Exception;
 				return value;
 			}
 			internal set {
@@ -61,19 +63,19 @@ namespace System.Threading.Tasks
 		}
 		
 		public Task (Func<TResult> function, CancellationToken cancellationToken)
-			: this ((o) => function (), null, cancellationToken, TaskCreationOptions.None)
+			: this (function == null ? (Func<object, TResult>)null : (o) => function(), null, cancellationToken, TaskCreationOptions.None)
 		{
 			
 		}
 		
 		public Task (Func<TResult> function, TaskCreationOptions creationOptions)
-			: this ((o) => function (), null, CancellationToken.None, creationOptions)
+			: this (function == null ? (Func<object, TResult>)null : (o) => function(), null, CancellationToken.None, creationOptions)
 		{
 			
 		}
 		
 		public Task (Func<TResult> function, CancellationToken cancellationToken, TaskCreationOptions creationOptions)
-			: this ((o) => function (), null, cancellationToken, creationOptions)
+			: this (function == null ? (Func<object, TResult>)null : (o) => function(), null, cancellationToken, creationOptions)
 		{
 			
 		}
