@@ -45,28 +45,25 @@ using System.Collections;
 using System.Globalization;
 using System.Runtime.CompilerServices;
 
-#if NET_2_0
 using System.Collections.Generic;
 using System.Runtime.ConstrainedExecution;
 using System.Runtime.InteropServices;
 using Mono.Globalization.Unicode;
-#endif
+
 
 namespace System
 {
 	[Serializable]
-#if NET_2_0
 	[ComVisible (true)]
 	public sealed class String : IConvertible, ICloneable, IEnumerable, IComparable, IComparable<String>, IEquatable <String>, IEnumerable<char>
-#else
-	public sealed class String : IConvertible, ICloneable, IEnumerable, IComparable
-#endif
 	{
 		[NonSerialized] private int length;
 		[NonSerialized] private char start_char;
 
 		public static readonly String Empty = "";
 
+		internal static readonly int LOS_limit = GetLOSLimit ();
+
 		public static unsafe bool Equals (string a, string b)
 		{
 			if ((a as object) == (b as object))
@@ -129,17 +126,13 @@ namespace System
 			return !Equals (a, b);
 		}
 
-#if NET_2_0
 		[ReliabilityContractAttribute (Consistency.WillNotCorruptState, Cer.MayFail)]
-#endif
 		public override bool Equals (Object obj)
 		{
 			return Equals (this, obj as String);
 		}
 
-#if NET_2_0
 		[ReliabilityContractAttribute (Consistency.WillNotCorruptState, Cer.MayFail)]
-#endif
 		public bool Equals (String value)
 		{
 			return Equals (this, value);
@@ -184,12 +177,9 @@ namespace System
 				CharCopy (dest + destinationIndex, src + sourceIndex, count);
 		}
 
-		public unsafe char[] ToCharArray ()
+		public char[] ToCharArray ()
 		{
-			char[] tmp = new char [length];
-			fixed (char* dest = tmp, src = this)
-				CharCopy (dest, src, length);
-			return tmp;
+			return ToCharArray (0, length);
 		}
 
 		public unsafe char[] ToCharArray (int startIndex, int length)
@@ -229,7 +219,6 @@ namespace System
 			return InternalSplit (separator, count, 0);
 		}
 
-#if NET_2_0
 		[ComVisible (false)]
 		[MonoDocumentationNote ("code should be moved to managed")]
 		public String[] Split (char[] separator, int count, StringSplitOptions options)
@@ -258,13 +247,15 @@ namespace System
 				throw new ArgumentOutOfRangeException ("count", "Count cannot be less than zero.");
 			if ((options != StringSplitOptions.None) && (options != StringSplitOptions.RemoveEmptyEntries))
 				throw new ArgumentException ("Illegal enum value: " + options + ".");
+			if (count == 1)
+				return new String [] { this };
 
 			bool removeEmpty = (options & StringSplitOptions.RemoveEmptyEntries) == StringSplitOptions.RemoveEmptyEntries;
 
 			if (count == 0 || (this == String.Empty && removeEmpty))
 				return new String [0];
 
-			ArrayList arr = new ArrayList ();
+			List<String> arr = new List<String> ();
 
 			int pos = 0;
 			int matchCount = 0;
@@ -288,34 +279,28 @@ namespace System
 				if (matchIndex == -1)
 					break;
 
-				if (!(matchPos == pos && removeEmpty))
+				if (!(matchPos == pos && removeEmpty)) {
+					if (arr.Count == count - 1)
+						break;
 					arr.Add (this.Substring (pos, matchPos - pos));
+				}
 
 				pos = matchPos + separator [matchIndex].Length;
 
 				matchCount ++;
-
-				if (matchCount == count - 1)
-					break;
 			}
 
 			if (matchCount == 0)
 				return new String [] { this };
-			else {
-				if (removeEmpty && pos == this.Length) {
-					String[] res = new String [arr.Count];
-					arr.CopyTo (0, res, 0, arr.Count);
 
-					return res;
-				}
-				else {
-					String[] res = new String [arr.Count + 1];
-					arr.CopyTo (0, res, 0, arr.Count);
-					res [arr.Count] = this.Substring (pos);
+			// string contained only separators
+			if (removeEmpty && matchCount != 0 && pos == this.Length && arr.Count == 0)
+				return new String [0];
 
-					return res;
-				}
-			}
+			if (!(removeEmpty && pos == this.Length))
+				arr.Add (this.Substring (pos));
+
+			return arr.ToArray ();
 		}
 
 		[ComVisible (false)]
@@ -329,22 +314,13 @@ namespace System
 		{
 			return Split (separator, Int32.MaxValue, options);
 		}
-#endif
 
 		public String Substring (int startIndex)
 		{
-#if NET_2_0
 			if (startIndex == 0)
 				return this;
 			if (startIndex < 0 || startIndex > this.length)
 				throw new ArgumentOutOfRangeException ("startIndex");
-#else
-			if (startIndex < 0)
-				throw new ArgumentOutOfRangeException ("startIndex", "Cannot be negative.");
-
-			if (startIndex > this.length)
-				throw new ArgumentOutOfRangeException ("length", "Cannot exceed length of string.");
-#endif
 
 			return SubstringUnchecked (startIndex, this.length - startIndex);
 		}
@@ -355,16 +331,12 @@ namespace System
 				throw new ArgumentOutOfRangeException ("length", "Cannot be negative.");
 			if (startIndex < 0)
 				throw new ArgumentOutOfRangeException ("startIndex", "Cannot be negative.");
-#if NET_2_0
 			if (startIndex > this.length)
 				throw new ArgumentOutOfRangeException ("startIndex", "Cannot exceed length of string.");
-#endif
 			if (startIndex > this.length - length)
 				throw new ArgumentOutOfRangeException ("length", "startIndex + length > this.length");
-#if NET_2_0
 			if (startIndex == 0 && length == this.length)
 				return this;
-#endif
 
 			return SubstringUnchecked (startIndex, length);
 		}
@@ -383,13 +355,17 @@ namespace System
 			return tmp;
 		}
 
-		private static readonly char[] WhiteChars = { (char) 0x9, (char) 0xA, (char) 0xB, (char) 0xC, (char) 0xD,
-#if NET_2_0
+		private static readonly char[] WhiteChars = {
+			(char) 0x9, (char) 0xA, (char) 0xB, (char) 0xC, (char) 0xD,
 			(char) 0x85, (char) 0x1680, (char) 0x2028, (char) 0x2029,
-#endif
 			(char) 0x20, (char) 0xA0, (char) 0x2000, (char) 0x2001, (char) 0x2002, (char) 0x2003, (char) 0x2004,
 			(char) 0x2005, (char) 0x2006, (char) 0x2007, (char) 0x2008, (char) 0x2009, (char) 0x200A, (char) 0x200B,
-			(char) 0x3000, (char) 0xFEFF };
+			(char) 0x3000, (char) 0xFEFF,
+#if NET_2_1
+		        // Silverlight 
+		        (char) 0x202f, (char) 0x205f,
+#endif
+		};
 
 		public String Trim ()
 		{
@@ -475,9 +451,12 @@ namespace System
 				}
 				else {
 					if (c != 0xA0 && c != 0xFEFF && c != 0x3000) {
-#if NET_2_0
-						if (c != 0x85 && c != 0x1680 && c != 0x2028 && c != 0x2029)
+						if (c != 0x85 && c != 0x1680 && c != 0x2028 && c != 0x2029
+#if NET_2_1
+						    // On Silverlight this whitespace participates in Trim
+						    && c != 0x202f && c != 0x205f
 #endif
+							)
 							if (c < 0x2000 || c > 0x200B)
 								return pos;
 					}
@@ -582,7 +561,7 @@ namespace System
 			// ENHANCE: Might call internal_compare_switch directly instead of doing all checks twice
 			return culture.CompareInfo.Compare (strA, indexA, len1, strB, indexB, len2, compopts);
 		}
-#if NET_2_0
+
 		public static int Compare (string strA, string strB, StringComparison comparisonType)
 		{
 			switch (comparisonType) {
@@ -659,7 +638,6 @@ namespace System
 
 			return culture.CompareInfo.Compare (strA, indexA, len1, strB, indexB, len2, options);
 		}
-#endif
 
 		public int CompareTo (Object value)
 		{
@@ -774,12 +752,7 @@ namespace System
 			return CultureInfo.CurrentCulture.CompareInfo.IsSuffix (this, value, CompareOptions.None);
 		}
 
-#if NET_2_0
-		public
-#else
-		internal
-#endif
-		bool EndsWith (String value, bool ignoreCase, CultureInfo culture)
+		public bool EndsWith (String value, bool ignoreCase, CultureInfo culture)
 		{
 			if (value == null)
 				throw new ArgumentNullException ("value");
@@ -874,7 +847,6 @@ namespace System
 		}
 
 
-#if NET_2_0
 		public int IndexOf (string value, StringComparison comparisonType)
 		{
 			return IndexOf (value, 0, this.Length, comparisonType);
@@ -905,7 +877,6 @@ namespace System
 				throw new ArgumentException (msg, "comparisonType");
 			}
 		}
-#endif
 
 		internal int IndexOfOrdinal (string value, int startIndex, int count, CompareOptions options)
 		{
@@ -930,7 +901,7 @@ namespace System
 			if (valueLen <= 1) {
 				if (valueLen == 1)
 					return IndexOfUnchecked (value[0], startIndex, count);
-				return 0;
+				return startIndex;
 			}
 
 			fixed (char* thisptr = this, valueptr = value) {
@@ -958,7 +929,7 @@ namespace System
 				return -1;
 
 			if (valueLen == 0)
-				return 0;
+				return startIndex;
 
 			fixed (char* thisptr = this, valueptr = value) {
 				char* ap = thisptr + startIndex;
@@ -976,8 +947,6 @@ namespace System
 			return -1;
 		}
 
-#if NET_2_0
-
 		public int LastIndexOf (string value, StringComparison comparisonType)
 		{
 			if (this.Length == 0)
@@ -1011,12 +980,15 @@ namespace System
 				throw new ArgumentException (msg, "comparisonType");
 			}
 		}
-#endif
 
 		internal int LastIndexOfOrdinal (string value, int startIndex, int count, CompareOptions options)
 		{
 			if (value == null)
 				throw new ArgumentNullException ("value");
+			if (this.Length == 0)
+				return value == String.Empty ? 0 : -1;
+			if (value.Length == 0)
+				return Math.Min (this.Length - 1, startIndex);
 			if (startIndex < 0 || startIndex > length)
 				throw new ArgumentOutOfRangeException ("startIndex");
 			if (count < 0 || (startIndex < count - 1))
@@ -1036,7 +1008,7 @@ namespace System
 			if (valueLen <= 1) {
 				if (valueLen == 1)
 					return LastIndexOfUnchecked (value[0], startIndex, count);
-				return 0;
+				return startIndex;
 			}
 
 			fixed (char* thisptr = this, valueptr = value) {
@@ -1064,7 +1036,7 @@ namespace System
 				return -1;
 
 			if (valueLen == 0)
-				return 0;
+				return startIndex;
 
 			fixed (char* thisptr = this, valueptr = value) {
 				char* ap = thisptr + startIndex - valueLen + 1;
@@ -1194,11 +1166,7 @@ namespace System
 		public int IndexOf (String value, int startIndex, int count)
 		{
 			if (value == null)
-#if NET_2_0
 				throw new ArgumentNullException ("value");
-#else
-				throw new ArgumentNullException ("string2");
-#endif
 			if (startIndex < 0 || startIndex > this.length)
 				throw new ArgumentOutOfRangeException ("startIndex", "Cannot be negative, and should not exceed length of string.");
 			if (count < 0 || startIndex > this.length - count)
@@ -1221,6 +1189,8 @@ namespace System
 		{
 			if (anyOf == null)
 				throw new ArgumentNullException ();
+			if (this.length == 0)
+				return -1;
 
 			return LastIndexOfAnyUnchecked (anyOf, this.length - 1, this.length);
 		}
@@ -1229,6 +1199,8 @@ namespace System
 		{
 			if (anyOf == null)
 				throw new ArgumentNullException ();
+			if (this.length == 0)
+				return -1;
 
 			if (startIndex < 0 || startIndex >= this.length)
 				throw new ArgumentOutOfRangeException ("startIndex", "Cannot be negative, and should be less than length of string.");
@@ -1243,6 +1215,8 @@ namespace System
 		{
 			if (anyOf == null) 
 				throw new ArgumentNullException ();
+			if (this.length == 0)
+				return -1;
 
 			if ((startIndex < 0) || (startIndex >= this.Length))
 				throw new ArgumentOutOfRangeException ("startIndex", "< 0 || > this.Length");
@@ -1297,9 +1271,9 @@ namespace System
 
 		public int LastIndexOf (char value, int startIndex, int count)
 		{
-			if (startIndex == 0 && this.length == 0)
+			if (this.length == 0)
 				return -1;
-
+ 
 			// >= for char (> for string)
 			if ((startIndex < 0) || (startIndex >= this.Length))
 				throw new ArgumentOutOfRangeException ("startIndex", "< 0 || >= this.Length");
@@ -1369,11 +1343,7 @@ namespace System
 		// Following methods are culture-sensitive
 		public int LastIndexOf (String value)
 		{
-			if (this.length == 0)
-				// This overload does additional checking
-				return LastIndexOf (value, 0, 0);
-			else
-				return LastIndexOf (value, this.length - 1, this.length);
+			return LastIndexOf (value, this.length - 1, this.length);
 		}
 
 		public int LastIndexOf (String value, int startIndex)
@@ -1387,12 +1357,10 @@ namespace System
 		public int LastIndexOf (String value, int startIndex, int count)
 		{
 			if (value == null)
-#if NET_2_0
 				throw new ArgumentNullException ("value");
-#else
-				throw new ArgumentNullException ("string2");
-#endif
 
+			if (this.length == 0)
+				return value == String.Empty ? 0 : -1;
 			// -1 > startIndex > for string (0 > startIndex >= for char)
 			if ((startIndex < -1) || (startIndex > this.Length))
 				throw new ArgumentOutOfRangeException ("startIndex", "< 0 || > this.Length");
@@ -1402,7 +1370,7 @@ namespace System
 				throw new ArgumentOutOfRangeException ("startIndex - count + 1 < 0");
 
 			if (value.Length == 0)
-				return startIndex;
+				return Math.Min (this.Length - 1, startIndex);
 
 			if (startIndex == 0 && this.length == 0)
 				return -1;
@@ -1419,7 +1387,6 @@ namespace System
 			return CultureInfo.CurrentCulture.CompareInfo.LastIndexOf (this, value, startIndex, count);
 		}
 
-#if NET_2_0
 		public bool Contains (String value)
 		{
 			return IndexOf (value) != -1;
@@ -1430,6 +1397,7 @@ namespace System
 			return (value == null) || (value.Length == 0);
 		}
 
+#if !MOONLIGHT
 		public string Normalize ()
 		{
 			return Normalization.Normalize (this, 0);
@@ -1467,6 +1435,7 @@ namespace System
 				return Normalization.IsNormalized (this, 3);
 			}
 		}
+#endif
 
 		public string Remove (int startIndex)
 		{
@@ -1477,7 +1446,6 @@ namespace System
 
 			return Remove (startIndex, this.length - startIndex);
 		}
-#endif
 
 		public String PadLeft (int totalWidth)
 		{
@@ -1521,6 +1489,8 @@ namespace System
 
 			if (totalWidth < this.length)
 				return this;
+			if (totalWidth == 0)
+				return String.Empty;
 
 			String tmp = InternalAllocateStr (totalWidth);
 
@@ -1543,7 +1513,6 @@ namespace System
 			return CultureInfo.CurrentCulture.CompareInfo.IsPrefix (this, value, CompareOptions.None);
 		}
 
-#if NET_2_0
 		[ComVisible (false)]
 		public bool StartsWith (string value, StringComparison comparisonType)
 		{
@@ -1593,14 +1562,8 @@ namespace System
 				throw new ArgumentException (msg, "comparisonType");
 			}
 		}
-#endif
 
-#if NET_2_0
-		public
-#else
-		internal
-#endif
-		bool StartsWith (String value, bool ignoreCase, CultureInfo culture)
+		public bool StartsWith (String value, bool ignoreCase, CultureInfo culture)
 		{
 			if (culture == null)
 				culture = CultureInfo.CurrentCulture;
@@ -1766,12 +1729,11 @@ namespace System
 			return culture.TextInfo.ToLower (this);
 		}
 
-#if NET_2_0
 		public unsafe String ToLowerInvariant ()
-#else
-		internal unsafe String ToLowerInvariant ()
-#endif
 		{
+			if (length == 0)
+				return String.Empty;
+
 			string tmp = InternalAllocateStr (length);
 			fixed (char* source = &start_char, dest = tmp) {
 
@@ -1803,12 +1765,11 @@ namespace System
 			return culture.TextInfo.ToUpper (this);
 		}
 
-#if NET_2_0
 		public unsafe String ToUpperInvariant ()
-#else
-		internal unsafe String ToUpperInvariant ()
-#endif
 		{
+			if (length == 0)
+				return String.Empty;
+
 			string tmp = InternalAllocateStr (length);
 			fixed (char* source = &start_char, dest = tmp) {
 
@@ -1987,34 +1948,9 @@ namespace System
 			return arg0.ToString ();
 		}
 
-		public unsafe static String Concat (Object arg0, Object arg1)
+		public static String Concat (Object arg0, Object arg1)
 		{
-			string s1, s2;
-
-			s1 = (arg0 != null) ? arg0.ToString () : null;
-			s2 = (arg1 != null) ? arg1.ToString () : null;
-			
-			if (s1 == null) {
-				if (s2 == null)
-					return String.Empty;
-				else
-					return s2;
-			} else if (s2 == null)
-				return s1;
-
-			String tmp = InternalAllocateStr (s1.Length + s2.Length);
-			if (s1.Length != 0) {
-				fixed (char *dest = tmp, src = s1) {
-					CharCopy (dest, src, s1.length);
-				}
-			}
-			if (s2.Length != 0) {
-				fixed (char *dest = tmp, src = s2) {
-					CharCopy (dest + s1.Length, src, s2.length);
-				}
-			}
-
-			return tmp;
+			return Concat ((arg0 != null) ? arg0.ToString () : null, (arg1 != null) ? arg1.ToString () : null);
 		}
 
 		public static String Concat (Object arg0, Object arg1, Object arg2)
@@ -2038,7 +1974,6 @@ namespace System
 			return Concat (s1, s2, s3);
 		}
 
-#if ! BOOTSTRAP_WITH_OLDLIB
 		[CLSCompliant(false)]
 		public static String Concat (Object arg0, Object arg1, Object arg2,
 					     Object arg3, __arglist)
@@ -2076,7 +2011,6 @@ namespace System
 
 			return Concat (s1, s2, s3, s4);			
 		}
-#endif
 
 		public unsafe static String Concat (String str0, String str1)
 		{
@@ -2201,8 +2135,6 @@ namespace System
 					len += strings[i].length;
 				}
 			}
-			if (len == 0)
-				return String.Empty;
 
 			return ConcatInternal (strings, len);
 		}
@@ -2218,14 +2150,15 @@ namespace System
 				if (s != null)
 					len += s.length;
 			}
-			if (len == 0)
-				return String.Empty;
 
 			return ConcatInternal (values, len);
 		}
 
 		private static unsafe String ConcatInternal (String[] values, int length)
 		{
+			if (length == 0)
+				return String.Empty;
+
 			String tmp = InternalAllocateStr (length);
 
 			fixed (char* dest = tmp) {
@@ -2284,7 +2217,11 @@ namespace System
 			return InternalIsInterned (str);
 		}
 	
+#if NET_4_0 || MOONLIGHT || MOBILE
+		public static string Join (string separator, params string [] value)
+#else
 		public static string Join (string separator, string [] value)
+#endif
 		{
 			if (value == null)
 				throw new ArgumentNullException ("value");
@@ -2408,63 +2345,37 @@ namespace System
 			return Convert.ToInt64 (this, provider);
 		}
 
-#if ONLY_1_1
-#pragma warning disable 3019
-		[CLSCompliant (false)]
-#endif
 		sbyte IConvertible.ToSByte (IFormatProvider provider)
 		{
 			return Convert.ToSByte (this, provider);
 		}
-#if ONLY_1_1
-#pragma warning restore 3019
-#endif
 
 		float IConvertible.ToSingle (IFormatProvider provider)
 		{
 			return Convert.ToSingle (this, provider);
 		}
 
-		object IConvertible.ToType (Type type, IFormatProvider provider)
+		object IConvertible.ToType (Type targetType, IFormatProvider provider)
 		{
-			return Convert.ToType (this, type, provider, false);
+			if (targetType == null)
+				throw new ArgumentNullException ("type");
+			return Convert.ToType (this, targetType, provider, false);
 		}
 
-#if ONLY_1_1
-#pragma warning disable 3019
-		[CLSCompliant (false)]
-#endif
 		ushort IConvertible.ToUInt16 (IFormatProvider provider)
 		{
 			return Convert.ToUInt16 (this, provider);
 		}
-#if ONLY_1_1
-#pragma warning restore 3019
-#endif
 
-#if ONLY_1_1
-#pragma warning disable 3019
-		[CLSCompliant (false)]
-#endif
 		uint IConvertible.ToUInt32 (IFormatProvider provider)
 		{
 			return Convert.ToUInt32 (this, provider);
 		}
-#if ONLY_1_1
-#pragma warning restore 3019
-#endif
 
-#if ONLY_1_1
-#pragma warning disable 3019
-		[CLSCompliant (false)]
-#endif
 		ulong IConvertible.ToUInt64 (IFormatProvider provider)
 		{
 			return Convert.ToUInt64 (this, provider);
 		}
-#if ONLY_1_1
-#pragma warning restore 3019
-#endif
 
 		public int Length {
 			get {
@@ -2477,12 +2388,10 @@ namespace System
 			return new CharEnumerator (this);
 		}
 
-#if NET_2_0
 		IEnumerator<char> IEnumerable<char>.GetEnumerator ()
 		{
 			return new CharEnumerator (this);
 		}
-#endif
 
 		IEnumerator IEnumerable.GetEnumerator ()
 		{
@@ -2492,68 +2401,66 @@ namespace System
 		private static void ParseFormatSpecifier (string str, ref int ptr, out int n, out int width,
 		                                          out bool left_align, out string format)
 		{
+			int max = str.Length;
+			
 			// parses format specifier of form:
 			//   N,[\ +[-]M][:F]}
 			//
 			// where:
-
-			try {
-				// N = argument number (non-negative integer)
-
-				n = ParseDecimal (str, ref ptr);
-				if (n < 0)
-					throw new FormatException ("Input string was not in a correct format.");
-
-				// M = width (non-negative integer)
-
-				if (str[ptr] == ',') {
-					// White space between ',' and number or sign.
-					++ptr;
-					while (Char.IsWhiteSpace (str [ptr]))
-						++ptr;
-					int start = ptr;
-
-					format = str.Substring (start, ptr - start);
-
-					left_align = (str [ptr] == '-');
-					if (left_align)
-						++ ptr;
-
-					width = ParseDecimal (str, ref ptr);
-					if (width < 0)
-						throw new FormatException ("Input string was not in a correct format.");
-				}
-				else {
-					width = 0;
-					left_align = false;
-					format = String.Empty;
-				}
-
-				// F = argument format (string)
-
-				if (str[ptr] == ':') {
-					int start = ++ ptr;
-					while (str[ptr] != '}')
-						++ ptr;
-
-					format += str.Substring (start, ptr - start);
-				}
-				else
-					format = null;
-
-				if (str[ptr ++] != '}')
-					throw new FormatException ("Input string was not in a correct format.");
-			}
-			catch (IndexOutOfRangeException) {
+			// N = argument number (non-negative integer)
+			
+			n = ParseDecimal (str, ref ptr);
+			if (n < 0)
 				throw new FormatException ("Input string was not in a correct format.");
+			
+			// M = width (non-negative integer)
+			
+			if (ptr < max && str[ptr] == ',') {
+				// White space between ',' and number or sign.
+				++ptr;
+				while (ptr < max && Char.IsWhiteSpace (str [ptr]))
+					++ptr;
+				int start = ptr;
+				
+				format = str.Substring (start, ptr - start);
+				
+				left_align = (ptr < max && str [ptr] == '-');
+				if (left_align)
+					++ ptr;
+				
+				width = ParseDecimal (str, ref ptr);
+				if (width < 0)
+					throw new FormatException ("Input string was not in a correct format.");
 			}
+			else {
+				width = 0;
+				left_align = false;
+				format = String.Empty;
+			}
+			
+			// F = argument format (string)
+			
+			if (ptr < max && str[ptr] == ':') {
+				int start = ++ ptr;
+				while (ptr < max && str[ptr] != '}')
+					++ ptr;
+				
+				format += str.Substring (start, ptr - start);
+			}
+			else
+				format = null;
+			
+			if ((ptr >= max) || str[ptr ++] != '}')
+				throw new FormatException ("Input string was not in a correct format.");
 		}
 
 		private static int ParseDecimal (string str, ref int ptr)
 		{
 			int p = ptr;
 			int n = 0;
-			while (true) {
+			int max = str.Length;
+			
+			while (p < max) {
 				char c = str[p];
 				if (c < '0' || '9' < c)
 					break;
@@ -2562,7 +2469,7 @@ namespace System
 				++ p;
 			}
 
-			if (p == ptr)
+			if (p == ptr || p == max)
 				return -1;
 
 			ptr = p;
@@ -2599,9 +2506,7 @@ namespace System
 			length = newLength;
 		}
 
-#if NET_2_0
 		[ReliabilityContractAttribute (Consistency.WillNotCorruptState, Cer.MayFail)]
-#endif
 		// When modifying it, GetCaseInsensitiveHashCode() should be modified as well.
 		public unsafe override int GetHashCode ()
 		{
@@ -2620,6 +2525,102 @@ namespace System
 			}
 		}
 
+#if MOONLIGHT || MOBILE || NET_4_0
+		[ComVisible(false)]
+		public static string Concat (IEnumerable<string> values)
+		{
+			if (values == null)
+				throw new ArgumentNullException ("values");
+
+			var stringList = new List<string> ();
+			int len = 0;
+			foreach (var v in values){
+				if (v == null)
+					continue;
+				len += v.Length;
+				stringList.Add (v);
+			}
+			return ConcatInternal (stringList.ToArray (), len);
+		}
+
+		[ComVisibleAttribute(false)]
+		public static string Concat<T> (IEnumerable<T> values)
+		{
+			if (values == null)
+				throw new ArgumentNullException ("values");
+
+			var stringList = new List<string> ();
+			int len = 0;
+			foreach (var v in values){
+				string sr = v.ToString ();
+				len += sr.Length;
+				stringList.Add (sr);
+			}
+			return ConcatInternal (stringList.ToArray (), len);
+		}
+
+		[ComVisibleAttribute(false)]
+		public static string Join (string separator, IEnumerable<string> values)
+		{
+			if (separator == null)
+				return Concat (values);
+			
+			if (values == null)
+				throw new ArgumentNullException ("values");
+			
+			var stringList = new List<string> ();
+			foreach (var v in values)
+				stringList.Add (v);
+
+			return JoinUnchecked (separator, stringList.ToArray (), 0, stringList.Count);
+		}
+
+		[ComVisibleAttribute(false)]
+		public static string Join (string separator, params object [] values)
+		{
+			if (separator == null)
+				return Concat (values);
+			
+			if (values == null)
+				throw new ArgumentNullException ("values");
+
+			var strCopy = new string [values.Length];
+			int i = 0;
+			foreach (var v in values)
+				strCopy [i++] = v.ToString ();
+
+			return JoinUnchecked (separator, strCopy, 0, strCopy.Length);
+		}
+		
+		[ComVisible (false)]
+		public static string Join<T> (string separator, IEnumerable<T> values)
+		{
+			if (separator == null)
+				return Concat<T> (values);
+				
+			if (values == null)
+				throw new ArgumentNullException ("values");
+			
+			var stringList = new List<string> ();
+			foreach (var v in values)
+				stringList.Add (v.ToString ());
+
+			return JoinUnchecked (separator, stringList.ToArray (), 0, stringList.Count);
+		}
+
+		public static bool IsNullOrWhiteSpace (string value)
+#else
+		internal static bool IsNullOrWhiteSpace (string value)
+#endif
+		{
+			if ((value == null) || (value.Length == 0))
+				return true;
+			foreach (char c in value)
+				if (!Char.IsWhiteSpace (c))
+					return false;
+			return true;
+		}
+
 		internal unsafe int GetCaseInsensitiveHashCode ()
 		{
 			fixed (char * c = this) {
@@ -2653,11 +2654,9 @@ namespace System
 					length++;
 			} catch (NullReferenceException) {
 				throw new ArgumentOutOfRangeException ("ptr", "Value does not refer to a valid string.");
-#if NET_2_0
 			} catch (AccessViolationException) {
 				throw new ArgumentOutOfRangeException ("ptr", "Value does not refer to a valid string.");
-#endif
-			}
+		}
 
 			return CreateString (value, 0, length, null);
 		}
@@ -2679,13 +2678,9 @@ namespace System
 			bool isDefaultEncoding;
 
 			if (isDefaultEncoding = (enc == null)) {
-#if NET_2_0
 				if (value == null)
 					throw new ArgumentNullException ("value");
 				if (length == 0)
-#else
-				if (value == null || length == 0)
-#endif
 					return String.Empty;
 
 				enc = Encoding.Default;
@@ -2698,19 +2693,12 @@ namespace System
 					try {
 						memcpy (bytePtr, (byte*) (value + startIndex), length);
 					} catch (NullReferenceException) {
-#if !NET_2_0
-						if (!isDefaultEncoding)
-							throw;
-#endif
-
 						throw new ArgumentOutOfRangeException ("ptr", "Value, startIndex and length do not refer to a valid string.");
-#if NET_2_0
 					} catch (AccessViolationException) {
 						if (!isDefaultEncoding)
 							throw;
 
 						throw new ArgumentOutOfRangeException ("value", "Value, startIndex and length do not refer to a valid string.");
-#endif
 					}
 
 			// GetString () is called even when length == 0
@@ -3039,46 +3027,19 @@ namespace System
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		public extern String (char c, int count);
 
-//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-//		private extern static string InternalJoin (string separator, string[] value, int sIndex, int count);
-
-//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-//		private extern String InternalReplace (String oldValue, string newValue, CompareInfo comp);
-
-//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-//		private extern void InternalCopyTo (int sIndex, char[] dest, int destIndex, int count);
-
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern String[] InternalSplit (char[] separator, int count, int options);
 
-//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-//		private extern String InternalTrim (char[] chars, int typ);
-
-//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-//		private extern int InternalLastIndexOfAny (char [] anyOf, int sIndex, int count);
-
-//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-//		private extern String InternalPad (int width, char chr, bool right);
-
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		internal extern static String InternalAllocateStr (int length);
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		internal extern static void InternalStrcpy (String dest, int destPos, String src);
-
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		internal extern static void InternalStrcpy (String dest, int destPos, char[] chars);
-
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		internal extern static void InternalStrcpy (String dest, int destPos, String src, int sPos, int count);
-
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		internal extern static void InternalStrcpy (String dest, int destPos, char[] chars, int sPos, int count);
-
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern static string InternalIntern (string str);
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern static string InternalIsInterned (string str);
+
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		private extern static int GetLOSLimit ();
 	}
 }
