@@ -27,8 +27,12 @@
 // OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
-#if (NET_2_0||BOOTSTRAP_NET_2_0) && !NET_2_1
+#if !NET_2_1
+
+//
+// Defining this writes the output to console.log
 //#define DEBUG
+
 using System.Collections;
 using System.IO;
 using System.Text;
@@ -58,7 +62,6 @@ namespace System {
 		string term;
 		StreamReader stdin;
 		CStreamWriter stdout;
-		internal byte verase;
 
 		int windowWidth;
 		int windowHeight;
@@ -73,6 +76,7 @@ namespace System {
 		string keypadXmit, keypadLocal;
 		bool controlCAsInput;
 		bool inited;
+		object initLock = new object ();
 		bool initKeys;
 		string origPair;
 		string origColors;
@@ -89,6 +93,7 @@ namespace System {
 		ByteMatcher rootmap;
 		bool home_1_1; // if true, we have to add 1 to x and y when using cursorAddress
 		int rl_startx = -1, rl_starty = -1;
+		byte [] control_characters; // Indexed by ControlCharacters.XXXXXX
 #if DEBUG
 		StreamWriter logger;
 #endif
@@ -174,80 +179,85 @@ namespace System {
 			if (inited)
 				return;
 
-			/* This should not happen any more, since it is checked for in Console */
-			if (!ConsoleDriver.IsConsole)
-				throw new IOException ("Not a tty.");
-
-			inited = true;
-
-			ConsoleDriver.SetEcho (false);
-
-			string endString = null;
-			keypadXmit = reader.Get (TermInfoStrings.KeypadXmit);
-			keypadLocal = reader.Get (TermInfoStrings.KeypadLocal);
-			if (keypadXmit != null) {
-				WriteConsole (keypadXmit); // Needed to get the arrows working
-				if (keypadLocal != null)
-					endString += keypadLocal;
-			}
-
-			origPair = reader.Get (TermInfoStrings.OrigPair);
-			origColors = reader.Get (TermInfoStrings.OrigColors);
-			setfgcolor = MangleParameters (reader.Get (TermInfoStrings.SetAForeground));
-			setbgcolor = MangleParameters (reader.Get (TermInfoStrings.SetABackground));
-			// lighter fg colours are 90 -> 97 rather than 30 -> 37
-			setlfgcolor = color16 ? setfgcolor.Replace ("[3", "[9") : setfgcolor;
-			// lighter bg colours are 100 -> 107 rather than 40 -> 47
-			setlbgcolor = color16 ? setbgcolor.Replace ("[4", "[10") : setbgcolor;
-			string resetColors = (origColors == null) ? origPair : origColors;
-			if (resetColors != null)
-				endString += resetColors;
-
-			byte vsusp;
-			byte intr;
-			unsafe {
-				if (!ConsoleDriver.TtySetup (keypadXmit, endString, out verase, out vsusp, out intr, out native_terminal_size))
-					throw new IOException ("Error initializing terminal.");
-			}
-
-			stdin = new StreamReader (Console.OpenStandardInput (0), Console.InputEncoding);
-			clear = reader.Get (TermInfoStrings.ClearScreen);
-			bell = reader.Get (TermInfoStrings.Bell);
-			if (clear == null) {
-				clear = reader.Get (TermInfoStrings.CursorHome);
-				clear += reader.Get (TermInfoStrings.ClrEos);
-			}
-
-			csrVisible = reader.Get (TermInfoStrings.CursorNormal);
-			if (csrVisible == null)
-				csrVisible = reader.Get (TermInfoStrings.CursorVisible);
-
-			csrInvisible = reader.Get (TermInfoStrings.CursorInvisible);
-			if (term == "cygwin" || term == "linux" || (term != null && term.StartsWith ("xterm")) ||
-				term == "rxvt" || term == "dtterm") {
-				titleFormat = "\x1b]0;{0}\x7"; // icon + window title
-			} else if (term == "iris-ansi") {
-				titleFormat = "\x1bP1.y{0}\x1b\\"; // not tested
-			} else if (term == "sun-cmd") {
-				titleFormat = "\x1b]l{0}\x1b\\"; // not tested
-			}
-
-			cursorAddress = reader.Get (TermInfoStrings.CursorAddress);
-			if (cursorAddress != null) {
-				string result = cursorAddress.Replace ("%i", String.Empty);
-				home_1_1 = (cursorAddress != result);
-				cursorAddress = MangleParameters (result);
-			}
-
-			GetCursorPosition ();
+			lock (initLock){
+				if (inited)
+					return;
+				inited = true;
+				
+				/* This should not happen any more, since it is checked for in Console */
+				if (!ConsoleDriver.IsConsole)
+					throw new IOException ("Not a tty.");
+				
+				ConsoleDriver.SetEcho (false);
+				
+				string endString = null;
+				keypadXmit = reader.Get (TermInfoStrings.KeypadXmit);
+				keypadLocal = reader.Get (TermInfoStrings.KeypadLocal);
+				if (keypadXmit != null) {
+					WriteConsole (keypadXmit); // Needed to get the arrows working
+					if (keypadLocal != null)
+						endString += keypadLocal;
+				}
+				
+				origPair = reader.Get (TermInfoStrings.OrigPair);
+				origColors = reader.Get (TermInfoStrings.OrigColors);
+				setfgcolor = MangleParameters (reader.Get (TermInfoStrings.SetAForeground));
+				setbgcolor = MangleParameters (reader.Get (TermInfoStrings.SetABackground));
+				// lighter fg colours are 90 -> 97 rather than 30 -> 37
+				setlfgcolor = color16 ? setfgcolor.Replace ("[3", "[9") : setfgcolor;
+				// lighter bg colours are 100 -> 107 rather than 40 -> 47
+				setlbgcolor = color16 ? setbgcolor.Replace ("[4", "[10") : setbgcolor;
+				string resetColors = (origColors == null) ? origPair : origColors;
+				if (resetColors != null)
+					endString += resetColors;
+				
+				unsafe {
+					if (!ConsoleDriver.TtySetup (keypadXmit, endString, out control_characters, out native_terminal_size)){
+						control_characters = new byte [17];
+						native_terminal_size = null;
+						//throw new IOException ("Error initializing terminal.");
+					}
+				}
+				
+				stdin = new StreamReader (Console.OpenStandardInput (0), Console.InputEncoding);
+				clear = reader.Get (TermInfoStrings.ClearScreen);
+				bell = reader.Get (TermInfoStrings.Bell);
+				if (clear == null) {
+					clear = reader.Get (TermInfoStrings.CursorHome);
+					clear += reader.Get (TermInfoStrings.ClrEos);
+				}
+				
+				csrVisible = reader.Get (TermInfoStrings.CursorNormal);
+				if (csrVisible == null)
+					csrVisible = reader.Get (TermInfoStrings.CursorVisible);
+				
+				csrInvisible = reader.Get (TermInfoStrings.CursorInvisible);
+				if (term == "cygwin" || term == "linux" || (term != null && term.StartsWith ("xterm")) ||
+				    term == "rxvt" || term == "dtterm") {
+					titleFormat = "\x1b]0;{0}\x7"; // icon + window title
+				} else if (term == "iris-ansi") {
+					titleFormat = "\x1bP1.y{0}\x1b\\"; // not tested
+				} else if (term == "sun-cmd") {
+					titleFormat = "\x1b]l{0}\x1b\\"; // not tested
+				}
+				
+				cursorAddress = reader.Get (TermInfoStrings.CursorAddress);
+				if (cursorAddress != null) {
+					string result = cursorAddress.Replace ("%i", String.Empty);
+					home_1_1 = (cursorAddress != result);
+					cursorAddress = MangleParameters (result);
+				}
+				
+				GetCursorPosition ();
 #if DEBUG
-			logger.WriteLine ("noGetPosition: {0} left: {1} top: {2}", noGetPosition, cursorLeft, cursorTop);
-			logger.Flush ();
+				logger.WriteLine ("noGetPosition: {0} left: {1} top: {2}", noGetPosition, cursorLeft, cursorTop);
+				logger.Flush ();
 #endif
-			if (noGetPosition) {
-				WriteConsole (clear);
-				cursorLeft = 0;
-				cursorTop = 0;
+				if (noGetPosition) {
+					WriteConsole (clear);
+					cursorLeft = 0;
+					cursorTop = 0;
+				}
 			}
 		}
 
@@ -356,8 +366,10 @@ namespace System {
 				break;
 			case ConsoleKey.Tab:
 				int n = 8 - (cursorLeft % 8);
-				for (int i = 0; i < n; i++)
+				for (int i = 0; i < n; i++){
 					IncrementX ();
+				}
+				WriteConsole ("\t");
 				break;
 			case ConsoleKey.Clear:
 				WriteConsole (clear);
@@ -466,14 +478,23 @@ namespace System {
 		void GetCursorPosition ()
 		{
 			int row = 0, col = 0;
+			int b;
 
+			// First, get any data in the input buffer.  Merely reduces the likelyhood of getting an error
+			int inqueue = ConsoleDriver.InternalKeyAvailable (0);
+			while (inqueue-- > 0){
+				b = stdin.Read ();
+				AddToBuffer (b);
+			}
+
+			// Then try to probe for the cursor coordinates
 			WriteConsole ("\x1b[6n");
 			if (ConsoleDriver.InternalKeyAvailable (1000) <= 0) {
 				noGetPosition = true;
 				return;
 			}
 
-			int b = stdin.Read ();
+			b = stdin.Read ();
 			while (b != '\x1b') {
 				AddToBuffer (b);
 				if (ConsoleDriver.InternalKeyAvailable (100) <= 0)
@@ -560,8 +581,7 @@ namespace System {
 				if (!inited) {
 					Init ();
 				}
-
-				throw new NotSupportedException ();
+				return false;
 			}
 		}
 
@@ -659,7 +679,7 @@ namespace System {
 					Init ();
 				}
 
-				throw new NotSupportedException ();
+				return false;
 			}
 		}
 
@@ -701,12 +721,14 @@ namespace System {
 			}
 		}
 
+		//
+		// Requries that caller calls Init () if not !inited.
+		//
 		unsafe void CheckWindowDimensions ()
 		{
-			if (terminal_size == *native_terminal_size)
+			if (native_terminal_size == null || terminal_size == *native_terminal_size)
 				return;
 
-			terminal_size = *native_terminal_size;
 			if (*native_terminal_size == -1){
 				int c = reader.Get (TermInfoNumbers.Columns);
 				if (c != 0)
@@ -716,6 +738,7 @@ namespace System {
 				if (c != 0)
 					windowHeight = c;
 			} else {
+				terminal_size = *native_terminal_size;
 				windowWidth = terminal_size >> 16;
 				windowHeight = terminal_size & 0xffff;
 			}
@@ -854,23 +877,45 @@ namespace System {
 			bool shift = false;
 			bool ctrl = false;
 
-			if (n == 10) {
+			switch (n){
+			case 10:
 				key = ConsoleKey.Enter;
-			} else if (n == 8 || n == 9 || n == 12 || n == 13 || n == 19) {
+				break;
+			case 0x20:
+				key = ConsoleKey.Spacebar;
+				break;
+			case 45:
+				key = ConsoleKey.Subtract;
+				break;
+			case 43:
+				key = ConsoleKey.Add;
+				break;
+			case 47:
+				key = ConsoleKey.Divide;
+				break;
+			case 42:
+				key = ConsoleKey.Multiply;
+				break;
+			case 8: case 9: case 12: case 13: case 19:
 				/* Values in ConsoleKey */
-			} else if (n >= 1 && n <= 26) {
-				// For Ctrl-a to Ctrl-z.
-				ctrl = true;
-				key = ConsoleKey.A + n - 1;
-			} else if (n == 27) {
+				break;
+			case 27:
 				key = ConsoleKey.Escape;
-			} else if (n >= 'a' && n <= 'z') {
-				key = ConsoleKey.A - 'a' + n;
-			} else if (n >= 'A' && n <= 'Z') {
-				shift = true;
-			} else if (n >= '0' && n <= '9') {
-			} else {
-				key = 0;
+				break;
+				
+			default:
+				if (n >= 1 && n <= 26) {
+					// For Ctrl-a to Ctrl-z.
+					ctrl = true;
+					key = ConsoleKey.A + n - 1;
+				} else if (n >= 'a' && n <= 'z') {
+					key = ConsoleKey.A - 'a' + n;
+				} else if (n >= 'A' && n <= 'Z') {
+					shift = true;
+				} else if (n >= '0' && n <= '9') {
+				} else
+					key = 0;
+				break;
 			}
 
 			return new ConsoleKeyInfo (c, key, shift, alt, ctrl);
@@ -931,11 +976,11 @@ namespace System {
 						do {
 							AddToBuffer (stdin.Read ());
 						} while (ConsoleDriver.InternalKeyAvailable (0) > 0);
-					} else if (stdin.Peek () != -1) {
+					} else if (stdin.DataAvailable ()) {
 						do {
 							AddToBuffer (stdin.Read ());
-						} while (stdin.Peek () != -1);
-					} else {						
+						} while (stdin.DataAvailable ());
+					} else {
 						if ((o = GetKeyFromBuffer (false)) != null)
 							break;
 
@@ -959,7 +1004,7 @@ namespace System {
 		bool InputPending ()
 		{
 			// check if we've got pending input we can read immediately
-			return readpos < writepos || stdin.Peek () != -1;
+			return readpos < writepos || stdin.DataAvailable ();
 		}
 
 		char [] echobuf = null;
@@ -1112,11 +1157,15 @@ namespace System {
 
 			rl_startx = cursorLeft;
 			rl_starty = cursorTop;
+			char eof = (char) control_characters [ControlCharacters.EOF];
 
 			do {
 				key = ReadKeyInternal (out fresh);
 				echo = echo || fresh;
 				c = key.KeyChar;
+				// EOF -> Ctrl-D (EOT) pressed.
+				if (c == eof && c != 0 && builder.Length == 0)
+					return null;
 
 				if (key.Key != ConsoleKey.Enter) {
 					if (key.Key != ConsoleKey.Backspace) {
@@ -1167,9 +1216,7 @@ namespace System {
 				Init ();
 			}
 
-			if (bufferWidth == 0)
-				CheckWindowDimensions ();
-
+			CheckWindowDimensions ();
 			if (left < 0 || left >= bufferWidth)
 				throw new ArgumentOutOfRangeException ("left", "Value must be positive and below the buffer width.");
 
@@ -1211,9 +1258,10 @@ namespace System {
 		void CreateKeyMap ()
 		{
 			keymap = new Hashtable ();
+			
 			keymap [TermInfoStrings.KeyBackspace] = new ConsoleKeyInfo ('\0', ConsoleKey.Backspace, false, false, false);
 			keymap [TermInfoStrings.KeyClear] = new ConsoleKeyInfo ('\0', ConsoleKey.Clear, false, false, false);
-			// Delete character...
+ 			// Delete character...
 			keymap [TermInfoStrings.KeyDown] = new ConsoleKeyInfo ('\0', ConsoleKey.DownArrow, false, false, false);
 			keymap [TermInfoStrings.KeyF1] = new ConsoleKeyInfo ('\0', ConsoleKey.F1, false, false, false);
 			keymap [TermInfoStrings.KeyF10] = new ConsoleKeyInfo ('\0', ConsoleKey.F10, false, false, false);
@@ -1226,7 +1274,6 @@ namespace System {
 			keymap [TermInfoStrings.KeyF8] = new ConsoleKeyInfo ('\0', ConsoleKey.F8, false, false, false);
 			keymap [TermInfoStrings.KeyF9] = new ConsoleKeyInfo ('\0', ConsoleKey.F9, false, false, false);
 			keymap [TermInfoStrings.KeyHome] = new ConsoleKeyInfo ('\0', ConsoleKey.Home, false, false, false);
-
 			keymap [TermInfoStrings.KeyLeft] = new ConsoleKeyInfo ('\0', ConsoleKey.LeftArrow, false, false, false);
 			keymap [TermInfoStrings.KeyLl] = new ConsoleKeyInfo ('\0', ConsoleKey.NumPad1, false, false, false);
 			keymap [TermInfoStrings.KeyNpage] = new ConsoleKeyInfo ('\0', ConsoleKey.PageDown, false, false, false);
@@ -1271,13 +1318,79 @@ namespace System {
 			keymap [TermInfoStrings.KeyF22] = new ConsoleKeyInfo ('\0', ConsoleKey.F22, false, false, false);
 			keymap [TermInfoStrings.KeyF23] = new ConsoleKeyInfo ('\0', ConsoleKey.F23, false, false, false);
 			keymap [TermInfoStrings.KeyF24] = new ConsoleKeyInfo ('\0', ConsoleKey.F24, false, false, false);
-
-
 			// These were previously missing:
 			keymap [TermInfoStrings.KeyDc] = new ConsoleKeyInfo ('\0', ConsoleKey.Delete, false, false, false);
 			keymap [TermInfoStrings.KeyIc] = new ConsoleKeyInfo ('\0', ConsoleKey.Insert, false, false, false);
 		}
 
+		//
+		// The keys that we know about and use
+		//
+		static TermInfoStrings [] UsedKeys = {
+			TermInfoStrings.KeyBackspace,
+			TermInfoStrings.KeyClear,
+			TermInfoStrings.KeyDown,
+			TermInfoStrings.KeyF1,
+			TermInfoStrings.KeyF10,
+			TermInfoStrings.KeyF2,
+			TermInfoStrings.KeyF3,
+			TermInfoStrings.KeyF4,
+			TermInfoStrings.KeyF5,
+			TermInfoStrings.KeyF6,
+			TermInfoStrings.KeyF7,
+			TermInfoStrings.KeyF8,
+			TermInfoStrings.KeyF9,
+			TermInfoStrings.KeyHome,
+			TermInfoStrings.KeyLeft,
+			TermInfoStrings.KeyLl,
+			TermInfoStrings.KeyNpage,
+			TermInfoStrings.KeyPpage,
+			TermInfoStrings.KeyRight,
+			TermInfoStrings.KeySf,
+			TermInfoStrings.KeySr,
+			TermInfoStrings.KeyUp,
+			TermInfoStrings.KeyA1,
+			TermInfoStrings.KeyA3,
+			TermInfoStrings.KeyB2,
+			TermInfoStrings.KeyC1,
+			TermInfoStrings.KeyC3,
+			TermInfoStrings.KeyBtab,
+			TermInfoStrings.KeyBeg,
+			TermInfoStrings.KeyCopy,
+			TermInfoStrings.KeyEnd,
+			TermInfoStrings.KeyEnter,
+			TermInfoStrings.KeyHelp,
+			TermInfoStrings.KeyPrint,
+			TermInfoStrings.KeyUndo,
+			TermInfoStrings.KeySbeg,
+			TermInfoStrings.KeyScopy,
+			TermInfoStrings.KeySdc,
+			TermInfoStrings.KeyShelp,
+			TermInfoStrings.KeyShome,
+			TermInfoStrings.KeySleft,
+			TermInfoStrings.KeySprint,
+			TermInfoStrings.KeySright,
+			TermInfoStrings.KeySundo,
+			TermInfoStrings.KeyF11,
+			TermInfoStrings.KeyF12,
+			TermInfoStrings.KeyF13,
+			TermInfoStrings.KeyF14,
+			TermInfoStrings.KeyF15,
+			TermInfoStrings.KeyF16,
+			TermInfoStrings.KeyF17,
+			TermInfoStrings.KeyF18,
+			TermInfoStrings.KeyF19,
+			TermInfoStrings.KeyF20,
+			TermInfoStrings.KeyF21,
+			TermInfoStrings.KeyF22,
+			TermInfoStrings.KeyF23,
+			TermInfoStrings.KeyF24,
+
+			// These were missing
+			TermInfoStrings.KeyDc,
+			TermInfoStrings.KeyIc
+		};
+		
 		void InitKeys ()
 		{
 			if (initKeys)
@@ -1285,70 +1398,11 @@ namespace System {
 
 			CreateKeyMap ();
 			rootmap = new ByteMatcher ();
-			AddStringMapping (TermInfoStrings.KeyBackspace);
-			AddStringMapping (TermInfoStrings.KeyClear);
-			AddStringMapping (TermInfoStrings.KeyDown);
-			AddStringMapping (TermInfoStrings.KeyF1);
-			AddStringMapping (TermInfoStrings.KeyF10);
-			AddStringMapping (TermInfoStrings.KeyF2);
-			AddStringMapping (TermInfoStrings.KeyF3);
-			AddStringMapping (TermInfoStrings.KeyF4);
-			AddStringMapping (TermInfoStrings.KeyF5);
-			AddStringMapping (TermInfoStrings.KeyF6);
-			AddStringMapping (TermInfoStrings.KeyF7);
-			AddStringMapping (TermInfoStrings.KeyF8);
-			AddStringMapping (TermInfoStrings.KeyF9);
-			AddStringMapping (TermInfoStrings.KeyHome);
-			AddStringMapping (TermInfoStrings.KeyLeft);
-			AddStringMapping (TermInfoStrings.KeyLl);
-			AddStringMapping (TermInfoStrings.KeyNpage);
-			AddStringMapping (TermInfoStrings.KeyPpage);
-			AddStringMapping (TermInfoStrings.KeyRight);
-			AddStringMapping (TermInfoStrings.KeySf);
-			AddStringMapping (TermInfoStrings.KeySr);
-			AddStringMapping (TermInfoStrings.KeyUp);
-			AddStringMapping (TermInfoStrings.KeyA1);
-			AddStringMapping (TermInfoStrings.KeyA3);
-			AddStringMapping (TermInfoStrings.KeyB2);
-			AddStringMapping (TermInfoStrings.KeyC1);
-			AddStringMapping (TermInfoStrings.KeyC3);
-			AddStringMapping (TermInfoStrings.KeyBtab);
-			AddStringMapping (TermInfoStrings.KeyBeg);
-			AddStringMapping (TermInfoStrings.KeyCopy);
-			AddStringMapping (TermInfoStrings.KeyEnd);
-			AddStringMapping (TermInfoStrings.KeyEnter);
-			AddStringMapping (TermInfoStrings.KeyHelp);
-			AddStringMapping (TermInfoStrings.KeyPrint);
-			AddStringMapping (TermInfoStrings.KeyUndo);
-			AddStringMapping (TermInfoStrings.KeySbeg);
-			AddStringMapping (TermInfoStrings.KeyScopy);
-			AddStringMapping (TermInfoStrings.KeySdc);
-			AddStringMapping (TermInfoStrings.KeyShelp);
-			AddStringMapping (TermInfoStrings.KeyShome);
-			AddStringMapping (TermInfoStrings.KeySleft);
-			AddStringMapping (TermInfoStrings.KeySprint);
-			AddStringMapping (TermInfoStrings.KeySright);
-			AddStringMapping (TermInfoStrings.KeySundo);
-			AddStringMapping (TermInfoStrings.KeyF11);
-			AddStringMapping (TermInfoStrings.KeyF12);
-			AddStringMapping (TermInfoStrings.KeyF13);
-			AddStringMapping (TermInfoStrings.KeyF14);
-			AddStringMapping (TermInfoStrings.KeyF15);
-			AddStringMapping (TermInfoStrings.KeyF16);
-			AddStringMapping (TermInfoStrings.KeyF17);
-			AddStringMapping (TermInfoStrings.KeyF18);
-			AddStringMapping (TermInfoStrings.KeyF19);
-			AddStringMapping (TermInfoStrings.KeyF20);
-			AddStringMapping (TermInfoStrings.KeyF21);
-			AddStringMapping (TermInfoStrings.KeyF22);
-			AddStringMapping (TermInfoStrings.KeyF23);
-			AddStringMapping (TermInfoStrings.KeyF24);
 
-			// These were missing
-			AddStringMapping (TermInfoStrings.KeyDc);
-			AddStringMapping (TermInfoStrings.KeyIc);
-
-			rootmap.AddMapping (TermInfoStrings.KeyBackspace, new byte [] { verase });
+			foreach (TermInfoStrings tis in UsedKeys)
+				AddStringMapping (tis);
+			
+			rootmap.AddMapping (TermInfoStrings.KeyBackspace, new byte [] { control_characters [ControlCharacters.Erase] });
 			rootmap.Sort ();
 			initKeys = true;
 		}
