@@ -98,7 +98,9 @@ public class AssemblyNameTest {
 		0x1A, 0x56, 0x76, 0x43, 0xDB, 0x64, 0x86, 0x41, 0x64, 0x8D,
 		0x4C, 0x91, 0x83, 0x4E, 0xF5, 0x6C };
 
+#if !NET_2_0
 	static byte [] kp_token = { 0xff, 0xef, 0x94, 0x53, 0x67, 0x69, 0xda, 0x06 };
+#endif
 
 	static byte [] publicKey1 = {
 		0x00, 0x24, 0x00, 0x00, 0x04, 0x80, 0x00, 0x00, 0x94, 0x00,
@@ -138,7 +140,9 @@ public class AssemblyNameTest {
 		0x50, 0xa4, 0x4d, 0x8b, 0x8a, 0x58, 0x17, 0x70, 0xa4, 0x53,
 		0xe4, 0xdc, 0x73, 0x5d, 0x8c, 0x4e, 0xb8, 0xd3, 0xa9, 0xbf };
 
+#if !NET_2_0
 	static byte [] pk_token2 = { 0x22, 0x7c, 0x9c, 0x2c, 0x3c, 0x00, 0x63, 0xe9 };
+#endif
 
 	static byte [] publicKey3 = {
 		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00,
@@ -1338,7 +1342,7 @@ public class AssemblyNameTest {
 		const string assemblyVersion = "1.2.3.4";
 
 		an = new AssemblyName (assemblyName + ", Version=" + assemblyVersion + 
-				", Culture=" + assemblyCulture + ", PublicKeyToken=" + GetTokenString (pk_token1));
+				", Culture=" + assemblyCulture + ", PublicKeyToken=" + GetTokenString (pk_token1) + ",ProcessorArchitecture=X86");
 		Assert.IsNull (an.CodeBase, "CodeBase");
 		Assert.AreEqual (CultureInfo.InvariantCulture, an.CultureInfo, "CultureInfo");
 		Assert.IsNull (an.EscapedCodeBase, "EscapedCodeBase");
@@ -1348,7 +1352,7 @@ public class AssemblyNameTest {
 		Assert.AreEqual (AssemblyHashAlgorithm.None, an.HashAlgorithm, "HashAlgorithm");
 		Assert.IsNull (an.KeyPair, "KeyPair");
 		Assert.AreEqual (assemblyName, an.Name, "Name");
-		Assert.AreEqual (ProcessorArchitecture.None, an.ProcessorArchitecture, "PA");
+		Assert.AreEqual (ProcessorArchitecture.X86, an.ProcessorArchitecture, "PA");
 		Assert.AreEqual (new Version (assemblyVersion), an.Version, "Version");
 		Assert.AreEqual (AssemblyVersionCompatibility.SameMachine, 
 			an.VersionCompatibility, "VersionCompatibility");
@@ -1988,6 +1992,16 @@ public class AssemblyNameTest {
 			Assert.IsNotNull (ex.Message, "#4");
 		}
 	}
+
+	[Test (Description="Xamarin bug #99 - whitespaces in key=value")]
+	public void WhiteSpaceInKeyValue ()
+	{
+		string nameWithSpaces = String.Format ("MySql.Data.Tests, PublicKey      = \t  {0},  Culture   =\tneutral, Version=\t1.2.3.4", GetTokenString (publicKey1));
+		string fullName = "MySql.Data.Tests, Version=1.2.3.4, Culture=neutral, PublicKeyToken=ce5276d8687ec6dc";
+		var an = new AssemblyName (nameWithSpaces);
+
+		Assert.AreEqual (fullName, an.FullName);
+	}
 #endif
 }
 
