@@ -1389,18 +1389,107 @@ namespace MonoTests.System.Security.Cryptography {
 #endif
 		}
 
-		class ExpandTransform : ICryptoTransform {
+		[Test]
+		public void ReadModeDispose_FinalBlock ()
+		{
+			using (SHA1 sha1 = SHA1.Create()) {
+				using (MemoryStream mem = new MemoryStream(new byte[] { 1, 2, 3 }, false))
+					using (CryptoStream cs = new CryptoStream(mem, sha1, CryptoStreamMode.Read))
+					{
+					}
+				byte b = sha1.Hash [0]; // This will throw if TransformFinalBlock not called in sha1
+				GC.KeepAlive (b); // just the warning...
+			}
+                }
 
-			private bool can_transform_multiple_blocks;
-			private int output_block_size;
+		[Test]
+		public void CustomDisposeCalled ()
+		{
+			using (MemoryStream mem = new MemoryStream(new byte[] { 1, 2, 3 }, false)) {
+				MyCryptoStream cs;
+				using (cs = new MyCryptoStream (mem, SHA1.Create()))
+				{
+				}
+				Assert.IsTrue (cs.DisposeCalled, "#1");
+			}
+		}
+
+		[Test]
+		public void ExplicitFlush ()
+		{
+			// Tests that explicitly calling Flush does not call Flush in the underlying stream
+			MyStream ms = new MyStream ();
+			using (CryptoStream cs = new CryptoStream (ms, SHA1.Create (), CryptoStreamMode.Read)) {
+				ms.FlushCounterEnabled = true;
+				cs.Flush ();
+				ms.FlushCounterEnabled = false;
+			}
+			Assert.IsTrue (ms.FlushCounter == 0);
+		}
+
+		[Test]
+		public void ImplicitFlush ()
+		{
+			// Tests that Dispose() calls Flush on the underlying stream
+			MyStream ms = new MyStream ();
+			ms.FlushCounterEnabled = true;
+			using (CryptoStream cs = new CryptoStream (ms, SHA1.Create (), CryptoStreamMode.Read)) {
+			}
+			Assert.IsTrue (ms.FlushCounter == 1);
+		}
+
+		[Test]
+		public void ImplicitFlushCascade ()
+		{
+			// Tests that Dispose() calls FlushFinalBlock() on the underlying stream
+			MyStream ms = new MyStream ();
+			ms.FlushCounterEnabled = true;
+			CryptoStream cs1 = new CryptoStream (ms, SHA1.Create (), CryptoStreamMode.Read);
+			using (CryptoStream cs = new CryptoStream (cs1, SHA1.Create (), CryptoStreamMode.Read)) {
+			}
+			Assert.IsTrue (ms.FlushCounter == 1);
+		}
+
+		[Test]
+		[ExpectedException (typeof (ArgumentException))]
+		public void Ctor_InvalidEnumValue ()
+		{
+			CryptoStream cs = new CryptoStream (Stream.Null, SHA1.Create (), (CryptoStreamMode) 0xff);
+		}
+
+		[Test]
+		public void OutputBlock_Smaller ()
+		{
+			// The OutputBlockSize is smaller than the InputBlockSize
+			using (CryptoStream cs = new CryptoStream(Stream.Null, new MyCryptAlgorithm(), CryptoStreamMode.Write)) {
+				byte[] buffer = new byte[512 * 1024];
+				cs.Write(buffer, 0, buffer.Length);
+			}
+		}
+
+		class MyCryptoStream : CryptoStream {
+			public bool DisposeCalled { get; private set;}
+
+			public MyCryptoStream(Stream stream, ICryptoTransform transform)
+						: base(stream, transform, CryptoStreamMode.Read)
+			{
+			}
+
+			protected override void Dispose(bool disposing)
+			{
+				base.Dispose(disposing);
+				DisposeCalled = true;
+			}
+		}
+
+		class ExpandTransform : ICryptoTransform {
 
 			public bool CanReuseTransform {
 				get { return true; }
 			}
 
 			public bool CanTransformMultipleBlocks {
-				get { return can_transform_multiple_blocks; }
-				private set { can_transform_multiple_blocks = value; }
+				get; private set;
 			}
 
 			public int InputBlockSize {
@@ -1408,8 +1497,7 @@ namespace MonoTests.System.Security.Cryptography {
 			}
 
 			public int OutputBlockSize {
-				get { return output_block_size; }
-				private set { output_block_size = value; }
+				get; private set;
 			}
 
 			public ExpandTransform (bool canTranformMultipleBlocks, int outputBlockSize)
@@ -1424,9 +1512,9 @@ namespace MonoTests.System.Security.Cryptography {
 
 			public int TransformBlock (byte [] inputBuffer, int inputOffset, int inputCount, byte [] outputBuffer, int outputOffset)
 			{
-				int ret = 0;
-				for (int i = 0; i < inputCount; i++, inputOffset++) {
-					for (int j = 0; j < OutputBlockSize; j++, outputOffset++, ret++) {
+				var ret = 0;
+				for (var i = 0; i < inputCount; i++, inputOffset++) {
+					for (var j = 0; j < OutputBlockSize; j++, outputOffset++, ret++) {
 						outputBuffer [outputOffset] = inputBuffer [inputOffset];
 					}
 				}
@@ -1435,7 +1523,7 @@ namespace MonoTests.System.Security.Cryptography {
 
 			public byte [] TransformFinalBlock (byte [] inputBuffer, int inputOffset, int inputCount)
 			{
-				byte[] outputBuffer = new byte [inputCount * OutputBlockSize];
+				var outputBuffer = new byte [inputCount * OutputBlockSize];
 				TransformBlock (inputBuffer, inputOffset, inputCount, outputBuffer, 0);
 				return outputBuffer;
 			}
@@ -1456,12 +1544,12 @@ namespace MonoTests.System.Security.Cryptography {
 		public void Expand ()
 		{
 			int n = 0;
-			foreach (bool transformMultiple in new bool [] { false, true }) {
-				foreach (int outputBlockSize in new int [] { 1, 2, 3, 4 }) {
-					MemoryStream expantedStream = new MemoryStream ();
-					byte[] inputData = new byte [] { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07 };
+			foreach (var transformMultiple in new [] { false, true }) {
+				foreach (var outputBlockSize in new [] { 1, 2, 3, 4 }) {
+					var expantedStream = new MemoryStream ();
+					var inputData = new byte [] { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07 };
 
-					using (CryptoStream stream = new CryptoStream (expantedStream, new ExpandTransform (transformMultiple, outputBlockSize), CryptoStreamMode.Write)) {
+					using (var stream = new CryptoStream (expantedStream, new ExpandTransform (transformMultiple, outputBlockSize), CryptoStreamMode.Write)) {
 						stream.Write (inputData, 0, inputData.Length);
 					}
 					expantedStream.Close ();
@@ -1473,9 +1561,6 @@ namespace MonoTests.System.Security.Cryptography {
 		}
 
 		class CompressTransform : ICryptoTransform {
-
-			private int input_block_size;
-
 			public bool CanReuseTransform {
 				get { return true; }
 			}
@@ -1485,8 +1570,7 @@ namespace MonoTests.System.Security.Cryptography {
 			}
 
 			public int InputBlockSize {
-				get { return input_block_size; }
-				private set { input_block_size = value; }
+				get; private set;
 			}
 
 			public int OutputBlockSize {
@@ -1506,8 +1590,8 @@ namespace MonoTests.System.Security.Cryptography {
 
 			public int TransformBlock (byte [] inputBuffer, int inputOffset, int inputCount, byte [] outputBuffer, int outputOffset)
 			{
-				int ret = 0;
-				for (int i = 0; i < inputCount; i++, inputOffset++) {
+				var ret = 0;
+				for (var i = 0; i < inputCount; i++, inputOffset++) {
 					if (++bufferedCount == InputBlockSize) {
 						outputBuffer [outputOffset++] = inputBuffer [inputOffset];
 						ret++;
@@ -1519,8 +1603,8 @@ namespace MonoTests.System.Security.Cryptography {
 
 			public byte [] TransformFinalBlock (byte [] inputBuffer, int inputOffset, int inputCount)
 			{
-				byte[] outputBuffer = new byte [inputCount * OutputBlockSize];
-				int ret = TransformBlock (inputBuffer, inputOffset, inputCount, outputBuffer, 0);
+				var outputBuffer = new byte [inputCount * OutputBlockSize];
+				var ret = TransformBlock (inputBuffer, inputOffset, inputCount, outputBuffer, 0);
 				byte[] result = new byte [ret];
 				Array.Copy (outputBuffer, result, ret);
 				return result;
@@ -1538,21 +1622,115 @@ namespace MonoTests.System.Security.Cryptography {
 		public void Compress ()
 		{
 			int n = 0;
-			foreach (int inputBlockSize in new int [] { 1, 2, 3, 4 }) {
-				byte[] inputData = new byte [] {
+			foreach (var inputBlockSize in new [] { 1, 2, 3, 4 }) {
+				var inputData = new byte [] {
 					0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
 					0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
 					0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
 					0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
 				};
 
-				using (CryptoStream stream = new CryptoStream (new MemoryStream (inputData), new CompressTransform (inputBlockSize), CryptoStreamMode.Read)) {
-					byte[] buffer = new byte [inputData.Length];
-					int ret = stream.Read (buffer, 0, buffer.Length);
+				using (var stream = new CryptoStream (new MemoryStream (inputData), new CompressTransform (inputBlockSize), CryptoStreamMode.Read)) {
+					var buffer = new byte [inputData.Length];
+					var ret = stream.Read (buffer, 0, buffer.Length);
 					string value = BitConverter.ToString (buffer, 0, ret);
 					Assert.AreEqual (compress_values [n++], value);
 				}
 			}
 		}
+
+		class MyCryptAlgorithm : ICryptoTransform {
+			public bool CanReuseTransform { get { return true; } }
+			public bool CanTransformMultipleBlocks { get { return false; } }
+			public int InputBlockSize { get { return 128 * 1024; } }
+			public int OutputBlockSize { get { return 64 * 1024; } }
+
+			public int TransformBlock(byte[] inputBuffer, int inputOffset, int inputCount, byte[] outputBuffer, int outputOffset)
+			{
+				return this.OutputBlockSize;
+			}
+
+			public byte[] TransformFinalBlock(byte[] inputBuffer, int inputOffset, int inputCount)
+			{
+				return new byte[this.OutputBlockSize];
+			}
+
+			public void Dispose() {}
+		}
+
+		class MyStream : Stream {
+			public bool FlushCounterEnabled;
+			public int FlushCounter;
+
+			public override bool CanRead
+			{
+				get {
+					return true;
+				}
+			}
+
+			public override bool CanSeek
+			{
+				get {
+					return true;
+				}
+			}
+
+			public override bool CanWrite
+			{
+				get {
+					return true;
+				}
+			}
+
+			public override long Length
+			{
+				get {
+					return 0;
+				}
+			}
+
+			public override long Position
+			{
+				get {
+					return 0;
+				}
+				set {
+				}
+			}
+
+			public override void Flush ()
+			{
+				if (FlushCounterEnabled)
+					FlushCounter++;
+			}
+
+			public override int Read (byte[] buffer, int offset, int count)
+			{
+				return 0;
+			}
+
+			public override int ReadByte ()
+			{
+				return -1;
+			}
+
+			public override long Seek (long offset, SeekOrigin origin)
+			{
+				return 0;
+			}
+
+			public override void SetLength (long value)
+			{
+			}
+
+			public override void Write (byte[] buffer, int offset, int count)
+			{
+			}
+
+			public override void WriteByte (byte value)
+			{
+			}
+		}
 	}
 }
