@@ -96,6 +96,10 @@ struct _SgenThreadInfo {
 	thread_port_t mach_port;
 #endif
 	
+#if defined(PLATFORM_ANDROID)
+	void *android_tid;
+#endif
+
 	unsigned int stop_count; /* to catch duplicate signals */
 	int signal;
 	int skip;
@@ -593,6 +597,7 @@ int mono_sgen_thread_handshake (int signum) MONO_INTERNAL;
 SgenThreadInfo* mono_sgen_thread_info_lookup (ARCH_THREAD_TYPE id) MONO_INTERNAL;
 SgenThreadInfo** mono_sgen_get_thread_table (void) MONO_INTERNAL;
 gboolean mono_sgen_suspend_thread (SgenThreadInfo *info) MONO_INTERNAL;
+int mono_sgen_pthread_kill (SgenThreadInfo *info, int signum) MONO_INTERNAL;
 
 
 void mono_sgen_wait_for_suspend_ack (int count) MONO_INTERNAL;
